/****************************************************************************
*                   nexus.c
*
*  This module implements routines for the nexus object.
*
*  from Persistence of Vision(tm) Ray Tracer
*  Copyright 2000 J. Grimbert
*---------------------------------------------------------------------------
*  NOTICE: This source code file is provided so that users may experiment
*  with enhancements to POV-Ray and to port the software to platforms other
*  than those supported by the POV-Ray Team.  There are strict rules under
*  which you are permitted to use this file.  The rules are in the file
*  named POVLEGAL.DOC which should be distributed with this file.
*  If POVLEGAL.DOC is not available or for more info please contact the POV-Ray
*  Team Coordinator by email to team-coord@povray.org or visit us on the web at
*  http://www.povray.org. The latest version of POV-Ray may be found at this site.
*
* This program is based on the popular DKB raytracer version 2.12.
* DKBTrace was originally written by David K. Buck.
* DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
*
*****************************************************************************/

/****************************************************************************
*
*  Explanation:
*
*    -
*
*  Syntax:
* 
*  nexus
*  {
*     object surface
*  }
* 
* 
*
*****************************************************************************/


#include "frame.h"
#include "povray.h"
#include "vector.h"
#include "povproto.h"
#include "bbox.h"
#include "csg.h"
#include "hfield.h"
#include "matrices.h"
#include "objects.h"
#include "planes.h"
#include "quadrics.h"
#include "nexus.h"



/*****************************************************************************
* Local preprocessor defines
******************************************************************************/




/*****************************************************************************
* Static functions
******************************************************************************/

static int All_NEXUS_Intersections (OBJECT *Object, RAY *Ray, ISTACK *Depth_Stack);
static int Inside_NEXUS (VECTOR point, OBJECT *Object);
static NEXUS *Copy_NEXUS (OBJECT *Object);
static void NEXUS_Normal (VECTOR Result, OBJECT *Object, INTERSECTION *Inter); 
static void Translate_NEXUS (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans);
static void Rotate_NEXUS (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans);
static void Scale_NEXUS (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans);
static void Transform_NEXUS (OBJECT *Object, TRANSFORM *Trans);
static void Destroy_NEXUS (OBJECT *Object);
static void Invert_NEXUS (OBJECT *Object);


/*****************************************************************************
* Local variables
******************************************************************************/

METHODS NEXUS_Methods =
{
  All_NEXUS_Intersections,
  Inside_NEXUS, NEXUS_Normal ,
  (COPY_METHOD)Copy_NEXUS,
  Translate_NEXUS, Rotate_NEXUS,
  Scale_NEXUS, Transform_NEXUS, Invert_NEXUS, Destroy_NEXUS
};

/*****************************************************************************
*
* FUNCTION
*
*   All_NEXUS_Intersections
*
* INPUT
*   
* OUTPUT
*   
* RETURNS
*   
* AUTHOR
*
*   J. Grimbert
*   
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static int All_NEXUS_Intersections (OBJECT *Object, RAY *Ray, ISTACK *Depth_Stack)
{
  int Found, Count;
  OBJECT *First_Sib;
  ISTACK *Local_Stack;
  INTERSECTION *Sibling_Intersection;
  VECTOR Norv,Dire;
  DBL ang;

  Increase_Counter(stats[Ray_NEXUS_Tests]);

  Count = Found = FALSE;
  Assign_Vector(Dire, Ray->Direction);
  if (Test_Flag(Object, INVERTED_FLAG))
  {
    VScaleEq(Dire,-1);
  }

  Local_Stack = open_istack();
  First_Sib = ((NEXUS *)Object)->Children;
  if (Ray_In_Bound (Ray, First_Sib->Bound))
  {
    if (All_Intersections (First_Sib, Ray, Local_Stack))
    {
      while ((Sibling_Intersection = pop_entry(Local_Stack)) != NULL)
      {
        if (Point_In_Clip (Sibling_Intersection->IPoint, Object->Clip))
        {
          Normal(Norv, First_Sib, Sibling_Intersection);
          VDot(ang, Norv, Dire);
          if (ang > 0)
          {
            itop(Depth_Stack).Depth = Sibling_Intersection->Depth;
            itop(Depth_Stack).Object = Sibling_Intersection->Object;
            Assign_Vector(itop(Depth_Stack).IPoint, 
                           Sibling_Intersection->IPoint);
            Assign_Vector(itop(Depth_Stack).INormal, 
                           Sibling_Intersection->INormal);
            itop(Depth_Stack).i1 = Sibling_Intersection->i1;
            itop(Depth_Stack).d1 = Sibling_Intersection->d1;
            itop(Depth_Stack).i2 = Sibling_Intersection->i2;
            itop(Depth_Stack).d2 = Sibling_Intersection->d2;
            itop(Depth_Stack).Pointer = Sibling_Intersection->Pointer;
            incstack(Depth_Stack);
            Found = TRUE;
            Count++;
          }
        }
      }
    }
  }
  close_istack (Local_Stack);

  if (Count)
  {
    Increase_Counter(stats[Ray_NEXUS_Tests_Succeeded]);
  }

  return (Count);
}

/*****************************************************************************
*
* FUNCTION
*
*   Inside_NEXUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static int Inside_NEXUS (VECTOR IPoint, OBJECT *Object)
{
  OBJECT *child;
  child = ((NEXUS *)Object)->Children;
  return (Inside_Object(IPoint,child));
}

static void NEXUS_Normal(VECTOR Result, OBJECT *Object, INTERSECTION *Inter)
{
  Normal(Result, Inter->Object, Inter);  
}                                                                               

/*****************************************************************************
*
* FUNCTION
*
*   Translate_NEXUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Translate_NEXUS (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans)
{
  OBJECT *Sib;

  for (Sib = ((NEXUS *) Object)->Children; Sib != NULL; Sib = Sib->Sibling)
  {
    Translate_Object(Sib, Vector, Trans);
  }

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Rotate_NEXUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Rotate_NEXUS (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans)
{
  OBJECT *Sib;

  for (Sib = ((NEXUS *) Object)->Children; Sib != NULL; Sib = Sib->Sibling)
  {
    Rotate_Object(Sib, Vector, Trans);
  }

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Scale_NEXUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Scale_NEXUS (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans)
{
  OBJECT *Sib;

  for (Sib = ((NEXUS *) Object)->Children; Sib != NULL; Sib = Sib->Sibling)
  {
    Scale_Object(Sib, Vector, Trans);
  }

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Transform_NEXUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Transform_NEXUS (OBJECT *Object, TRANSFORM *Trans)
{
  OBJECT *Sib;

  for (Sib = ((NEXUS *) Object)->Children; Sib != NULL; Sib = Sib->Sibling)
  {
    Transform_Object (Sib, Trans);
  }

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Invert_NEXUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Invert_NEXUS (OBJECT *Object)
{
  Invert_Flag(Object, INVERTED_FLAG);  
}



/*****************************************************************************
*
* FUNCTION
*
*   Create_NEXUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

NEXUS *Create_NEXUS ()
{
  NEXUS *New;

  New = (NEXUS *)POV_MALLOC(sizeof (NEXUS), "nexus");

  INIT_OBJECT_FIELDS(New, NEXUS_OBJECT, &NEXUS_Methods)

  New->Children = NULL;

  return (New);
  
}


/*****************************************************************************
*
* FUNCTION
*
*   Copy_NEXUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static NEXUS *Copy_NEXUS (OBJECT *Object)
{
  NEXUS *New;
  OBJECT *Old_Sib, *New_Sib, *Prev_Sib;

  New = (NEXUS *)POV_MALLOC(sizeof (NEXUS), "nexus");

  *New = *(NEXUS *)Object;

  New->Children = Prev_Sib = NULL;

  for (Old_Sib = ((NEXUS *)Object)->Children; Old_Sib != NULL; Old_Sib = Old_Sib->Sibling)
  {
    New_Sib = Copy_Object (Old_Sib);

    if (New->Children == NULL)
    {
      New->Children = New_Sib;
    }

    if (Prev_Sib != NULL)
    {
      Prev_Sib->Sibling = New_Sib;
    }

    Prev_Sib = New_Sib;
  }

  return (New);
}



/*****************************************************************************
*
* FUNCTION
*
*   Destroy_NEXUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Destroy_NEXUS (OBJECT *Object)
{
  Destroy_Object (((NEXUS *) Object)->Children);
  POV_FREE (Object);
}



/*****************************************************************************
*
* FUNCTION
*
*   Compute_NEXUS_BBox
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

void Compute_NEXUS_BBox (OBJECT *Object)
{
  Make_BBox(Object->BBox, ((NEXUS *)Object)->Children->BBox.Lower_Left[X],
   ((NEXUS *)Object)->Children->BBox.Lower_Left[Y],
   ((NEXUS *)Object)->Children->BBox.Lower_Left[Z],
   ((NEXUS *)Object)->Children->BBox.Lengths[X],
   ((NEXUS *)Object)->Children->BBox.Lengths[Y],
   ((NEXUS *)Object)->Children->BBox.Lengths[Z]);
}
