/****************************************************************************
*                   ovus.c
*
*  This module implements routines for the ovus object.
*
*  This module was written by Jerome Grimbert.
*
*  from Persistence of Vision(tm) Ray Tracer
*  Copyright 2002 J. Grimbert
*---------------------------------------------------------------------------
*  NOTICE: This source code file is provided so that users may experiment
*  with enhancements to POV-Ray and to port the software to platforms other
*  than those supported by the POV-Ray Team.  There are strict rules under
*  which you are permitted to use this file.  The rules are in the file
*  named POVLEGAL.DOC which should be distributed with this file.
*  If POVLEGAL.DOC is not available or for more info please contact the POV-Ray
*  Team Coordinator by email to team-coord@povray.org or visit us on the web at
*  http://www.povray.org. The latest version of POV-Ray may be found at this site.
*
* This program is based on the popular DKB raytracer version 2.12.
* DKBTrace was originally written by David K. Buck.
* DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
*
*****************************************************************************/

/****************************************************************************
*
*  Explanation:
*
*    -
*
*  Syntax:
* 
*  ovus
*  {
*     bottom_radius,top_radius
*  }
* 
*  The so long awaited 'Egg' forms.
* 
*  Normally, the bottom_radius is bigger than the top_radius
*  the center of the top sphere is at the zenith of the bottom sphere
*  and the bottom sphere is centered at 0.0
*  The radius of the connecting surface is the double of the biggest radius
*   (yes, the biggest diameter is used as the curvature of the connection)
*
*  The hard part was just to find where the connection starts and ends.
*
*  The name is ovus because:
*    - it's rather like torus (similar interface pushs for similar ending)
*    - the 2D curve is called 'ove' (latin for egg)
*
*****************************************************************************/


#include "frame.h"
#include "povray.h"
#include "vector.h"
#include "povproto.h"
#include "bbox.h"
#include "matrices.h"
#include "objects.h"
#include "polysolv.h"
#include "ovus.h"



/*****************************************************************************
* Local preprocessor defines
******************************************************************************/

#define ROOT_TOLERANCE 1.0e-4



/*****************************************************************************
* Static functions
******************************************************************************/

static int All_OVUS_Intersections(OBJECT * Object, RAY * Ray,
                                  ISTACK * Depth_Stack);
static int Inside_OVUS(VECTOR point, OBJECT * Object);
static OVUS *Copy_OVUS(OBJECT * Object);
static void OVUS_Normal(VECTOR Result, OBJECT * Object, INTERSECTION * Inter);
static void Translate_OVUS(OBJECT * Object, VECTOR Vector, TRANSFORM * Trans);
static void Rotate_OVUS(OBJECT * Object, VECTOR Vector, TRANSFORM * Trans);
static void Scale_OVUS(OBJECT * Object, VECTOR Vector, TRANSFORM * Trans);
static void Transform_OVUS(OBJECT * Object, TRANSFORM * Trans);
static void Destroy_OVUS(OBJECT * Object);
static void Invert_OVUS(OBJECT * Object);


/*****************************************************************************
* Local variables
******************************************************************************/

METHODS OVUS_Methods = {
   All_OVUS_Intersections,
   Inside_OVUS, OVUS_Normal,
   (COPY_METHOD) Copy_OVUS,
   Translate_OVUS, Rotate_OVUS,
   Scale_OVUS, Transform_OVUS, Invert_OVUS, Destroy_OVUS
};


static void Intersect_Ovus_Spheres(OVUS * Ovus, RAY * Ray, DBL radius1,
                                   DBL radius2, DBL xp, DBL yp, DBL rad,
                                   DBL * Depth1, DBL * Depth2, DBL * Depth3,
                                   DBL * Depth4, DBL * Depth5, DBL * Depth6)
{
   DBL OCSquared, t_Closest_Approach, Half_Chord, t_Half_Chord_Squared;
   VECTOR Padj, Ipoint;
   VECTOR IPoint;
   DBL R2, r2, Py2, Dy2, PDy2, k1, k2, horizontal, vertical;
   DBL Rad1, Rad2;
   int n;
   int lcount=0;
   VECTOR Second_Center;
   DBL c[5], r[4];

   *Depth1 = *Depth2 = *Depth3 = *Depth4 = *Depth5 = *Depth6 = -100;
   /* no hit unless... */
   VScale(Padj, Ray->Initial, -1);
   Rad1 = Sqr(radius1);
   Rad2 = Sqr(radius2);
   VDot(OCSquared, Padj, Padj);

   VDot(t_Closest_Approach, Padj, Ray->Direction);

   if ((OCSquared < Rad1) || (t_Closest_Approach > EPSILON))
   {

      t_Half_Chord_Squared = Rad1 - OCSquared + Sqr(t_Closest_Approach);

      if (t_Half_Chord_Squared > EPSILON)
      {
         Half_Chord = sqrt(t_Half_Chord_Squared);

         *Depth1 = t_Closest_Approach - Half_Chord;
         *Depth2 = t_Closest_Approach + Half_Chord;
      VEvaluateRay(IPoint, Ray->Initial, *Depth1, Ray->Direction);
      if (IPoint[Y] < Ovus->by)
      {
         lcount++;
      }
      VEvaluateRay(IPoint, Ray->Initial, *Depth2, Ray->Direction);
      if (IPoint[Y] < Ovus->by)
      {
         lcount++;
      }
      }
   }
   if (lcount > 1) return;
   Make_Vector(Second_Center, 0, radius1, 0);
   VSub(Padj, Second_Center, Ray->Initial);

   VDot(OCSquared, Padj, Padj);

   VDot(t_Closest_Approach, Padj, Ray->Direction);

   if ((OCSquared < Rad2) || (t_Closest_Approach > EPSILON))
   {

      t_Half_Chord_Squared = Rad2 - OCSquared + Sqr(t_Closest_Approach);

      if (t_Half_Chord_Squared > EPSILON)
      {
         Half_Chord = sqrt(t_Half_Chord_Squared);

         *Depth3 = t_Closest_Approach - Half_Chord;
         *Depth4 = t_Closest_Approach + Half_Chord;
      VEvaluateRay(IPoint, Ray->Initial, *Depth3, Ray->Direction);
      if (IPoint[Y] > Ovus->ty)
      {
         lcount++;
      }
      VEvaluateRay(IPoint, Ray->Initial, *Depth4, Ray->Direction);
      if (IPoint[Y] > Ovus->ty)
      {
         lcount++;
      }

      }
   }
   if (lcount > 1) return;
   Make_Vector(Second_Center, 0, yp, 0);
   VSub(Padj, Ray->Initial, Second_Center);
   R2 = Sqr(xp);
   r2 = Sqr(rad);               /* Notice : rad > xp here ! */

   Py2 = Padj[Y] * Padj[Y];
   Dy2 = Ray->Direction[Y] * Ray->Direction[Y];
   PDy2 = Padj[Y] * Ray->Direction[Y];

   k1 = Padj[X] * Padj[X] + Padj[Z] * Padj[Z] + Py2 - R2 - r2;
   k2 = Padj[X] * Ray->Direction[X] + Padj[Z] * Ray->Direction[Z] + PDy2;
   /* this is just like a big torus */
   c[0] = 1.0;

   c[1] = 4.0 * k2;

   c[2] = 2.0 * (k1 + 2.0 * (k2 * k2 + R2 * Dy2));

   c[3] = 4.0 * (k2 * k1 + 2.0 * R2 * PDy2);

   c[4] = k1 * k1 + 4.0 * R2 * (Py2 - r2);

   n = Solve_Polynomial(4, c, r, Test_Flag(Ovus, STURM_FLAG), ROOT_TOLERANCE);
   while (n--)
   {
      /* here we only keep the 'lemon' inside the torus 
      ** and dismiss the 'apple' 
      */
      /* 
      ** If you find a solution to resolve the rotation of
      **   (x + r)^2 + y^2 = R^2 around y (so replacing x by sqrt(x^2+z^2))
      ** with something which is faster than a 4th degree polynome,
      ** please feel welcome to update and share...
      */
      VEvaluateRay(Ipoint, Ray->Initial, r[n], Ray->Direction);

      vertical = Ipoint[Y];
      if ((vertical > Ovus->by) && (vertical < Ovus->ty))
      {
         horizontal = sqrt(Sqr(Ipoint[X]) + Sqr(Ipoint[Z]));
         OCSquared = Sqr(horizontal + Ovus->xp) + Sqr((vertical - Ovus->yp));
         if (fabs(OCSquared - Sqr(Ovus->cr)) < ROOT_TOLERANCE)
         {
            if (*Depth5 < 0)
            {
               *Depth5 = r[n];
            }
            else
            {
               *Depth6 = r[n];
            }
         }
      }
   }
}

/*****************************************************************************
*
* FUNCTION
*
*   All_OVUS_Intersections
*
* INPUT
*   
* OUTPUT
*   
* RETURNS
*   
* AUTHOR
*
*   J. Grimbert
*   
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static int All_OVUS_Intersections(OBJECT * Object, RAY * Ray,
                                  ISTACK * Depth_Stack)
{
   OVUS *Ovus = (OVUS *) Object;
   int Count = 0;
   VECTOR Real_Normal, Real_Pt, INormal, IPoint;
   DBL Depth1, Depth2, Depth3, Depth4, Depth5, Depth6;
   DBL len, horizontal;
   RAY New_Ray;

   Increase_Counter(stats[Ray_OVUS_Tests]);
   MInvTransPoint(New_Ray.Initial, Ray->Initial, Ovus->Trans);
   MInvTransDirection(New_Ray.Direction, Ray->Direction, Ovus->Trans);
   VLength(len, New_Ray.Direction);
   VInverseScaleEq(New_Ray.Direction, len);

   /* Bottom sphere */
   Intersect_Ovus_Spheres(Ovus, &New_Ray, Ovus->br, Ovus->tr, Ovus->xp,
                          Ovus->yp, Ovus->cr, &Depth1, &Depth2, &Depth3,
                          &Depth4, &Depth5, &Depth6);
   if (Depth1 > EPSILON)
   {
      VEvaluateRay(IPoint, New_Ray.Initial, Depth1, New_Ray.Direction);
      if (IPoint[Y] < Ovus->by)
      {
         MTransPoint(Real_Pt, IPoint, Ovus->Trans);
         if (Point_In_Clip(Real_Pt, Object->Clip))
         {
            Assign_Vector(INormal, IPoint);
            VInverseScaleEq(INormal, Ovus->br);
            MTransNormal(Real_Normal, INormal, Ovus->Trans);
            VNormalizeEq(Real_Normal);
            push_normal_entry(Depth1 / len, Real_Pt, Real_Normal, Object,
                              Depth_Stack);
            Count++;
         }
      }
   }

   if (Depth2 > EPSILON)
   {
      VEvaluateRay(IPoint, New_Ray.Initial, Depth2, New_Ray.Direction);

      if (IPoint[Y] < Ovus->by)
      {
         MTransPoint(Real_Pt, IPoint, Ovus->Trans);
         if (Point_In_Clip(Real_Pt, Object->Clip))
         {
            Assign_Vector(INormal, IPoint);
            VInverseScaleEq(INormal, Ovus->br);
            MTransNormal(Real_Normal, INormal, Ovus->Trans);
            VNormalizeEq(Real_Normal);
            push_normal_entry(Depth2 / len, Real_Pt, Real_Normal, Object,
                              Depth_Stack);
            Count++;
         }
      }
   }

   if (Depth3 > EPSILON)
   {
      VEvaluateRay(IPoint, New_Ray.Initial, Depth3, New_Ray.Direction);

      if (IPoint[Y] > Ovus->ty)
      {
         MTransPoint(Real_Pt, IPoint, Ovus->Trans);
         if (Point_In_Clip(Real_Pt, Object->Clip))
         {
            Assign_Vector(INormal, IPoint);
            INormal[Y] -= Ovus->br;
            VInverseScaleEq(INormal, Ovus->tr);
            MTransNormal(Real_Normal, INormal, Ovus->Trans);
            VNormalizeEq(Real_Normal);
            push_normal_entry(Depth3 / len, Real_Pt, Real_Normal, Object,
                              Depth_Stack);
            Count++;
         }
      }
   }
   if (Depth4 > EPSILON)
   {
      VEvaluateRay(IPoint, New_Ray.Initial, Depth4, New_Ray.Direction);

      if (IPoint[Y] > Ovus->ty)
      {
         MTransPoint(Real_Pt, IPoint, Ovus->Trans);
         if (Point_In_Clip(Real_Pt, Object->Clip))
         {
            Assign_Vector(INormal, IPoint);
            INormal[Y] -= Ovus->br;
            VInverseScaleEq(INormal, Ovus->tr);
            MTransNormal(Real_Normal, INormal, Ovus->Trans);
            VNormalizeEq(Real_Normal);
            push_normal_entry(Depth4 / len, Real_Pt, Real_Normal, Object,
                              Depth_Stack);
            Count++;
         }
      }
   }

   if (Depth5 > EPSILON)
   {
      VEvaluateRay(IPoint, New_Ray.Initial, Depth5, New_Ray.Direction);
      MTransPoint(Real_Pt, IPoint, Ovus->Trans);
      if (Point_In_Clip(Real_Pt, Object->Clip))
      {
         Assign_Vector(INormal, IPoint);
         
         INormal[Y] -= Ovus->yp;
         horizontal = sqrt(Sqr(INormal[X]) + Sqr(INormal[Z]));
         INormal[X] += (INormal[X] * Ovus->xp / horizontal);
         INormal[Z] += (INormal[Z] * Ovus->xp / horizontal);
         VNormalizeEq(INormal);
         MTransNormal(Real_Normal, INormal, Ovus->Trans);
         VNormalizeEq(Real_Normal);
         push_normal_entry(Depth5 / len, Real_Pt, Real_Normal, Object,
                           Depth_Stack);
         Count++;
      }
   }
   if (Depth6 > EPSILON)
   {
      VEvaluateRay(IPoint, New_Ray.Initial, Depth6, New_Ray.Direction);
      MTransPoint(Real_Pt, IPoint, Ovus->Trans);
      MTransNormal(Real_Normal, INormal, Ovus->Trans);

      if (Point_In_Clip(Real_Pt, Object->Clip))
      {
         Assign_Vector(INormal, IPoint);
         INormal[Y] -= Ovus->yp;
         horizontal = sqrt(Sqr(INormal[X]) + Sqr(INormal[Z]));
         INormal[X] += (INormal[X] * Ovus->xp / horizontal);
         INormal[Z] += (INormal[Z] * Ovus->xp / horizontal);
         VNormalizeEq(INormal);
         MTransNormal(Real_Normal, INormal, Ovus->Trans);
         VNormalizeEq(Real_Normal);
         push_normal_entry(Depth6 / len, Real_Pt, Real_Normal, Object,
                           Depth_Stack);
         Count++;
      }
   }
   if (Count)
   {
      Increase_Counter(stats[Ray_OVUS_Tests_Succeeded]);
   }

   return (Count);
}

/*****************************************************************************
*
* FUNCTION
*
*   Inside_OVUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static int Inside_OVUS(VECTOR IPoint, OBJECT * Object)
{
   OVUS *Ovus = (OVUS *) Object;
   DBL OCSquared;
   DBL horizontal, vertical;
   int INSide = 0;
   VECTOR Origin, New_Point, Other;
   Make_Vector(Origin, 0, Ovus->br, 0);
   MInvTransPoint(New_Point, IPoint, Ovus->Trans);
   VDot(OCSquared, New_Point, New_Point);
   if (OCSquared < Sqr(Ovus->br))
   {
      INSide = 1;
   }
   VSub(Other, New_Point, Origin);
   VDot(OCSquared, Other, Other);
   if (OCSquared < Sqr(Ovus->tr))
   {
      INSide = 1;
   }
   vertical = New_Point[Y];
   if ((vertical > Ovus->by) && (vertical < Ovus->ty))
   {
      horizontal = sqrt(Sqr(New_Point[X]) + Sqr(New_Point[Z]));
      OCSquared = Sqr(horizontal + Ovus->xp) + Sqr((vertical - Ovus->yp));
      if (OCSquared < Sqr(Ovus->cr))
      {
         INSide = 1;
      }
   }
   if (Test_Flag(Object, INVERTED_FLAG))
   {
      return !INSide;
   }
   else
   {
      return INSide;
   }
}

static void OVUS_Normal(VECTOR Result, OBJECT * Object, INTERSECTION * Inter)
{
   Assign_Vector(Result, Inter->INormal);
}

/*****************************************************************************
*
* FUNCTION
*
*   Translate_OVUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Translate_OVUS(OBJECT * Object, VECTOR Vector, TRANSFORM * Trans)
{
   Compose_Transforms(((OVUS *) Object)->Trans, Trans);
   Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Rotate_OVUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Rotate_OVUS(OBJECT * Object, VECTOR Vector, TRANSFORM * Trans)
{
   Compose_Transforms(((OVUS *) Object)->Trans, Trans);
   Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Scale_OVUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Scale_OVUS(OBJECT * Object, VECTOR Vector, TRANSFORM * Trans)
{
   Compose_Transforms(((OVUS *) Object)->Trans, Trans);
   Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Transform_OVUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Transform_OVUS(OBJECT * Object, TRANSFORM * Trans)
{
   Compose_Transforms(((OVUS *) Object)->Trans, Trans);
   Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Invert_OVUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Invert_OVUS(OBJECT * Object)
{
   Invert_Flag(Object, INVERTED_FLAG);
}



/*****************************************************************************
*
* FUNCTION
*
*   Create_OVUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

OVUS *Create_OVUS()
{
   OVUS *New;

   New = (OVUS *) POV_MALLOC(sizeof(OVUS), "ovus");

   INIT_OBJECT_FIELDS(New, OVUS_OBJECT, &OVUS_Methods) New->Trans =
      Create_Transform();

   New->br = 0.0;
   New->tr = 0.0;
   New->xp = 0.0;
   New->yp = 0.0;
   New->by = 0.0;
   New->ty = 0.0;
   New->cr = 0.0;
   return (New);

}


/*****************************************************************************
*
* FUNCTION
*
*   Copy_OVUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static OVUS *Copy_OVUS(OBJECT * Object)
{
   OVUS *New, *Ovus = (OVUS *) Object;

   New = Create_OVUS();

   Destroy_Transform(New->Trans);

   *New = *Ovus;

   New->Trans = Copy_Transform(Ovus->Trans);

   return (New);
}



/*****************************************************************************
*
* FUNCTION
*
*   Destroy_OVUS
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Destroy_OVUS(OBJECT * Object)
{
   Destroy_Transform(((OVUS *) Object)->Trans);
   POV_FREE(Object);
}



/*****************************************************************************
*
* FUNCTION
*
*   Compute_OVUS_BBox
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

void Compute_OVUS_BBox(OBJECT * Object)
{
   OVUS *Ovus = (OVUS *) Object;
   DBL biggest;
   biggest = Ovus->cr - Ovus->xp;
   if (biggest < Ovus->br)
   {
      biggest = Ovus->br;
   }
   if (biggest < Ovus->tr)
   {
      biggest = Ovus->tr;
   }
   Make_BBox(Object->BBox, -biggest, -Ovus->br , -biggest, 2 * biggest,
             2 * Ovus->br + Ovus->tr, 2 * biggest);
   Recompute_BBox(&Ovus->BBox, Ovus->Trans);
}
