/**************************************************************************
*                warps.c
*
*  This module implements functions that warp or modify the point at which
*  a texture pattern is evaluated.
*
*  from Persistence of Vision(tm) Ray Tracer
*  Copyright 1996,1999 Persistence of Vision Team
*---------------------------------------------------------------------------
*  NOTICE: This source code file is provided so that users may experiment
*  with enhancements to POV-Ray and to port the software to platforms other
*  than those supported by the POV-Ray Team.  There are strict rules under
*  which you are permitted to use this file.  The rules are in the file
*  named POVLEGAL.DOC which should be distributed with this file.
*  If POVLEGAL.DOC is not available or for more info please contact the POV-Ray
*  Team Coordinator by email to team-coord@povray.org or visit us on the web at
*  http://www.povray.org. The latest version of POV-Ray may be found at this site.
*
* This program is based on the popular DKB raytracer version 2.12.
* DKBTrace was originally written by David K. Buck.
* DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
*
*****************************************************************************/

#include "frame.h"
#include "vector.h"
#include "povproto.h"
#include "matrices.h"
#include "warps.h"
#include "pattern.h"
#include "texture.h"

/*****************************************************************************
* Local preprocessor defines
******************************************************************************/

#define COORDINATE_LIMIT 1.0e17


/*****************************************************************************
* Static functions
******************************************************************************/



/*****************************************************************************
*
* FUNCTION
*
*   Warp_EPoint
*
* INPUT
*
*   EPoint -- The original point in 3d space at which a pattern
*   is evaluated.
*   TPat   -- Texture pattern struct
*   
* OUTPUT
*
*   TPoint -- Point after turbulence and transform
*   have been applied
*   
* RETURNS
*   
* AUTHOR
*
*   POV-Ray Team
*   
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Warp_EPoint (VECTOR TPoint, VECTOR EPoint, TPATTERN *TPat)
{
   VECTOR PTurbulence,RP;
   int Axis,i,temp_rand;
   int blockX = 0, blockY = 0, blockZ = 0 ;
   SNGL BlkNum;
   DBL  Length;
   DBL  Strength,xx,yy,zz,ee,ff,gg;
   DBL amount;
   WARP *Warp=TPat->Warps;
   TURB *Turb;
   TRANS *Tr;
   REPEAT *Repeat;
   BLACK_HOLE *Black_Hole;
   VECTOR Delta, Center, Angle;
/* JG3 start */
   SPHERIC_WARP *Spheric_warp;
   CYLINDRIC_WARP *Cylindric_warp;
   TORIC_WARP *Toric_warp;
   CONIC_WARP *Conic_warp;
   SCREWING_WARP *Screwing_warp;
   ROLLING_WARP *Rolling_warp;
   BENDING_WARP *Bending_warp;
   TRANSFORM Trans;
/* JG3 end */

   Assign_Vector(TPoint, EPoint);

   while (Warp != NULL)
   {
      if (Warp->Warp_Transform != NULL)
      {
        MInvTransPoint(TPoint, TPoint, Warp->Warp_Transform);
      }
      switch(Warp->Warp_Type)
      {
        case CLASSIC_TURB_WARP:
          if ((TPat->Type == MARBLE_PATTERN) ||
              (TPat->Type == NO_PATTERN)     ||
              (TPat->Type == WOOD_PATTERN))
          {
             break;
          }
        /* If not a special type, fall through to next case */

        case EXTRA_TURB_WARP:
          Turb=(TURB *)Warp;
          DTurbulence (PTurbulence, TPoint, Turb);
          TPoint[X] += PTurbulence[X] * Turb->Turbulence[X];
          TPoint[Y] += PTurbulence[Y] * Turb->Turbulence[Y];
          TPoint[Z] += PTurbulence[Z] * Turb->Turbulence[Z];
          break;

        case NO_WARP:
          break;

        case TRANSFORM_WARP:
          Tr=(TRANS *)Warp;
          MInvTransPoint(TPoint, TPoint, &(Tr->Trans));
          break;

        case REPEAT_WARP:
          Repeat=(REPEAT *)Warp;
          Assign_Vector(RP,TPoint);
          Axis=Repeat->Axis;
          BlkNum=floor(TPoint[Axis]/Repeat->Width);
          
          RP[Axis]=TPoint[Axis]-BlkNum*Repeat->Width;
          
          if (((int)BlkNum) & 1)
          {          
             VEvaluateEq(RP,Repeat->Flip);
             if ( Repeat->Flip[Axis] < 0 ) 
             {
                RP[Axis] = Repeat->Width+RP[Axis];
             }
          }

          VAddScaledEq(RP,BlkNum,Repeat->Offset);
          Assign_Vector(TPoint,RP);
          break;

        case BLACK_HOLE_WARP:
          Black_Hole = (BLACK_HOLE *) Warp ;
          Assign_Vector (Center, Black_Hole->Center) ;

          if (Black_Hole->Repeat)
          {
            /* first, get the block number we're in for each dimension  */
            /* block numbers are (currently) calculated relative to 0   */
            /* we use floor () since it correctly returns -1 for the
               first block below 0 in each axis                         */
            /* one final point - we could run into overflow problems if
               the repeat vector was small and the scene very large.    */
            if (Black_Hole->Repeat_Vector [X] >= Small_Tolerance)
              blockX = (int) floor (TPoint [X] / Black_Hole->Repeat_Vector [X]) ;

            if (Black_Hole->Repeat_Vector [Y] >= Small_Tolerance)
              blockY = (int) floor (TPoint [Y] / Black_Hole->Repeat_Vector [Y]) ;

            if (Black_Hole->Repeat_Vector [Z] >= Small_Tolerance)
              blockZ = (int) floor (TPoint [Z] / Black_Hole->Repeat_Vector [Z]) ;

            if (Black_Hole->Uncertain)
            {
              /* if the position is uncertain calculate the new one first */
              /* this will allow the same numbers to be returned by frand */
              
              temp_rand = POV_GET_OLD_RAND(); /*protect seed*/
  
              POV_SRAND (Hash3d (blockX, blockY, blockZ)) ;
              Center [X] += FRAND () * Black_Hole->Uncertainty_Vector [X] ;
              Center [Y] += FRAND () * Black_Hole->Uncertainty_Vector [Y] ;
              Center [Z] += FRAND () * Black_Hole->Uncertainty_Vector [Z] ;
              POV_SRAND (temp_rand) ;  /*restore*/
            }

            Center [X] += Black_Hole->Repeat_Vector [X] * blockX ;
            Center [Y] += Black_Hole->Repeat_Vector [Y] * blockY ;
            Center [Z] += Black_Hole->Repeat_Vector [Z] * blockZ ;
          }

          VSub (Delta, TPoint, Center) ;
          VLength (Length, Delta) ;

          /* Length is the distance from the centre of the black hole */
          if (Length >= Black_Hole->Radius) break ;

          if (Black_Hole->Type == 0)
          {
            /* now convert the length to a proportion (0 to 1) that the point
               is from the edge of the black hole. a point on the perimeter
               of the black hole will be 0.0 ; a point at the centre will be
               1.0 ; a point exactly halfway will be 0.5, and so forth. */
            Length = (Black_Hole->Radius - Length) / Black_Hole->Radius ;

            /* Strength is the magnitude of the transformation effect. firstly,
               apply the Power variable to Length. this is meant to provide a
               means of controlling how fast the power of the Black Hole falls
               off from its centre. if Power is 2.0, then the effect is inverse
               square. increasing power will cause the Black Hole to be a lot
               weaker in its effect towards its perimeter. 
               
               finally we multiply Strength with the Black Hole's Strength
               variable. if the resultant value exceeds 1.0 we clip it to 1.0.
               this means a point will never be transformed by more than its
               original distance from the centre. the result of this clipping
               is that you will have an 'exclusion' area near the centre of
               the black hole where all points whose final value exceeded or
               equalled 1.0 were moved by a fixed amount. this only happens
               if the Strength value of the Black Hole was greater than one. */

            Strength = pow (Length, Black_Hole->Power) * Black_Hole->Strength ;
            if (Strength > 1.0) Strength = 1.0 ;
            
            /* if the Black Hole is inverted, it gives the impression of 'push-
               ing' the pattern away from its centre. otherwise it sucks. */
            VScaleEq (Delta, Black_Hole->Inverted ? -Strength : Strength) ;

            /* add the scaled Delta to the input point to end up with TPoint. */
            VAddEq (TPoint, Delta) ;
          }
          break;
          
/* JG3 start */
        case SPHERE_WARP:
          Spheric_warp = (SPHERIC_WARP *) Warp ;
          Assign_Vector (Delta, TPoint) ;
          Assign_Vector (Center, TPoint) ;
          if (!Spheric_warp->Inversed)
          {
            VLength (Length, Delta) ;
            TPoint [X] = Length;
            Center [Y] = atan2(Delta[Z],Delta[X]);
            Center [X] = 0.0;
            Center [Z] = 0.0;
            TPoint [Z] = Center[Y]/M_PI;
            Center [Y] *= 180.0/M_PI;
            Compute_Rotation_Transform(&Trans,Center);
            MTransPoint(Center, Delta, &Trans);
            TPoint [Y] = atan2(Center[Y],Center[X])/M_PI_2 ;
          }
          else
          {
            /* Delta[]
            ** X is length [0...infinity]
            ** Y is angle (elevation) [-1,1]
            ** Z is angle  [-1,1]
            */
            Length = Delta [X];
            Center [Y] = fmod( Delta [Y], 2.0)*M_PI_2;
            Center [Z] = fmod( Delta [Z], 2.0)*M_PI;
            TPoint [X] = Length * cos (Center [Y])*cos(Center[Z]) ;
            TPoint [Y] = Length * sin(Center[Y]);
            TPoint [Z] = Length * cos (Center [Y])*sin(Center[Z]);
          }
          VSub(Center,TPoint,Delta); /* from old to new point */
          VAddScaled(TPoint,Delta,Spheric_warp->Amount,Center);
          break;

        case CYLINDER_WARP:
          Cylindric_warp = (CYLINDRIC_WARP *) Warp;
          Assign_Vector (Delta, TPoint) ;
          Assign_Vector (Center, TPoint) ;
          if (!Cylindric_warp->Inversed) 
          {
            TPoint [Y] = Delta[Y] ;
            Delta[Y] = 0.0;
            VLength (Length, Delta) ;
            TPoint [X] = Length ;
            TPoint [Z] = atan2(Delta[Z],Delta[X])/M_PI;
          }
          else
          {
            /* Delta[]
            ** X is radial length in x,z plane
            ** Y is height
            ** Z is angle [-1,1]
            */
            Center [Z] = fmod( Delta [Z], 2.0)*M_PI;
            Length = Delta [X];
            TPoint [Y] = Delta[Y] ;
            TPoint [X] = Length * cos(Center[Z]) ;
            TPoint [Z] = Length * sin(Center[Z]) ;
          }
          VSub(Center,TPoint,Delta); /* from old to new point */
          VAddScaled(TPoint,Delta,Cylindric_warp->Amount,Center);
          break;

        case TORUS_WARP:
          Toric_warp = (TORIC_WARP *) Warp;
          Assign_Vector (Delta, TPoint) ;
          Assign_Vector (Center, TPoint) ;
          if (!Toric_warp->Inversed)
          {
            Center[Y] = atan2(Delta[Z],Delta[X]);
            Center[X] = 0.0;
            Center[Z] = 0.0;
            TPoint [Z] = Center[Y]/M_PI ;
            Center [Y] *= 180.0/M_PI;
            Compute_Rotation_Transform(&Trans,Center);
            MTransPoint(Center, Delta, &Trans);
            Center[X] -= Toric_warp->Radius;
            VLength (Length, Center) ;
            TPoint [X] = Length ;
            TPoint [Y] = atan2(Center[Y],Center[X])/M_PI ;
          }
          else
          {
            /* Delta[]
            ** X is length from circle
            ** Y is angle [-1,1], elevation
            ** Z is angle [-1,1]
            */
            Center [Y] = fmod( Delta [Y], 2.0)*M_PI;
            Center [Z] = fmod( Delta [Z], 2.0)*M_PI;
            Length = Delta [X];
            TPoint[Y] = Length * sin(Center[Y]) ;
            TPoint[X] = (Toric_warp->Radius + (Length * cos(Center[Y])))*cos(Center[Z]);
            TPoint[Z] = (Toric_warp->Radius + (Length * cos(Center[Y])))*sin(Center[Z]);
          }
          VSub(Center,TPoint,Delta); /* from old to new point */
          VAddScaled(TPoint,Delta,Toric_warp->Amount,Center);
          break;


        case CONE_WARP:
          Conic_warp = (CONIC_WARP *) Warp;
          Assign_Vector (Delta, TPoint) ;
          Assign_Vector (Center, TPoint) ;
          if (!Conic_warp->Inversed)
          {
            Center [Y] = Delta [Y];
            Delta[Y] = 0.0;
            VLength (Length, Delta) ;
            if (Center[Y] != 0.0)
            {
                TPoint [X] = Length / fabs( Center[Y]);
            }
            else
            {  
                /* Special case: the <xz> plane at the origin cannot be described
                   with the conic coordinates
                   Let's keep the Z valid, but the X is at the origin 
                */
                TPoint [X] = 0.0;
            }
            TPoint [Y] = Center[Y] ;
            TPoint [Z] = atan2(Delta[Z],Delta[X])/M_PI ;
          }
          else
          {
            /* Delta[]
            ** X is cone slope
            ** Y is height
            ** Z is angle [-1,1]
            */
            Center [Z] = fmod( Delta [Z], 2.0)*M_PI;
            TPoint[Y] = Delta[Y] ;
            Length = Delta[X] * fabs(Delta[Y]) ;
            TPoint[Z] = Length * sin(Center[Z]) ;
            TPoint[X] = Length * cos(Center[Z]) ;

          }
          VSub(Center,TPoint,Delta); /* from old to new point */
          VAddScaled(TPoint,Delta,Conic_warp->Amount,Center);
          break;

        case ROLL_WARP:
          Rolling_warp = (ROLLING_WARP *) Warp;
          xx=Rolling_warp->Axis[X];
          yy=Rolling_warp->Axis[Y];
          zz=Rolling_warp->Axis[Z];
          Assign_Vector(Delta,Rolling_warp->Center);
          Assign_Vector(Center,Rolling_warp->Param);

          VSubEq(TPoint,Delta);
          VCross(Angle,Center,TPoint);
          VLength(amount,Angle);
          amount *= (Rolling_warp->Inversed ? 1:-1);
          if ((Rolling_warp->Has_mini)&&(amount<Rolling_warp->Minimal))
          {
             amount = Rolling_warp->Minimal;
          }
          if ((Rolling_warp->Has_maxi)&&(amount>Rolling_warp->Maximal))
          {
             amount = Rolling_warp->Maximal;
          }
          ee=cos(amount*M_PI_180);
          ff=sin(amount*M_PI_180);
          gg=1-ee;           

          Trans.matrix[0][0]= xx*xx*gg+ee;
          Trans.matrix[0][1]= xx*yy*gg+ff*zz;
          Trans.matrix[0][2]= xx*zz*gg-ff*yy;
          Trans.matrix[1][0]= xx*yy*gg-ff*zz;
          Trans.matrix[1][1]= yy*yy*gg+ee;
          Trans.matrix[1][2]= yy*zz*gg+ff*xx;
          Trans.matrix[2][0]= xx*zz*gg+ff*yy;
          Trans.matrix[2][1]= yy*zz*gg-ff*xx;
          Trans.matrix[2][2]= zz*zz*gg+ee;

          Trans.matrix[3][0]= 0;
          Trans.matrix[3][1]= 0;
          Trans.matrix[3][2]= 0;
          Trans.matrix[0][3]= 0;
          Trans.matrix[1][3]= 0;
          Trans.matrix[2][3]= 0;
          MTransPoint(EPoint,TPoint,&Trans);
          VAdd(TPoint,EPoint,Delta);
          break;
        case BEND_WARP:
          Bending_warp = (BENDING_WARP *) Warp;
          xx=Bending_warp->Axis[X];
          yy=Bending_warp->Axis[Y];
          zz=Bending_warp->Axis[Z];
          Assign_Vector(Delta,Bending_warp->Center);
          VSubEq(TPoint,Delta);
          MInvTransPoint(Center,TPoint,&(Bending_warp->Projection));
          amount = Center[Z];
          Center[Z] = 0.0;
          if ((Bending_warp->Has_mini)&&(amount<Bending_warp->Minimal))
          {
             Center[Z] = amount-Bending_warp->Minimal;
          }
          if ((Bending_warp->Has_maxi)&&(amount>Bending_warp->Maximal))
          {
             Center[Z] = amount-Bending_warp->Maximal;
          }
          amount *= Bending_warp->Amount;
          ee=cos(amount*M_PI_180);
          ff=sin(amount*M_PI_180);
          gg=1-ee;           

          Trans.matrix[0][0]= xx*xx*gg+ee;
          Trans.matrix[0][1]= xx*yy*gg+ff*zz;
          Trans.matrix[0][2]= xx*zz*gg-ff*yy;
          Trans.matrix[1][0]= xx*yy*gg-ff*zz;
          Trans.matrix[1][1]= yy*yy*gg+ee;
          Trans.matrix[1][2]= yy*zz*gg+ff*xx;
          Trans.matrix[2][0]= xx*zz*gg+ff*yy;
          Trans.matrix[2][1]= yy*zz*gg-ff*xx;
          Trans.matrix[2][2]= zz*zz*gg+ee;

          Trans.matrix[3][0]= 0;
          Trans.matrix[3][1]= 0;
          Trans.matrix[3][2]= 0;
          Trans.matrix[0][3]= 0;
          Trans.matrix[1][3]= 0;
          Trans.matrix[2][3]= 0;
          MTransPoint(EPoint,Center,&Trans);
          MTransPoint(Center,EPoint,&(Bending_warp->Projection));
          VAdd(TPoint,Center,Delta);
          break;
        case SCREW_WARP:
          Screwing_warp = (SCREWING_WARP *) Warp;
          xx=Screwing_warp->Axis[X];
          yy=Screwing_warp->Axis[Y];
          zz=Screwing_warp->Axis[Z];
          Assign_Vector(Delta,Screwing_warp->Center);
          Assign_Vector(Center,Screwing_warp->Param);

          VSubEq(TPoint,Delta);
          VDot(amount,TPoint,Center);
          amount *= (Screwing_warp->Inversed ? 1:-1);
          if ((Screwing_warp->Has_mini)&&(amount<Screwing_warp->Minimal))
          {
             amount = Screwing_warp->Minimal;
          }
          if ((Screwing_warp->Has_maxi)&&(amount>Screwing_warp->Maximal))
          {
             amount = Screwing_warp->Maximal;
          }
          ee=cos(amount*M_PI_180);
          ff=sin(amount*M_PI_180);
          gg=1-ee;           

          Trans.matrix[0][0]= xx*xx*gg+ee;
          Trans.matrix[0][1]= xx*yy*gg+ff*zz;
          Trans.matrix[0][2]= xx*zz*gg-ff*yy;
          Trans.matrix[1][0]= xx*yy*gg-ff*zz;
          Trans.matrix[1][1]= yy*yy*gg+ee;
          Trans.matrix[1][2]= yy*zz*gg+ff*xx;
          Trans.matrix[2][0]= xx*zz*gg+ff*yy;
          Trans.matrix[2][1]= yy*zz*gg-ff*xx;
          Trans.matrix[2][2]= zz*zz*gg+ee;

          Trans.matrix[3][0]= 0;
          Trans.matrix[3][1]= 0;
          Trans.matrix[3][2]= 0;
          Trans.matrix[0][3]= 0;
          Trans.matrix[1][3]= 0;
          Trans.matrix[2][3]= 0;
          MTransPoint(EPoint,TPoint,&Trans);
          VAdd(TPoint,EPoint,Delta);
          break;
/* JG3 end */

        default:
          Error("Warp type %d not yet implemented",Warp->Warp_Type);
      }
      if (Warp->Warp_Transform != NULL)
      {
        MTransPoint(TPoint, TPoint, Warp->Warp_Transform);
      }
      Warp=Warp->Next_Warp;
   }

   for (i=X; i<=Z; i++)
     if (TPoint[i] > COORDINATE_LIMIT)
       TPoint[i]= COORDINATE_LIMIT;
     else
       if (TPoint[i] < -COORDINATE_LIMIT)
         TPoint[i] = -COORDINATE_LIMIT;

}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*   
* OUTPUT
*   
* RETURNS
*   
* AUTHOR
*   
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

WARP *Create_Warp (int Warp_Type)
{
 WARP *New;
 TURB *TNew;
 REPEAT *RNew;
 TRANS *TRNew;
 BLACK_HOLE *BNew;
/* JG3 start */
 SPHERIC_WARP *SNew;
 CYLINDRIC_WARP *CNew;
 TORIC_WARP *TONew;
 CONIC_WARP *CONew;
 ROLLING_WARP *RONew;
 SCREWING_WARP *SCNew;
 BENDING_WARP *BENew;
/* JG3 end */
   
 New = NULL;

 switch (Warp_Type)
 {
   case CLASSIC_TURB_WARP:
   case EXTRA_TURB_WARP:
     
     TNew = (TURB *)POV_MALLOC(sizeof(TURB),"turbulence struct");

     Make_Vector(TNew->Turbulence,0.0,0.0,0.0);

     TNew->Octaves = 6;
     TNew->Omega = 0.5;
     TNew->Lambda = 2.0;

     New = (WARP *)TNew;

     break;
     
   case REPEAT_WARP:

     RNew = (REPEAT *)POV_MALLOC(sizeof(REPEAT),"repeat warp");

     RNew->Axis = -1;
     RNew->Width = 0.0;

     Make_Vector(RNew->Offset,0.0,0.0,0.0);
     Make_Vector(RNew->Flip,1.0,1.0,1.0);

     New = (WARP *)RNew;

     break;

   case BLACK_HOLE_WARP:
     BNew = (BLACK_HOLE *)POV_MALLOC (sizeof (BLACK_HOLE), "black hole warp") ;
     Make_Vector (BNew->Center, 0.0, 0.0, 0.0) ;
     Make_Vector (BNew->Repeat_Vector, 0.0, 0.0, 0.0) ;
     Make_Vector (BNew->Uncertainty_Vector, 0.0, 0.0, 0.0) ;
     BNew->Strength = 1.0 ;
     BNew->Power = 2.0 ;
     BNew->Radius = 1.0 ;
     BNew->Radius_Squared = 1.0 ;
     BNew->Inverse_Radius = 1.0 ;
     BNew->Inverted = FALSE ;
     BNew->Type = 0 ;
     BNew->Repeat = FALSE ;
     BNew->Uncertain = FALSE ;
     New = (WARP *) BNew ;
     break ;
/* JG3 start */
   case SPHERE_WARP:
     SNew = (SPHERIC_WARP *)POV_MALLOC (sizeof (SPHERIC_WARP), "spheric warp") ;
     SNew->Amount = 1;
     SNew->Inversed = 0;
     New = (WARP *) SNew;
     break;

   case BEND_WARP:
     BENew = (BENDING_WARP *)POV_MALLOC (sizeof (BENDING_WARP), "bending warp") ;
     Make_Vector (BENew->Center, 0.0, 0.0, 0.0);
     Make_Vector (BENew->Angle, 1.0, 1.0, 1.0);
     Make_Vector (BENew->No_move, 1.0, 1.0, 1.0);
     BENew->Amount = 0;
     BENew->Has_mini=0;
     BENew->Has_maxi=0;
     New = (WARP *) BENew;
     break;
   case ROLL_WARP:
     RONew = (ROLLING_WARP *)POV_MALLOC (sizeof (ROLLING_WARP), "rolling warp") ;
     Make_Vector (RONew->Center, 0.0, 0.0, 0.0);
     Make_Vector (RONew->Param, 1.0, 1.0, 1.0);
     RONew->Inversed = 0;
     RONew->Has_mini=0;
     RONew->Has_maxi=0;
     New = (WARP *) RONew;
     break;

   case SCREW_WARP:
     SCNew = (SCREWING_WARP *)POV_MALLOC (sizeof (SCREWING_WARP), "screwing warp") ;
     Make_Vector (SCNew->Center, 0.0, 0.0, 0.0);
     Make_Vector (SCNew->Param, 1.0, 1.0, 1.0);
     SCNew->Inversed = 0;
     SCNew->Has_mini=0;
     SCNew->Has_maxi=0;
     New = (WARP *) SCNew;
     break;

   case CYLINDER_WARP:
     CNew = (CYLINDRIC_WARP *)POV_MALLOC (sizeof (CYLINDRIC_WARP), "cylindric warp") ;
     CNew->Amount = 1;
     CNew->Inversed = 0;
     New = (WARP *) CNew;
     break;

   case TORUS_WARP:
     TONew = (TORIC_WARP *)POV_MALLOC (sizeof (TORIC_WARP), "toric warp") ;
     TONew->Amount = 1.0;
     TONew->Radius = 1.0;
     TONew->Inversed = 0;
     New = (WARP *) TONew;
     break;

   case CONE_WARP:
     CONew = (CONIC_WARP *)POV_MALLOC (sizeof (CONIC_WARP), "conic warp") ;
     CONew->Amount = 1.0;
     CONew->Inversed = 0;
     New = (WARP *) CONew;
     break;
/* JG3 end */

   case TRANSFORM_WARP:

     TRNew = (TRANS *)POV_MALLOC(sizeof(TRANS),"pattern transform");

     MIdentity (TRNew->Trans.matrix);
     MIdentity (TRNew->Trans.inverse);

     New = (WARP *)TRNew;

     break;
     
   default:

     Error("Unknown Warp type %d.",Warp_Type);
  }
  
  New->Warp_Transform = NULL;
  New->Warp_Type = Warp_Type;
  New->Next_Warp = NULL;
  
  return(New);
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*   
* OUTPUT
*   
* RETURNS
*   
* AUTHOR
*   
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Destroy_Warps (WARP *Warps)
{
 WARP *Temp1 = Warps;
 WARP *Temp2;

 while (Temp1!=NULL)
 {
   Temp2 = Temp1->Next_Warp;
   if (Temp1->Warp_Transform != NULL)
   {
     Destroy_Transform(Temp1->Warp_Transform);                               
   }
   POV_FREE(Temp1);
   
   Temp1 = Temp2;
 }
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*   
* OUTPUT
*   
* RETURNS
*   
* AUTHOR
*   
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

WARP *Copy_Warps (WARP *Old)
{
  WARP *New;

  if (Old != NULL)
  {
    New=Create_Warp(Old->Warp_Type);
    New->Warp_Transform = Copy_Transform(Old->Warp_Transform);
    switch (Old->Warp_Type)
    {
       case CLASSIC_TURB_WARP:
       case EXTRA_TURB_WARP:
         memcpy(New,Old,sizeof(TURB));
         break;
     
       case REPEAT_WARP:
         memcpy(New,Old,sizeof(REPEAT));
         break;
     
       case BLACK_HOLE_WARP:
         memcpy(New,Old,sizeof(BLACK_HOLE));
         break;

/* JG3 start */
       case SPHERE_WARP:
         memcpy(New,Old,sizeof(SPHERIC_WARP));
         break;

       case CYLINDER_WARP:
         memcpy(New,Old,sizeof(CYLINDRIC_WARP));
         break;

       case TORUS_WARP:
         memcpy(New,Old,sizeof(TORIC_WARP));
         break;

       case CONE_WARP:
         memcpy(New,Old,sizeof(CONIC_WARP));
         break;

       case BEND_WARP:
         memcpy(New,Old,sizeof(BENDING_WARP));
         break;

       case ROLL_WARP:
         memcpy(New,Old,sizeof(ROLLING_WARP));
         break;

       case SCREW_WARP:
         memcpy(New,Old,sizeof(SCREWING_WARP));
         break;

/* JG3 end */
     
       case TRANSFORM_WARP:
         memcpy(New,Old,sizeof(TRANS));
         break;
    }
    New->Next_Warp=Copy_Warps(Old->Next_Warp);
  }
  else
  {
    New=NULL;
  }
  return(New);
}

