/**************************************************************************
*                pattern.c
*
*  This module implements texturing functions that return a value to be
*  used in a pigment or normal.
*
*  from Persistence of Vision(tm) Ray Tracer
*  Copyright 1996,1999 Persistence of Vision Team
*---------------------------------------------------------------------------
*  NOTICE: This source code file is provided so that users may experiment
*  with enhancements to POV-Ray and to port the software to platforms other
*  than those supported by the POV-Ray Team.  There are strict rules under
*  which you are permitted to use this file.  The rules are in the file
*  named POVLEGAL.DOC which should be distributed with this file.
*  If POVLEGAL.DOC is not available or for more info please contact the POV-Ray
*  Team Coordinator by email to team-coord@povray.org or visit us on the web at
*  http://www.povray.org. The latest version of POV-Ray may be found at this site.
*
* This program is based on the popular DKB raytracer version 2.12.
* DKBTrace was originally written by David K. Buck.
* DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
*
* Modifications by Hans-Detlev Fink, January 1999, used with permission
* Modifications by Thomas Willhalm, March 1999, used with permission
*
*****************************************************************************/

/*
 * Some texture ideas garnered from SIGGRAPH '85 Volume 19 Number 3,
 * "An Image Synthesizer" By Ken Perlin.
 * Further Ideas Garnered from "The RenderMan Companion" (Addison Wesley).
 */

#include "frame.h"
#include "vector.h"
#include "povproto.h"
#include "matrices.h"
#include "pattern.h"
#include "povray.h"
#include "texture.h"
#include "image.h"
#include "txttest.h"
#include "colour.h"
#include "camera.h"


/*****************************************************************************
* Local preprocessor defines
******************************************************************************/


/*****************************************************************************
* Static functions
******************************************************************************/

static DBL agate (VECTOR EPoint, TPATTERN *TPat);
static DBL brick (VECTOR EPoint, TPATTERN *TPat);
static DBL checker (VECTOR EPoint);
static DBL square (VECTOR EPoint);
static DBL ternaire (VECTOR EPoint);
static DBL quadran (VECTOR EPoint);
static DBL classique (VECTOR EPoint);
static DBL tintin (VECTOR EPoint, VECTOR NPoint, TPATTERN *TPat);
static DBL tintins (VECTOR EPoint, VECTOR NPoint, TPATTERN *TPat);
static DBL cellulo (VECTOR NPoint, RAY *ORay);
static DBL crackle (VECTOR EPoint);
static DBL gradient (VECTOR EPoint, TPATTERN *TPat);
static DBL granite (VECTOR EPoint);
static DBL leopard (VECTOR EPoint);
static DBL mandel (VECTOR EPoint, TPATTERN *TPat);
static DBL marble (VECTOR EPoint, TPATTERN *TPat);
static DBL onion (VECTOR EPoint);
static DBL radial (VECTOR EPoint);
static DBL spiral1 (VECTOR EPoint, TPATTERN *TPat);
static DBL spiral2 (VECTOR EPoint, TPATTERN *TPat);
static DBL wood (VECTOR EPoint, TPATTERN *TPat);
static DBL hexagon (VECTOR EPoint);
static DBL planar_pattern (VECTOR EPoint);
static DBL spherical (VECTOR EPoint);
static DBL boxed (VECTOR EPoint);
static DBL cylindrical (VECTOR EPoint);
static DBL density_file (VECTOR EPoint, TPATTERN *TPat);

static long PickInCube (VECTOR tv, VECTOR p1);

static DBL ripples_pigm (VECTOR EPoint, TPATTERN *TPat);
static DBL waves_pigm  (VECTOR EPoint, TPATTERN *TPat);
static DBL dents_pigm  (VECTOR EPoint);
static DBL wrinkles_pigm (VECTOR EPoint);
static DBL quilted_pigm (VECTOR EPoint, TPATTERN *TPat);
static TURB *Search_For_Turb (WARP *Warps);
/* static TURB *Copy_Turb (TURB *Old);   Unused function [AED] */
static unsigned short readushort(FILE *infile);

/* JG2 start */
static DBL blotches (VECTOR EPoint);
static DBL cells (VECTOR tv);
static DBL bands (VECTOR EPoint, TPATTERN *TPat);
static DBL sheet (VECTOR EPoint, TPATTERN *TPat);    
/* JG2 end */
/* JG8 start */
static DBL voronoi (VECTOR EPoint);
/* JG8 end */
/* Carrelage */
static DBL car_carre  (VECTOR EPoint);         
static DBL car_hexagone  (VECTOR EPoint);         
static DBL car_triangle  (VECTOR EPoint);         
static DBL car_lozange  (VECTOR EPoint);         
static DBL car_rombhe  (VECTOR EPoint);         
static DBL car_rectangle  (VECTOR EPoint);         
static DBL car_ortho_carre  (VECTOR EPoint);         
static DBL car_carre_triangle  (VECTOR EPoint);         
static DBL car_hexa_triangle  (VECTOR EPoint);         
static DBL car_carre_decale  (VECTOR EPoint);         
static DBL car_rectangle_carre  (VECTOR EPoint);         
static DBL car_carre_rectangle  (VECTOR EPoint);         
static DBL car_carre_interne  (VECTOR EPoint);         
static DBL car_carre_interne_5  (VECTOR EPoint);         
static DBL car_carre_double  (VECTOR EPoint);                         
static DBL car_hexa_carre_triangle  (VECTOR EPoint);                         
static DBL car_hexa_carre_triangle_6  (VECTOR EPoint);                         
static DBL car_rectangle_paire  (VECTOR EPoint);                         
static DBL car_hexa_tri_droite  (VECTOR EPoint);                         
static DBL car_hexa_tri_gauche  (VECTOR EPoint);                         
static DBL car_carre_tri  (VECTOR EPoint);                         
static DBL car_dodeca_tri  (VECTOR EPoint);                         
static DBL car_dodeca_hex  (VECTOR EPoint);                         
static DBL car_dodeca_hex_5  (VECTOR EPoint);                         
static DBL tetragonal  (VECTOR EPoint, TPATTERN *TPat);                         
static DBL sexagonal  (VECTOR EPoint, TPATTERN *TPat);                         
static DBL trigonal  (VECTOR EPoint, TPATTERN *TPat);                         
/* Fin de Carrelage */

/*****************************************************************************
*
* FUNCTION
*
*   agate
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern is evaluated.
*   TPat   -- Texture pattern struct
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
* CHANGES
*   Oct 1994    : adapted from agate pigment by [CY]
*
******************************************************************************/

static DBL agate (VECTOR EPoint, TPATTERN *TPat)
{
  register DBL noise, turb_val;
  TURB* Turb;

  Turb=Search_For_Turb(TPat->Warps);

  turb_val = TPat->Vals.Agate_Turb_Scale * Turbulence(EPoint,Turb);

  noise = 0.5 * (cycloidal(1.3 * turb_val + 1.1 * EPoint[Z]) + 1.0);

  if (noise < 0.0)
  {
    noise = 0.0;
  }
  else
  {
    noise = min(1.0, noise);
    noise = pow(noise, 0.77);
  }

  return(noise);
}


/*****************************************************************************
*
* FUNCTION
*
*   brick
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   TPat   -- Texture pattern struct
*   
* OUTPUT
*   
* RETURNS
*
*   DBL value exactly 0.0 or 1.0
*   
* AUTHOR
*
*   Dan Farmer
*   
* DESCRIPTION
*
* CHANGES
*   Oct 1994    : adapted from pigment by [CY]
*
******************************************************************************/

static DBL brick (VECTOR EPoint, TPATTERN *TPat)
{
  int ibrickx, ibricky, ibrickz;
  DBL brickheight, brickwidth, brickdepth;
  DBL brickmortar, mortarheight, mortarwidth, mortardepth;
  DBL brickx, bricky, brickz;
  DBL x, y, z, fudgit;

  fudgit=Small_Tolerance+TPat->Vals.Brick.Mortar;

  x =  EPoint[X]+fudgit;
  y =  EPoint[Y]+fudgit;
  z =  EPoint[Z]+fudgit;

  brickwidth  = TPat->Vals.Brick.Size[X];
  brickheight = TPat->Vals.Brick.Size[Y];
  brickdepth  = TPat->Vals.Brick.Size[Z];
  brickmortar = (DBL)TPat->Vals.Brick.Mortar;

  mortarwidth  = brickmortar / brickwidth;
  mortarheight = brickmortar / brickheight;
  mortardepth  = brickmortar / brickdepth;

  /* 1) Check mortar layers in the X-Z plane (ie: top view) */

  bricky = y / brickheight;
  ibricky = (int) bricky;
  bricky -= (DBL) ibricky;

  if (bricky < 0.0)
  {
    bricky += 1.0;
  }

  if (bricky <= mortarheight)
  {
    return(0.0);
  }

  bricky = (y / brickheight) * 0.5;
  ibricky = (int) bricky;
  bricky -= (DBL) ibricky;

  if (bricky < 0.0)
  {
    bricky += 1.0;
  }


  /* 2) Check ODD mortar layers in the Y-Z plane (ends) */

  brickx = (x / brickwidth);
  ibrickx = (int) brickx;
  brickx -= (DBL) ibrickx;

  if (brickx < 0.0)
  {
    brickx += 1.0;
  }

  if ((brickx <= mortarwidth) && (bricky <= 0.5))
  {
    return(0.0);
  }

  /* 3) Check EVEN mortar layers in the Y-Z plane (ends) */

  brickx = (x / brickwidth) + 0.5;
  ibrickx = (int) brickx;
  brickx -= (DBL) ibrickx;

  if (brickx < 0.0)
  {
    brickx += 1.0;
  }

  if ((brickx <= mortarwidth) && (bricky > 0.5))
  {
    return(0.0);
  }

  /* 4) Check ODD mortar layers in the Y-X plane (facing) */

  brickz = (z / brickdepth);
  ibrickz = (int) brickz;
  brickz -= (DBL) ibrickz;

  if (brickz < 0.0)
  {
    brickz += 1.0;
  }

  if ((brickz <= mortardepth) && (bricky > 0.5))
  {
    return(0.0);
  }

  /* 5) Check EVEN mortar layers in the X-Y plane (facing) */

  brickz = (z / brickdepth) + 0.5;
  ibrickz = (int) brickz;
  brickz -= (DBL) ibrickz;

  if (brickz < 0.0)
  {
    brickz += 1.0;
  }

  if ((brickz <= mortardepth) && (bricky <= 0.5))
  {
    return(0.0);
  }

  /* If we've gotten this far, color me brick. */

  return(1.0);
}


/*****************************************************************************
*
* FUNCTION
*
*   checker
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value exactly 0.0 or 1.0
*
* AUTHOR
*
*   POV-Team
*
* DESCRIPTION
*
* CHANGES
*   Oct 1994    : adapted from pigment by [CY]
*
******************************************************************************/

static DBL checker (VECTOR EPoint)
{
  int value;

  value = (int)(floor(EPoint[X]+Small_Tolerance) +
                floor(EPoint[Y]+Small_Tolerance) +
                floor(EPoint[Z]+Small_Tolerance));

  if (value & 1)
  {
    return (1.0);
  }
  else
  {
    return (0.0);
  }
}


/*****************************************************************************
*
* FUNCTION
*
*   tintin
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   NPoint -- The normal in 3d space 
*   TPat   -- Pattern info
*
* OUTPUT
*
* RETURNS
*
*   DBL value exactly 0.0 or 1.0
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

static DBL tintin (VECTOR EPoint, VECTOR NPoint, TPATTERN *TPat )
{
  int value;
  int valu;
  DBL xx,zz,angle;
  xx = NPoint[X];
  zz = NPoint[Z];
  angle= (atan2(zz,xx) * TPat->Vals.Tintin.Num_x/TWO_M_PI ) + 0.5 ;
  valu = (int)(floor(angle));
  value = (int)( floor(EPoint[Y]+Small_Tolerance) );
  value ^= valu;
  if ( value & 1)
  {
    return (1.0);
  }
  else
  {
    return (0.0);
  }
}


/*****************************************************************************
*
* FUNCTION
*
*   tintins
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   NPoint -- The normal in 3d space 
*   TPat   -- Pattern info
*
* OUTPUT
*
* RETURNS
*
*   DBL value exactly 0.0 or 1.0
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

static DBL tintins (VECTOR EPoint, VECTOR NPoint, TPATTERN *TPat )
{
  int value;
  int valu;
  DBL xx,zz,angle;
  xx = NPoint[X];
  zz = NPoint[Z];
  angle= (atan2(zz,xx) * TPat->Vals.Tintin.Num_x/TWO_M_PI ) + 0.5;
  valu = (int)(floor(angle));
  value = (int)( floor(atan2(NPoint[Y],sqrt(xx*xx+zz*zz) )
                  *TPat->Vals.Tintin.Num_y/TWO_M_PI));
  value ^= valu;
  if ( value & 1)
  {
    return (1.0);
  }
  else
  {
    return (0.0);
  }
}

/*****************************************************************************
*
* FUNCTION
*
*   cellulo
*
* INPUT
*
*   NPoint -- The normal in 3d space 
*   Ray    -- Ray info    
*
* OUTPUT
*
* RETURNS
*
*   DBL value between 0.0 and 1.0
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

static DBL cellulo (VECTOR NPoint, RAY *Ray)
{
  VECTOR a,b;
  DBL angle;
  VNormalize(a,NPoint);
  VNormalize(b,Ray->Direction);
  VDot(angle,a,b);
  angle=fabs(angle);
  if (angle > 1.0) { angle = 1.0; }
  return angle;
}


/*****************************************************************************
*
* FUNCTION
*
*   square 
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value exactly 0.0, 1.0, 2.0 or 3.0
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*   Paving the XZ plan with 4 'colours', in square
*
* CHANGES
*
******************************************************************************/

static DBL square (VECTOR EPoint)
{
  int valueX,valueZ;

  valueX = (int)(floor(EPoint[X])); 
  valueZ = (int)(floor(EPoint[Z])); 

  if (valueX & 1)
  {
    if (valueZ & 1)
    {
       return (2.0);
    }
    else
    {
       return (3.0);
    }
  }
  else
  {
    if (valueZ & 1)
    {
       return (1.0);
    }
    else
    {
       return (0.0);
    }
  }
}

/*****************************************************************************
*
* FUNCTION
*
*   ternaire
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value exactly 0.0, 1.0, 2.0, 3.0, 4.0 or 5.0
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*   Paving the XZ plan with 6 'colours', in triangle around the origin
*
* CHANGES
*
******************************************************************************/

/* 1.73205080756887729352 is sqrt(3) */
/* .86602540378443864676 is sqrt(3)/2 */
#define SQR3_2  .866025403784438646763723170752
#define SQR3    1.732050807568877293527446341505
#define SQR2    1.414213562373095048801688724209
#define SQR2_2  .707106781186547524400844362104
static DBL ternaire (VECTOR EPoint)
{
DBL answer=0;
DBL x,z;
DBL xs,zs;
int a,b;
DBL k,slop1,slop2;
int mask;

x=EPoint[X];
z=EPoint[Z];

 xs = x-3.0*floor(x/3.0);
 zs = z-SQR3*floor(z/SQR3);
 /* xs,zs is in { [0.0, 3.0 [, [0.0, SQR3 [ } 
 ** but there is some symetry to simplify the testing
 */
 a = (int)floor(xs);
 xs -= a;
 b = (zs <SQR3_2 ? 0: 1);
 if (b)
 {
    zs = SQR3 - zs; /* mirror */
 } 

 k = 1.0 - xs;
 if ((xs != 0.0)&&( k != 0.0 )) /* second condition should never occurs */
 {
    slop1 = zs/xs;
    slop2 = zs/k; /* just in case */
   switch( (slop1<SQR3?1:0)+(slop2<SQR3?2:0))
    { 
     case 3:
      answer = 0.0;
      break;
     case 2:
      answer = 1.0;
       break;
     case 1: 
      answer = 3.0;
      break;
    }
 }
 else
 {
   answer = 1.0;
 }
 mask = (int) answer;
 answer = (mask & 1) ? fmod(6.0+answer+2.0*a,6.0): fmod(6.0+answer-2.0*a,6.0);
 if (b)
 {
   answer = 5.0 - answer;
 }

 return answer;
}

static DBL quadran (VECTOR EPoint)
{
DBL answer;
DBL x,y,z;
DBL xx,yy,zz;

 x=EPoint[X];
 y=EPoint[Y];
 z=EPoint[Z];
 xx=fabs(x);
 yy=fabs(y);
 zz=fabs(z);
 if ((xx>yy)&&(xx>zz))
 {
    answer = (x<0)?1.0:0.0;
 }
 else if ((yy>xx)&&(yy>zz))
 {
    answer = (y<0)?3.0:2.0;
 }
 else
 {
    answer = (z<0)?5.0:4.0;
 }
 return answer;
}

static DBL classique (VECTOR EPoint)
{
DBL answer;
DBL x,y,z;
DBL xx,yy,zz;

 x=EPoint[X];
 y=EPoint[Y];
 z=EPoint[Z];
 xx=fabs(x);
 yy=fabs(y);
 zz=fabs(z);
 if (y*y<(x*x+z*z))
 {
 if (xx>zz)
 {
    answer = (x<0)?1.0:0.0;
 }
 else
 {
    answer = (z<0)?5.0:4.0;
 }
 }
 else
 {
   answer = (y<0)?3.0:2.0;
 }
 return answer;
}



/*****************************************************************************
*
* FUNCTION
*
*   crackle
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Jim McElhiney
*
* DESCRIPTION
*
*   "crackle":
*
*   New colour function by Jim McElhiney,
*     CompuServe 71201,1326, aka mcelhiney@acm.org
*
*   Large scale, without turbulence, makes a pretty good stone wall.
*   Small scale, without turbulence, makes a pretty good crackle ceramic glaze.
*   Highly turbulent (with moderate displacement) makes a good marble, solving
*   the problem of apparent parallel layers in Perlin's method.
*   2 octaves of full-displacement turbulence make a great "drizzled paint"
*   pattern, like a 1950's counter top.
*   Rule of thumb:  put a single colour transition near 0 in your colour map.
*
*   Mathematically, the set crackle(p)=0 is a 3D Voronoi diagram of a field of
*   semirandom points, and crackle(p)>0 is distance from set along shortest path.
*   (A Voronoi diagram is the locus of points equidistant from their 2 nearest
*   neighbours from a set of disjoint points, like the membranes in suds are
*   to the centres of the bubbles).
*
*   All "crackle" specific source code and examples are in the public domain.
*
* CHANGES
*   Oct 1994    : adapted from pigment by [CY]
*
******************************************************************************/

static DBL crackle (VECTOR EPoint)
{
  int    i;
  long   thisseed;
  DBL    sum, minsum, minsum2, tf;
  VECTOR sv, tv, dv, t1, add;

  static int cvc;
  static long lastseed = 0x80000000;
  static VECTOR cv[81];

  Assign_Vector(tv,EPoint);

  /*
   * Check to see if the input point is in the same unit cube as the last
   * call to this function, to use cache of cubelets for speed.
   */

  thisseed = PickInCube(tv, t1);

  if (thisseed != lastseed)
  {
    /*
     * No, not same unit cube.  Calculate the random points for this new
     * cube and its 80 neighbours which differ in any axis by 1 or 2.
     * Why distance of 2?  If there is 1 point in each cube, located
     * randomly, it is possible for the closest random point to be in the
     * cube 2 over, or the one two over and one up.  It is NOT possible
     * for it to be two over and two up.  Picture a 3x3x3 cube with 9 more
     * cubes glued onto each face.
     */

    /* Now store a points for this cube and each of the 80 neighbour cubes. */

    cvc = 0;

    for (add[X] = -2.0; add[X] < 2.5; add[X] +=1.0)
    {
      for (add[Y] = -2.0; add[Y] < 2.5; add[Y] += 1.0)
      {
        for (add[Z] = -2.0; add[Z] < 2.5; add[Z] += 1.0)
        {
          /* For each cubelet in a 5x5 cube. */

          if ((fabs(add[X])>1.5)+(fabs(add[Y])>1.5)+(fabs(add[Z])>1.5) <= 1.0)
          {
            /* Yes, it's within a 3d knight move away. */

            VAdd(sv, tv, add);

            PickInCube(sv, t1);

            cv[cvc][X] = t1[X];
            cv[cvc][Y] = t1[Y];
            cv[cvc][Z] = t1[Z];
            cvc++;
          }
        }
      }
    }

    lastseed = thisseed;
  }

  /*
   * Find the 2 points with the 2 shortest distances from the input point.
   * Loop invariant:  minsum is shortest dist, minsum2 is 2nd shortest
   */

  /* Set up the loop so the invariant is true:  minsum <= minsum2 */

  VSub(dv, cv[0], tv);  minsum  = VSumSqr(dv);
  VSub(dv, cv[1], tv);  minsum2 = VSumSqr(dv);

  if (minsum2 < minsum)
  {
    tf = minsum; minsum = minsum2; minsum2 = tf;
  }

  /* Loop for the 81 cubelets to find closest and 2nd closest. */

  for (i = 2; i < cvc; i++)
  {
    VSub(dv, cv[i], tv);

    sum = VSumSqr(dv);

    if (sum < minsum)
    {
      minsum2 = minsum;
      minsum = sum;
    }
    else
    {
      if (sum < minsum2)
      {
        minsum2 = sum;
      }
    }
  }

  /* Crackle value is absolute value of diff in dist to closest 2 points. */

  tf = sqrt(minsum2) - sqrt(minsum);      /* minsum is known <= minsum2 */

  /*
   * Note that the theoretical range of this function is 0 to root 3.
   * In practice, it rarely exceeds 0.9, and only very rarely 1.0
   */

  return min(tf, 1.);
}


/*****************************************************************************
*
* FUNCTION
*
*   gradient
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   
* OUTPUT
*   
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*   
* AUTHOR
*
*   POV-Ray Team
*   
* DESCRIPTION
*
*   Gradient Pattern - gradient based on the fractional values of
*   x, y or z, based on whether or not the given directional vector is
*   a 1.0 or a 0.0.
*   The basic concept of this is from DBW Render, but Dave Wecker's
*   only supports simple Y axis gradients.
*
* CHANGES
*   Oct 1994    : adapted from pigment by [CY]
*
******************************************************************************/

static DBL gradient (VECTOR EPoint, TPATTERN *TPat)
{
  register int i;
  register DBL temp;
  DBL value = 0.0;

  for (i=X; i<=Z; i++)
  {
    if (TPat->Vals.Gradient[i] != 0.0)
    {
      temp = fabs(EPoint[i]);

      value += fmod(temp,1.0);
    }
  }

  /* Clamp to 1.0. */

  value = ((value > 1.0) ? fmod(value, 1.0) : value);

  return(value);
}



/*****************************************************************************
*
* FUNCTION
*
*   granite
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   
* OUTPUT
*   
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*   
* AUTHOR
*
*   POV-Ray Team
*   
* DESCRIPTION
*
*   Granite - kind of a union of the "spotted" and the "dented" textures,
*   using a 1/f fractal noise function for color values. Typically used
*   with small scaling values. Should work with colour maps for pink granite.
*
* CHANGES
*   Oct 1994    : adapted from pigment by [CY]
*
******************************************************************************/

static DBL granite (VECTOR EPoint)
{
  register int i;
  register DBL temp, noise = 0.0, freq = 1.0;
  VECTOR tv1,tv2;

  VScale(tv1,EPoint,4.0);

  for (i = 0; i < 6 ; freq *= 2.0, i++)
  {
    VScale(tv2,tv1,freq);
    temp = 0.5 - Noise (tv2);

    temp = fabs(temp);

    noise += temp / freq;
  }

  return(noise);
}



/*****************************************************************************
*
* FUNCTION
*
*   leopard
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Scott Taylor
*
* DESCRIPTION
*
* CHANGES
*   Jul 1991 : Creation.
*   Oct 1994 : adapted from pigment by [CY]
*
******************************************************************************/

static DBL leopard (VECTOR EPoint)
{
  register DBL value, temp1, temp2, temp3;

  /* This form didn't work with Zortech 386 compiler */
  /* value = Sqr((sin(x)+sin(y)+sin(z))/3); */
  /* So we break it down. */

  temp1 = sin(EPoint[X]);
  temp2 = sin(EPoint[Y]);
  temp3 = sin(EPoint[Z]);

  value = Sqr((temp1 + temp2 + temp3) / 3.0);

  return(value);
}



/*****************************************************************************
*
* FUNCTION
*
*   mandel
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   submitted by user, name lost (sorry)
*
* DESCRIPTION
* The mandel pattern computes the standard Mandelbrot fractal pattern and
* projects it onto the X-Y plane.  It uses the X and Y coordinates to compute
* the Mandelbrot set.
*
* CHANGES
*   Oct 1994 : adapted from pigment by [CY]
*
******************************************************************************/

static DBL mandel (VECTOR EPoint, TPATTERN *TPat)
{
  int it_max, col;
  DBL a, b, cf, a2, b2, x, y;

  a = x = EPoint[X]; a2 = Sqr(a);
  b = y = EPoint[Y]; b2 = Sqr(b);

  it_max = TPat->Vals.Iterations;

  for (col = 0; col < it_max; col++)
  {
    b  = 2.0 * a * b + y;
    a  = a2 - b2 + x;

    a2 = Sqr(a);
    b2 = Sqr(b);

    if ((a2 + b2) > 4.0)
    {
      break;
    }
  }

  cf = (DBL)col / (DBL)it_max;

  return(cf);
}



/*****************************************************************************
*
* FUNCTION
*
*   marble
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   TPat   -- Texture pattern struct
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
* CHANGES
*   Oct 1994 : adapted from pigment by [CY]
*
******************************************************************************/

static DBL marble (VECTOR EPoint, TPATTERN *TPat)
{
  register DBL turb_val;
  TURB *Turb;

  if ((Turb=Search_For_Turb(TPat->Warps)) != NULL)
  {
    turb_val = Turb->Turbulence[X] * Turbulence(EPoint,Turb);
  }
  else
  {
    turb_val = 0.0;
  }

  return(EPoint[X] + turb_val);
}



/*****************************************************************************
*
* FUNCTION
*
*   onion
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Scott Taylor
*
* DESCRIPTION
*
* CHANGES
*   Jul 1991 : Creation.
*   Oct 1994 : adapted from pigment by [CY]
*
******************************************************************************/

static DBL onion (VECTOR EPoint)
{
  /* The variable noise is not used as noise in this function */

  register DBL noise;

/*
   This ramp goes 0-1,1-0,0-1,1-0...

   noise = (fmod(sqrt(Sqr(x)+Sqr(y)+Sqr(z)),2.0)-1.0);

   if (noise<0.0) {noise = 0.0-noise;}
*/

  /* This ramp goes 0-1, 0-1, 0-1, 0-1 ... */

  noise = (fmod(sqrt(Sqr(EPoint[X])+Sqr(EPoint[Y])+Sqr(EPoint[Z])), 1.0));

  return(noise);
}



/*****************************************************************************
*
* FUNCTION
*
*   radial
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Chris Young -- new in vers 2.0
*
* DESCRIPTION
*
* CHANGES
*   Oct 1994 : adapted from pigment by [CY]
*
******************************************************************************/

static DBL radial (VECTOR EPoint)
{
  register DBL value;

  if ((fabs(EPoint[X])<0.001) && (fabs(EPoint[Z])<0.001))
  {
    value = 0.25;
  }
  else
  {
    value = 0.25 + (atan2(EPoint[X],EPoint[Z]) + M_PI) / TWO_M_PI;
  }

  return(value);
}



/*****************************************************************************
*
* FUNCTION
*
*   spiral1
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   TPat   -- Texture pattern struct
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Dieter Bayer
*
* DESCRIPTION
*   Spiral whirles around z-axis.
*   The number of "arms" is defined in the TPat.
*
* CHANGES
*   Aug 1994 : Creation.
*   Oct 1994 : adapted from pigment by [CY]
*
******************************************************************************/

static DBL spiral1(VECTOR EPoint, TPATTERN *TPat)
{
  DBL rad, phi, turb_val;
  DBL x = EPoint[X];
  DBL y = EPoint[Y];
  DBL z = EPoint[Z];
  TURB *Turb;

  if ((Turb=Search_For_Turb(TPat->Warps)) != NULL)
  {
    turb_val = Turb->Turbulence[X] * Turbulence(EPoint,Turb);
  }
  else
  {
    turb_val = 0.0;
  }

  /* Get distance from z-axis. */

  rad = sqrt(x * x + y * y);

  /* Get angle in x,y-plane (0...2 PI). */

  if (rad == 0.0)
  {
    phi = 0.0;
  }
  else
  {
    if (x < 0.0)
    {
      phi = 3.0 * M_PI_2 - asin(y / rad);
    }
    else
    {
      phi = M_PI_2 + asin(y / rad);
    }
  }

  return(z + rad + (DBL)TPat->Vals.Arms * phi / TWO_M_PI + turb_val);
}



/*****************************************************************************
*
* FUNCTION
*
*   spiral2
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   TPat   -- Texture pattern struct
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Dieter Bayer
*
* DESCRIPTION
*   Spiral whirles around z-axis.
*   The number of "arms" is defined in the TPat.
*
* CHANGES
*   Aug 1994 : Creation.
*   Oct 1994 : adapted from pigment by [CY]
*
******************************************************************************/


static DBL spiral2(VECTOR EPoint, TPATTERN *TPat)
{
  DBL rad, phi, turb_val;
  DBL x = EPoint[X];
  DBL y = EPoint[Y];
  DBL z = EPoint[Z];
  TURB *Turb;

  if ((Turb=Search_For_Turb(TPat->Warps)) != NULL)
  {
    turb_val = Turb->Turbulence[X] * Turbulence(EPoint,Turb);
  }
  else
  {
    turb_val = 0.0;
  }

  /* Get distance from z-axis. */

  rad = sqrt(x * x + y * y);

  /* Get angle in x,y-plane (0...2 PI) */

  if (rad == 0.0)
  {
    phi = 0.0;
  }
  else
  {
    if (x < 0.0)
    {
      phi = 3.0 * M_PI_2 - asin(y / rad);
    }
    else
    {
      phi = M_PI_2 + asin(y / rad);
    }
  }

  turb_val = Triangle_Wave(z + rad + (DBL)TPat->Vals.Arms * phi / TWO_M_PI +
                           turb_val);

  return(Triangle_Wave(rad) + turb_val);
}



/*****************************************************************************
*
* FUNCTION
*
*   wood
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   TPat   -- Texture pattern struct
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
* CHANGES
*   Oct 1994 : adapted from pigment by [CY]
*
******************************************************************************/


static DBL wood (VECTOR EPoint, TPATTERN *TPat)
{
  register DBL length;
  VECTOR WoodTurbulence;
  VECTOR point;
  DBL x=EPoint[X];
  DBL y=EPoint[Y];
  TURB *Turb;

  if ((Turb=Search_For_Turb(TPat->Warps)) != NULL)
  {
    DTurbulence (WoodTurbulence, EPoint,Turb);
    point[X] = cycloidal((x + WoodTurbulence[X]) * Turb->Turbulence[X]);
    point[Y] = cycloidal((y + WoodTurbulence[Y]) * Turb->Turbulence[Y]);
  }
  else
  {
    point[X] = 0.0;
    point[Y] = 0.0;
  }
  point[Z] = 0.0;

  point[X] += x;
  point[Y] += y;

  /* point[Z] += z; Deleted per David Buck --  BP 7/91 */

  VLength (length, point);

  return(length);
}


/*****************************************************************************
*
* FUNCTION
*
*   hexagon
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value exactly 0.0, 1.0 or 2.0
*
* AUTHOR
*
*   Ernest MacDougal Campbell III
*   
* DESCRIPTION
*
*   TriHex pattern -- Ernest MacDougal Campbell III (EMC3) 11/23/92
*
*   Creates a hexagon pattern in the XZ plane.
*
*   This algorithm is hard to explain.  First it scales the point to make
*   a few of the later calculations easier, then maps some points to be
*   closer to the Origin.  A small area in the first quadrant is subdivided
*   into a 6 x 6 grid.  The position of the point mapped into that grid
*   determines its color.  For some points, just the grid location is enough,
*   but for others, we have to calculate which half of the block it's in
*   (this is where the atan2() function comes in handy).
*
* CHANGES
*   Nov 1992 : Creation.
*   Oct 1994 : adapted from pigment by [CY]
*
******************************************************************************/

#define xfactor 0.5;         /* each triangle is split in half for the grid */
#define zfactor 0.866025404; /* sqrt(3)/2 -- Height of an equilateral triangle */

static DBL hexagon (VECTOR EPoint)
{
  int xm, zm;
  int brkindx;
  DBL xs, zs, xl, zl, value = 0.0;
  DBL x=EPoint[X];
  DBL z=EPoint[Z];


  /* Keep all numbers positive.  Also, if z is negative, map it in such a
   * way as to avoid mirroring across the x-axis.  The value 5.196152424
   * is (sqrt(3)/2) * 6 (because the grid is 6 blocks high)
   */

  x = fabs(x);

  /* Avoid mirroring across x-axis. */

  z = z < 0.0 ? 5.196152424 - fabs(z) : z;

  /* Scale point to make calcs easier. */

  xs = x/xfactor;
  zs = z/zfactor;

  /* Map points into the 6 x 6 grid where the basic formula works. */

  xs -= floor(xs/6.0) * 6.0;
  zs -= floor(zs/6.0) * 6.0;

  /* Get a block in the 6 x 6 grid. */

  xm = (int) FLOOR(xs) % 6;
  zm = (int) FLOOR(zs) % 6;

  switch (xm)
  {
    /* These are easy cases: Color depends only on xm and zm. */

    case 0:
    case 5:

      switch (zm)
      {
        case 0:
        case 5: value = 0; break;

        case 1:
        case 2: value = 1; break;

        case 3:
        case 4: value = 2; break;
      }

      break;

    case 2:
    case 3:

      switch (zm)
      {
        case 0:
        case 1: value = 2; break;

        case 2:
        case 3: value = 0; break;

        case 4:
        case 5: value = 1; break;
      }

      break;

    /* These cases are harder.  These blocks are divided diagonally
     * by the angled edges of the hexagons.  Some slope positive, and
     * others negative.  We flip the x value of the negatively sloped
     * pieces.  Then we check to see if the point in question falls
     * in the upper or lower half of the block.  That info, plus the
     * z status of the block determines the color.
     */

    case 1:
    case 4:

      /* Map the point into the block at the origin. */

      xl = xs-xm;
      zl = zs-zm;

      /* These blocks have negative slopes so we flip it horizontally. */

      if (((xm + zm) % 2) == 1)
      {
        xl = 1.0 - xl;
      }

      /* Avoid a divide-by-zero error. */

      if (xl == 0.0)
      {
        xl = 0.0001;
      }

      /* Is the angle less-than or greater-than 45 degrees? */

      brkindx = (zl / xl) < 1.0;

      /* was...
       * brkindx = (atan2(zl,xl) < (45 * M_PI_180));
       * ...but because of the mapping, it's easier and cheaper,
       * CPU-wise, to just use a good ol' slope.
       */

      switch (brkindx)
      {
        case TRUE:

          switch (zm)
          {
            case 0:
            case 3: value = 0; break;

            case 2:
            case 5: value = 1; break;

            case 1:
            case 4: value = 2; break;
          }

          break;

        case FALSE:

          switch (zm)
          {
            case 0:
            case 3: value = 2; break;

            case 2:
            case 5: value = 0; break;

            case 1:
            case 4: value = 1; break;
          }

          break;
      }
  }

  value = fmod(value, 3.0);

  return(value);
}

/* In addition to clipping the value to 
   lie between 0.0 to 1.0, it also fudges 1.0-value.
 */

#define CLIP_DENSITY(r) if((r)<0.0){(r)=1.0;}else{if((r)>1.0){(r)=0.0;}else{(r)=1.0-(r);}}

static DBL planar_pattern (VECTOR EPoint)
{
  register DBL value;

  value = fabs(EPoint[Y]);
  CLIP_DENSITY(value);

  return(value);
}

static DBL spherical (VECTOR EPoint)
{
  register DBL value;

  VLength(value, EPoint);
  CLIP_DENSITY(value);

  return(value);
}

static DBL boxed (VECTOR EPoint)
{
  register DBL value;

  value = max(fabs(EPoint[X]), max(fabs(EPoint[Y]), fabs(EPoint[Z])));
  CLIP_DENSITY(value);

  return(value);
}

static DBL cylindrical (VECTOR EPoint)
{
  register DBL value;

  value = sqrt(Sqr(EPoint[X]) + Sqr(EPoint[Z]));
  CLIP_DENSITY(value);

  return(value);
}



/*****************************************************************************
*
* FUNCTION
*
*   PickInCube(tv, p1)
*
* INPUT
*
*   ?
*
* OUTPUT
*   
* RETURNS
*
*   long integer hash function used, to speed up cacheing.
*   
* AUTHOR
*
*   Jim McElhiney
*   
* DESCRIPTION
*
*   A subroutine to go with crackle.
*
*   Pick a random point in the same unit-sized cube as tv, in a
*   predictable way, so that when called again with another point in
*   the same unit cube, p1 is picked to be the same.
*
* CHANGES
*
******************************************************************************/

static long PickInCube(VECTOR tv, VECTOR  p1)
{
  int seed, temp;
  VECTOR flo;

  /*
   * This uses floor() not FLOOR, so it will not be a mirror
   * image about zero in the range -1.0 to 1.0. The viewer
   * won't see an artefact around the origin.
   */

  flo[X] = floor(tv[X] - EPSILON);
  flo[Y] = floor(tv[Y] - EPSILON);
  flo[Z] = floor(tv[Z] - EPSILON);

  seed = Hash3d((int)flo[X], (int)flo[Y], (int)flo[Z]);

  temp = POV_GET_OLD_RAND(); /* save current seed */
  
  POV_SRAND(seed);

  p1[X] = flo[X] + FRAND();
  p1[Y] = flo[Y] + FRAND();
  p1[Z] = flo[Z] + FRAND();

  POV_SRAND(temp);  /* restore */

  return((long)seed);
}



/*****************************************************************************
*
* FUNCTION
*
*   Evaluate_Pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   TPat   -- Texture pattern struct
*   
* OUTPUT
*   
* RETURNS
*
*   DBL result usual 0.0 to 1.0 but may be 2.0 in hexagon
*   
* AUTHOR
*
*   adapted from Add_Pigment by Chris Young
*   
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

DBL Evaluate_TPat (TPATTERN *TPat, VECTOR EPoint, VECTOR NPoint, VECTOR TPoint, RAY *ORay)
{
  DBL value = 0.0;
  /* VECTOR TPoint; */
  VECTOR APoint;
  VECTOR BPoint;
  Warp_EPoint (TPoint, EPoint, TPat);

  switch (TPat->Type)
  {
    case AGATE_PATTERN:    value = agate    (TPoint, TPat);   break;

    case BOZO_PATTERN:
    case SPOTTED_PATTERN:
    case BUMPS_PATTERN:    value = Noise    (TPoint);         break;

    case BRICK_PATTERN:    value = brick    (TPoint, TPat);   break;
    case CHECKER_PATTERN:  value = checker  (TPoint);         break;
    case SQUARE_PATTERN:   value = square   (TPoint);         break;
    case TERNAIRE_PATTERN: value = ternaire (TPoint);         break;
    case QUADRAN_PATTERN: value = quadran (TPoint);         break;
    case CLASSIQUE_PATTERN: value = classique (TPoint);         break;
    case TINTIN_PATTERN: 
        VAdd(APoint,EPoint,NPoint);
        Warp_EPoint (BPoint, APoint, TPat);
        VSub(APoint,BPoint,TPoint);
     value = tintin (TPoint,APoint,TPat);         break;
    case TINTINS_PATTERN: 
        VAdd(APoint,EPoint,NPoint);
        Warp_EPoint (BPoint, APoint, TPat);
        VSub(APoint,BPoint,TPoint);
       value = tintins (TPoint,APoint,TPat);         break;
    case CELLULO_PATTERN: 
        VAdd(APoint,EPoint,NPoint);
        Warp_EPoint (BPoint, APoint, TPat);
        VSub(APoint,BPoint,TPoint);
       value = cellulo (APoint,ORay);         break;
    case CRACKLE_PATTERN:  value = crackle  (TPoint);         break;
    case MATRIX_PATTERN:   value = cells    (TPoint);         break;
    case BLOTCH_PATTERN:   value = blotches (TPoint);         break;
/* JG8 start */
    case VORONOI_PATTERN:  value = voronoi  (TPoint);         break;
/* JG8 end */
/* Carrelage */
    case   CAR_CARRE_PATTERN:
              value = car_carre  (TPoint);         break;
    case   CAR_HEXAGONE_PATTERN:
              value = car_hexagone  (TPoint);         break;
    case   CAR_TRIANGLE_PATTERN:
              value = car_triangle  (TPoint);         break;
    case   CAR_LOZANGE_PATTERN:
              value = car_lozange  (TPoint);         break;
    case   CAR_ROMBHE_PATTERN:
              value = car_rombhe  (TPoint);         break;
    case   CAR_RECTANGLE_PATTERN:
              value = car_rectangle  (TPoint);         break;
    case   CAR_ORTHO_CARRE_PATTERN:
              value = car_ortho_carre  (TPoint);         break;
    case   CAR_CARRE_TRIANGLE_PATTERN:
              value = car_carre_triangle  (TPoint);         break;
    case   CAR_HEXA_TRIANGLE_PATTERN:
              value = car_hexa_triangle  (TPoint);         break;
    case   CAR_CARRE_DECALE_PATTERN:
              value = car_carre_decale  (TPoint);         break;
    case   CAR_RECTANGLE_CARRE_PATTERN:
              value = car_rectangle_carre  (TPoint);         break;
    case   CAR_CARRE_RECTANGLE_PATTERN:
              value = car_carre_rectangle  (TPoint);         break;
    case   CAR_CARRE_INTERNE_PATTERN:
              value = car_carre_interne  (TPoint);         break;
    case   CAR_CARRE_DOUBLE_PATTERN: 
              value = car_carre_double  (TPoint);         break;
    case   CAR_HEXA_CARRE_TRIANGLE_PATTERN: 
              value = car_hexa_carre_triangle  (TPoint);         break;
    case   CAR_RECTANGLE_PAIRE_PATTERN: 
              value = car_rectangle_paire  (TPoint);         break;
    case   CAR_HEXA_TRI_DROITE_PATTERN: 
              value = car_hexa_tri_droite  (TPoint);         break;
    case   CAR_HEXA_TRI_GAUCHE_PATTERN: 
              value = car_hexa_tri_gauche  (TPoint);         break;
    case   CAR_CARRE_TRI_PATTERN: 
              value = car_carre_tri  (TPoint);         break;
    case   CAR_DODECA_TRI_PATTERN: 
              value = car_dodeca_tri  (TPoint);         break;
    case   CAR_DODECA_HEX_PATTERN: 
              value = car_dodeca_hex  (TPoint);         break;
    case   CAR_DODECA_HEX5_PATTERN:
              value = car_dodeca_hex_5(TPoint);         break;
    case   CAR_HEXA_CARRE_TRIANGLE6_PATTERN:
              value = car_hexa_carre_triangle_6(TPoint);         break;
    case   CAR_CARRE_INTERNE5_PATTERN:
              value = car_carre_interne_5(TPoint);         break;
    case   CAR_TETRAG_PATTERN:
              value = tetragonal  (TPoint, TPat);         break;
    case   CAR_TRIG_PATTERN:
              value = trigonal  (TPoint, TPat);         break;
    case   CAR_HEXAG_PATTERN:
              value = sexagonal  (TPoint, TPat);         break;
/* Fin de Carrelage */

    case GRADIENT_PATTERN: value = gradient (TPoint, TPat);   break;
    case BAND_PATTERN:     value = bands    (TPoint, TPat);   break;
    case SHEET_PATTERN:    value = sheet    (TPoint, TPat);   break;
    case GRANITE_PATTERN:  value = granite  (TPoint);         break;
    case HEXAGON_PATTERN:  value = hexagon  (TPoint);         break;
    case LEOPARD_PATTERN:  value = leopard  (TPoint);         break;
    case MANDEL_PATTERN:   value = mandel   (TPoint, TPat);   break;
    case MARBLE_PATTERN:   value = marble   (TPoint, TPat);   break;
    case ONION_PATTERN:    value = onion    (TPoint);         break;
    case RADIAL_PATTERN:   value = radial   (TPoint);         break;
    case SPIRAL1_PATTERN:  value = spiral1  (TPoint, TPat);   break;
    case SPIRAL2_PATTERN:  value = spiral2  (TPoint, TPat);   break;
    case WOOD_PATTERN:     value = wood     (TPoint, TPat);   break;

    case WAVES_PATTERN:    value = waves_pigm    (TPoint, TPat);   break;
    case RIPPLES_PATTERN:  value = ripples_pigm  (TPoint, TPat);   break;
    case WRINKLES_PATTERN: value = wrinkles_pigm (TPoint);   break;
    case DENTS_PATTERN:    value = dents_pigm    (TPoint);   break;
    case QUILTED_PATTERN:  value = quilted_pigm  (TPoint, TPat);   break;

    case PLANAR_PATTERN:      value = planar_pattern (TPoint);      break;
    case BOXED_PATTERN:       value = boxed          (TPoint);      break;
    case SPHERICAL_PATTERN:   value = spherical      (TPoint);      break;
    case CYLINDRICAL_PATTERN: value = cylindrical    (TPoint);      break;
    case DENSITY_FILE_PATTERN:value = density_file (TPoint, TPat);  break;


    default: Error("Problem in Evaluate_TPat.");
  }

  if (TPat->Frequency !=0.0)
  {
    value = fmod(value * TPat->Frequency + TPat->Phase, 1.00001);
  }

  /* allow negative Frequency */

  if (value < 0.0)
  {
    value -= floor(value);
  }

  switch (TPat->Wave_Type)
  {
    case RAMP_WAVE:
      break;

    case SINE_WAVE:
      value = (1.0+cycloidal(value))*0.5;
      break;

    case TRIANGLE_WAVE:
      value = Triangle_Wave(value);
      break;

    case SCALLOP_WAVE:
      value = fabs(cycloidal(value*0.5));
      break;

    case CUBIC_WAVE:
      value = Sqr(value)*((-2.0 * value) + 3.0);
      break;

    case POLY_WAVE:
      value = pow(value, TPat->Exponent);
      break;

    default: Error("Unknown Wave Type %d.",TPat->Wave_Type);
   }

  return(value);
}


/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Init_TPat_Fields (TPATTERN *Tpat)
{
  Tpat->Type       = NO_PATTERN;
  Tpat->Wave_Type  = RAMP_WAVE;
  Tpat->Flags      = NO_FLAGS;
  Tpat->References = 1;
  Tpat->Exponent   = 1.0;
  Tpat->Frequency  = 1.0;
  Tpat->Phase      = 0.0;
  Tpat->Warps      = NULL;
  Tpat->Next       = NULL;
  Tpat->Blend_Map  = NULL;
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Copy_TPat_Fields (TPATTERN *New, TPATTERN  *Old)
{
  *New = *Old;
  
  /* Copy warp chain */
  New->Warps = Copy_Warps(Old->Warps);

  New->Blend_Map = Copy_Blend_Map(Old->Blend_Map);

  /* Note, cannot copy Old->Next because we don't know what kind of
     thing this is.  It must be copied by Copy_Pigment, Copy_Tnormal etc.
  */

  if (Old->Type == BITMAP_PATTERN)
  {
     New->Vals.Image = Copy_Image(Old->Vals.Image);
  }
  if (Old->Type == DENSITY_FILE_PATTERN)
  {
     New->Vals.Density_File = Copy_Density_File(Old->Vals.Density_File);
  }
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Destroy_TPat_Fields(TPATTERN *Tpat)
{
  Destroy_Warps(Tpat->Warps);
  Destroy_Blend_Map(Tpat->Blend_Map);
  /* Note, cannot destroy Tpat->Next nor pattern itself because we don't
     know what kind of thing this is.  It must be destroied by Destroy_Pigment, etc.
  */

  if (Tpat->Type == BITMAP_PATTERN)
  {
     Destroy_Image(Tpat->Vals.Image);
  }

  if (Tpat->Type == DENSITY_FILE_PATTERN)
  {
     Destroy_Density_File(Tpat->Vals.Density_File);
  }
}




/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

TURB *Create_Turb()
{
  TURB *New;

  New = (TURB *)POV_MALLOC(sizeof(TURB),"turbulence struct");

  Make_Vector(New->Turbulence, 0.0, 0.0, 0.0);

  New->Octaves = 6;
  New->Omega = 0.5;
  New->Lambda = 2.0;

  return(New);
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

#if 0   /* Unused function [AED] */
static TURB *Copy_Turb(TURB *Old)
{
  TURB *New;

  if (Old != NULL)
  {
    New = Create_Turb();

    *New = *Old;
  }
  else
  {
    New=NULL;
  }

  return(New);
}
#endif


/*****************************************************************************
*
* FUNCTION
*
*   Translate_Tpattern
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Translate_Tpattern(TPATTERN *Tpattern,VECTOR Vector)
{
  TRANSFORM Trans;

  if (Tpattern != NULL)
  {
    Compute_Translation_Transform (&Trans, Vector);

    Transform_Tpattern (Tpattern, &Trans);
  }
}



/*****************************************************************************
*
* FUNCTION
*
*   Rotate_Tpattern
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Rotate_Tpattern(TPATTERN *Tpattern,VECTOR Vector)
{
  TRANSFORM Trans;

  if (Tpattern != NULL)
  {
    Compute_Rotation_Transform (&Trans, Vector);

    Transform_Tpattern (Tpattern, &Trans);
  }
}



/*****************************************************************************
*
* FUNCTION
*
*   Scale_Tpattern
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Scale_Tpattern(TPATTERN *Tpattern,VECTOR Vector)
{
  TRANSFORM Trans;

  if (Tpattern != NULL)
  {
    Compute_Scaling_Transform (&Trans, Vector);

    Transform_Tpattern (Tpattern, &Trans);
  }
}



/*****************************************************************************
*
* FUNCTION
*
*   Transform_Tpattern
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Transform_Tpattern(TPATTERN *Tpattern,TRANSFORM *Trans)
{
  WARP *Temp;

  if (Tpattern != NULL)
  {
    if (Tpattern->Warps == NULL)
    {
      Tpattern->Warps=Create_Warp(TRANSFORM_WARP);
    }
    else
    {
      if (Tpattern->Warps->Warp_Type != TRANSFORM_WARP)
      {
        Temp=Tpattern->Warps;

        Tpattern->Warps=Create_Warp(TRANSFORM_WARP);

        Tpattern->Warps->Next_Warp=Temp;
      }
    }

    Compose_Transforms (&( ((TRANS *)(Tpattern->Warps))->Trans), Trans);
  }
}


/*****************************************************************************
*
* FUNCTION
*
*   ripples_pigm
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   TPat   -- Texture pattern struct
*
* OUTPUT
*   
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*   
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION   : Note this pattern is only used for pigments and textures.
*                 Normals have a specialized pattern for this.
*
* CHANGES
*   Nov 1994 : adapted from normal by [CY]
*
******************************************************************************/

static DBL ripples_pigm (VECTOR EPoint, TPATTERN *TPat)
{
  register unsigned int i;
  register DBL length, index;
  DBL scalar =0.0;
  VECTOR point;

  for (i = 0 ; i < Number_Of_Waves ; i++)
  {
    VSub (point, EPoint, Wave_Sources[i]);
    VLength (length, point);

    if (length == 0.0)
      length = 1.0;

    index = length * TPat->Frequency + TPat->Phase;

    scalar += cycloidal(index);
  }

  scalar = 0.5*(1.0+(scalar / (DBL)Number_Of_Waves));

  return(scalar);
}


/*****************************************************************************
*
* FUNCTION
*
*   waves_pigm
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   TPat   -- Texture pattern struct
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION   : Note this pattern is only used for pigments and textures.
*                 Normals have a specialized pattern for this.
*
* CHANGES
*   Nov 1994 : adapted from normal by [CY]
*
******************************************************************************/

static DBL waves_pigm (VECTOR EPoint, TPATTERN *TPat)
{
  register unsigned int i;
  register DBL length, index;
  DBL scalar = 0.0;
  VECTOR point;

  for (i = 0 ; i < Number_Of_Waves ; i++)
  {
    VSub (point, EPoint, Wave_Sources[i]);
    VLength (length, point);

    if (length == 0.0)
    {
      length = 1.0;
    }

    index = length * TPat->Frequency * frequency[i] + TPat->Phase;

    scalar += cycloidal(index)/frequency[i];
  }

  scalar = 0.2*(2.5+(scalar / (DBL)Number_Of_Waves));

  return(scalar);
}


/*****************************************************************************
*
* FUNCTION
*
*   dents_pigm
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*   
* AUTHOR
*
*   POV-Ray Team
*   
* DESCRIPTION   : Note this pattern is only used for pigments and textures.
*                 Normals have a specialized pattern for this.
*
* CHANGES
*   Nov 1994 : adapted from normal by [CY]
*
******************************************************************************/

static DBL dents_pigm (VECTOR EPoint)
{
  DBL noise;

  noise = Noise (EPoint);

  return(noise * noise * noise);
}



/*****************************************************************************
*
* FUNCTION
*
*   wrinkles_pigm
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   
* OUTPUT
*   
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*   
* AUTHOR
*
*   POV-Ray Team
*   
* DESCRIPTION   : Note this pattern is only used for pigments and textures.
*                 Normals have a specialized pattern for this.
*
* CHANGES
*   Nov 1994 : adapted from normal by [CY]
*
******************************************************************************/


static DBL wrinkles_pigm (VECTOR EPoint)
{
  register int i;
  DBL lambda = 2.0;
  DBL omega = 0.5;
  DBL value;
  VECTOR temp;

  value = Noise(EPoint);

  for (i = 1; i < 10; i++)
  {
    VScale(temp,EPoint,lambda);

    value += omega * Noise(temp);

    lambda *= 2.0;

    omega *= 0.5;
  }

  return(value/2.0);
}



/*****************************************************************************
*
* FUNCTION
*
*   quilted_pigm
*
* INPUT
*   
* OUTPUT
*   
* RETURNS
*   
* AUTHOR
*
*   Dan Farmer & Chris Young
*   
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

static DBL quilted_pigm (VECTOR EPoint, TPATTERN *TPat)
{
  VECTOR value;
  DBL t;

  value[X] = EPoint[X]-FLOOR(EPoint[X])-0.5;
  value[Y] = EPoint[Y]-FLOOR(EPoint[Y])-0.5;
  value[Z] = EPoint[Z]-FLOOR(EPoint[Z])-0.5;

  t = sqrt(value[X]*value[X]+value[Y]*value[Y]+value[Z]*value[Z]);

  t = quilt_cubic(t, TPat->Vals.Quilted.Control0, TPat->Vals.Quilted.Control1);

  value[X] *= t;
  value[Y] *= t;
  value[Z] *= t;

  return((fabs(value[X])+fabs(value[Y])+fabs(value[Z]))/3.0);
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

#define INV_SQRT_3_4 1.154700538
DBL quilt_cubic(DBL t,DBL p1,DBL p2)
{
 DBL it=(1-t);
 DBL itsqrd=it*it;
/* DBL itcubed=it*itsqrd; */
 DBL tsqrd=t*t;
 DBL tcubed=t*tsqrd;
 DBL val;

/* Originally coded as...

 val= (DBL)(itcubed*n1+(tcubed)*n2+3*t*(itsqrd)*p1+3*(tsqrd)*(it)*p2);

 re-written by CEY to optimise because n1=0 n2=1 always.
 
*/

 val = (tcubed + 3.0*t*itsqrd*p1 + 3.0*tsqrd*it*p2) * INV_SQRT_3_4;
 
 return(val);
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Search_Blend_Map (DBL value,BLEND_MAP *Blend_Map,BLEND_MAP_ENTRY **Prev,BLEND_MAP_ENTRY  **Cur)
{
  BLEND_MAP_ENTRY *P, *C;
  int Max_Ent=Blend_Map->Number_Of_Entries-1;

  /* if greater than last, use last. */

  if (value >= Blend_Map->Blend_Map_Entries[Max_Ent].value)
  {
    P = C = &(Blend_Map->Blend_Map_Entries[Max_Ent]);
  }
  else
  {
    P = C = &(Blend_Map->Blend_Map_Entries[0]);

    while (value > C->value)
    {
      P = C++;
    }
  }

  if (value == C->value)
  {
    P = C;
  }

  *Prev = P;
  *Cur  = C;
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

static TURB *Search_For_Turb(WARP *Warps)
{
  WARP* Temp=Warps;

  if (Temp!=NULL)
  {
    while (Temp->Next_Warp != NULL)
    {
      Temp=Temp->Next_Warp;
    }

    if (Temp->Warp_Type != CLASSIC_TURB_WARP)
    {
       Temp=NULL;
    }
  }

  return ((TURB *)Temp);
}


/*****************************************************************************
*
* FUNCTION
*
*   density_file
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Dieter Bayer
*
* DESCRIPTION
*
* CHANGES
*
*   Dec 1996 : Creation.
*
******************************************************************************/

static DBL density_file(VECTOR EPoint, TPATTERN *TPat)
{
  int x, y, z;
  int x1, y1, z1;
  int x2, y2, z2;
  DBL xx, yy, zz;
  DBL xi, yi, zi;
  DBL f111, f112, f121, f122, f211, f212, f221, f222;
  DBL density = 0.0;
  DENSITY_FILE_DATA *Data;

  if ((TPat->Vals.Density_File != NULL) &&
      ((Data = TPat->Vals.Density_File->Data) != NULL))
  {
    if ((EPoint[X] >= 0.0) && (EPoint[X] < 1.0) &&
        (EPoint[Y] >= 0.0) && (EPoint[Y] < 1.0) &&
        (EPoint[Z] >= 0.0) && (EPoint[Z] < 1.0))
    {
      switch (TPat->Vals.Density_File->Interpolation)
      {
        case NO_INTERPOLATION:

          x = (int)(EPoint[X] * (DBL)Data->Sx);
          y = (int)(EPoint[Y] * (DBL)Data->Sy);
          z = (int)(EPoint[Z] * (DBL)Data->Sz);

          if ((x < 0) || (x >= Data->Sx) ||
              (y < 0) || (y >= Data->Sy) ||
              (z < 0) || (z >= Data->Sz))
          {
            density = 0.0;
          }
          else
          {
            density = (DBL)Data->Density[z][y][x] / 255.0;
          }

          break;

        case TRILINEAR_INTERPOLATION:

          xx = EPoint[X] * (DBL)(Data->Sx - 1);
          yy = EPoint[Y] * (DBL)(Data->Sy - 1);
          zz = EPoint[Z] * (DBL)(Data->Sz - 1);

          x1 = (int)xx;
          y1 = (int)yy;
          z1 = (int)zz;

          x2 = x1 + 1;
          y2 = y1 + 1;
          z2 = z1 + 1;

          xx -= floor(xx);
          yy -= floor(yy);
          zz -= floor(zz);

          xi = 1.0 - xx;
          yi = 1.0 - yy;
          zi = 1.0 - zz;

          f111 = (DBL)Data->Density[z1][y1][x1] / 255.0;
          f112 = (DBL)Data->Density[z1][y1][x2] / 255.0;
          f121 = (DBL)Data->Density[z1][y2][x1] / 255.0;
          f122 = (DBL)Data->Density[z1][y2][x2] / 255.0;
          f211 = (DBL)Data->Density[z2][y1][x1] / 255.0;
          f212 = (DBL)Data->Density[z2][y1][x2] / 255.0;
          f221 = (DBL)Data->Density[z2][y2][x1] / 255.0;
          f222 = (DBL)Data->Density[z2][y2][x2] / 255.0;

          density = f111 * zi * yi * xi +
                    f112 * zi * yi * xx +
                    f121 * zi * yy * xi +
                    f122 * zi * yy * xx +
                    f211 * zz * yi * xi +
                    f212 * zz * yi * xx +
                    f221 * zz * yy * xi +
                    f222 * zz * yy * xx;

          break;
      }
    }
    else
    {
      density = 0.0;
    }

/*
    fprintf(stderr, "x = %3d, y = %3d, z = %3d, density = %5.4f\n", x, y, z, density);
*/
  }
  return(density);
}


/*****************************************************************************
*
* FUNCTION
*
*   Create_Density_File
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Dieter Bayer
*
* DESCRIPTION
*
*   Create a density file structure.
*
* CHANGES
*
*   Dec 1996 : Creation.
*
******************************************************************************/

DENSITY_FILE *Create_Density_File()
{
  DENSITY_FILE *New;

  New = (DENSITY_FILE *)POV_MALLOC(sizeof(DENSITY_FILE), "density file");

  New->Interpolation = NO_INTERPOLATION;

  New->Data = (DENSITY_FILE_DATA *)POV_MALLOC(sizeof(DENSITY_FILE_DATA), "density file data");

  New->Data->References = 1;

  New->Data->Name = NULL;

  New->Data->Sx =
  New->Data->Sy =
  New->Data->Sz = 0;

  New->Data->Density = NULL;

  return (New);
}



/*****************************************************************************
*
* FUNCTION
*
*   Copy_Density_File
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Dieter Bayer
*
* DESCRIPTION
*
*   Copy a density file structure.
*
* CHANGES
*
*   Dec 1996 : Creation.
*
******************************************************************************/

DENSITY_FILE *Copy_Density_File(DENSITY_FILE *Old)
{
  DENSITY_FILE *New;

  if (Old != NULL)
  {
    New = (DENSITY_FILE *)POV_MALLOC(sizeof(DENSITY_FILE), "density file");

    *New = *Old;

    New->Data->References++;
  }
  else          /* tw */
    New = NULL; /* tw */
    
  return(New);
}



/*****************************************************************************
*
* FUNCTION
*
*   Destroy_Density_File
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Dieter Bayer
*
* DESCRIPTION
*
*   Destroy a density file structure.
*
* CHANGES
*
*   Dec 1996 : Creation.
*
******************************************************************************/

void Destroy_Density_File(DENSITY_FILE *Density_File)
{
  int y, z;

  if (Density_File != NULL)
  {
    if ((--(Density_File->Data->References)) == 0)
    {
      POV_FREE(Density_File->Data->Name);

      for (z = 0; z < Density_File->Data->Sz; z++)
      {
        for (y = 0; y < Density_File->Data->Sy; y++)
        {
          POV_FREE(Density_File->Data->Density[z][y]);
        }

        POV_FREE(Density_File->Data->Density[z]);
      }
      POV_FREE(Density_File->Data->Density);
      POV_FREE(Density_File->Data);
    }

    POV_FREE(Density_File);
  }
}


void Read_Density_File(DENSITY_FILE *df)
{
  unsigned char ***map;
  int y, z, sx, sy, sz;
  FILE *file;

  if (df == NULL)
  {
     return;
  }
  
  /* Allocate and read density file. */

  if ((df != NULL) && (df->Data->Name != NULL))
  {
    if ((file = Locate_File(df->Data->Name, READ_BINFILE_STRING, ".df3", ".DF3",NULL,TRUE)) == NULL)
    {
      Error("Cannot read media density file.\n");
    }
    
    sx = df->Data->Sx = readushort(file);
    sy = df->Data->Sy = readushort(file);
    sz = df->Data->Sz = readushort(file);

    map = (unsigned char ***)POV_MALLOC(sz*sizeof(unsigned char **), "media density file data");

    for (z = 0; z < sz; z++)
    {
      map[z] = (unsigned char **)POV_MALLOC(sy*sizeof(unsigned char *), "media density file data");

      for (y = 0; y < sy; y++)
      {
        map[z][y] = (unsigned char *)POV_MALLOC(sx*sizeof(unsigned char), "media density file data");

        fread(map[z][y], sizeof(unsigned char), (size_t)sx, file);
      }
    }

    df->Data->Density = map;

    if (file != NULL)		/* -hdf99- */
    {
      fclose(file);
    }

  }
}

static unsigned short readushort(FILE *infile)
{
  int i0, i1 = 0; /* To quiet warnings */

  if ((i0  = fgetc(infile)) == EOF || (i1  = fgetc(infile)) == EOF)
  {
    Error("Error reading density_file\n");
  }

  return (unsigned short)((((unsigned short)i0) << 8) | ((unsigned short)i1));
}


/* JG2 start */

/*****************************************************************************
*
* FUNCTION
*
*   blotches
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   John VanSickle, based on code by Jim McElhiney
*
* DESCRIPTION
*
*   "blotches":
*
*   New colour function by John VanSickle,
*     vansickl@erols.com
*
*   Assigns a pseudorandom point to each unit cube, and another float from 0 to
*   1 to each cube.  The nearest pseudorandom point to the evaluated point is
*   selected, and its associated pseudorandom value is returned.  It creates
*   regions of uniform value.
*
*   All "blotches" specific source code and examples are in the public domain.
*
* CHANGES
*   Oct 1994    : original crackle code adapted by [CY]
*   Jul 1999    : adapted for a different pattern by [JV]
*
******************************************************************************/

static DBL blotches (VECTOR EPoint)
{
  int    i,j,temp,seed;
  long   thisseed;
  DBL    sum, minsum, tf;
  VECTOR sv, tv, dv, t1, add;

  static int bvc;
  static long lastbeed = 0x80000000;
  static VECTOR bv[81];

  Assign_Vector(tv,EPoint);

  /*
   * Check to see if the input point is in the same unit cube as the last
   * call to this function, to use cache of cubelets for speed.
   */

  thisseed = PickInCube(tv, t1);

  if (thisseed != lastbeed)
  {
    /*
     * No, not same unit cube.  Calculate the random points for this new
     * cube and its 80 neighbours which differ in any axis by 1 or 2.
     * Why distance of 2?  If there is 1 point in each cube, located
     * randomly, it is possible for the closest random point to be in the
     * cube 2 over, or the one two over and one up.  It is NOT possible
     * for it to be two over and two up.  Picture a 3x3x3 cube with 9 more
     * cubes glued onto each face.
     */

    /* Now store a points for this cube and each of the 80 neighbour cubes. */

    bvc = 0;

    for (add[X] = -2.0; add[X] < 2.5; add[X] +=1.0)
    {
      for (add[Y] = -2.0; add[Y] < 2.5; add[Y] += 1.0)
      {
        for (add[Z] = -2.0; add[Z] < 2.5; add[Z] += 1.0)
        {
          /* For each cubelet in a 5x5 cube. */

          if ((fabs(add[X])>1.5)+(fabs(add[Y])>1.5)+(fabs(add[Z])>1.5) <= 1.0)
          {
            /* Yes, it's within a 3d knight move away. */

            VAdd(sv, tv, add);

            PickInCube(sv, t1);

            bv[bvc][X] = t1[X];
            bv[bvc][Y] = t1[Y];
            bv[bvc][Z] = t1[Z];
            bvc++;
          }
        }
      }
    }

    lastbeed = thisseed;
  }

  /*
   * Find the point with the shortest distance from the input point.
   * Loop invariant:  minsum is shortest dist
   */

  VSub(dv, bv[0], tv);  minsum  = VSumSqr(dv);
  j=0;

  /* Loop for the 81 cubelets to find closest. */

  for (i = 1; i < bvc; i++)
  {
    VSub(dv, bv[i], tv);

    sum = VSumSqr(dv);

    if (sum < minsum) {
      minsum = sum;
      j=i;
    }
  }

  /* select a random value based on the cube from which this came. */

  seed = Hash3d((int)bv[j][X], (int)bv[j][Y], (int)bv[j][Z]);

  temp = POV_GET_OLD_RAND(); /* save current seed */
  
  POV_SRAND(seed);

  /* The first three values were used for the random point, and shouldn't
     be used for the value of the region */

  FRAND();
  FRAND();
  FRAND();

  tf = FRAND();

  POV_SRAND(temp);  /* restore */

  return min(tf,1.0);
}

/*****************************************************************************
*
* FUNCTION
*
*   cells
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   John VanSickle
*
* DESCRIPTION
*
*   "cells":
*
*   New colour function by John VanSickle,
*     vansickl@erols.com
*
*   Assigns a pseudorandom value to each unit cube.  The value for the cube in
*   which the evaluted point lies is returned.
*
*   All "cells" specific source code and examples are in the public domain.
*
* CHANGES
*   Oct 1994    : original crackle code adapted by [CY]
*   Jul 1999    : adapted for a different pattern by [JV]
*
******************************************************************************/

static DBL cells (VECTOR tv)
{
  int    temp,seed;
  DBL    tf;

  VECTOR flo;

  /*
   * This uses floor() not FLOOR, so it will not be a mirror
   * image about zero in the range -1.0 to 1.0. The viewer
   * won't see an artefact around the origin.
   */

  flo[X] = floor(tv[X] - EPSILON);
  flo[Y] = floor(tv[Y] - EPSILON);
  flo[Z] = floor(tv[Z] - EPSILON);

  seed = Hash3d((int)flo[X], (int)flo[Y], (int)flo[Z]);

  /* select a random value based on the cube from which this came. */

  temp = POV_GET_OLD_RAND(); /* save current seed */
  
  POV_SRAND(seed);

  tf = FRAND();

  POV_SRAND(temp);  /* restore */

  return min(tf,1.0);
}

/*****************************************************************************
*
* FUNCTION
*
*   bands
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   
* OUTPUT
*   
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*   
* AUTHOR
*
*   John VanSickle, based on the gradient code
*   
* DESCRIPTION
*
*   banded Pattern - just like gradient, but is not mirrored around the origin,
*   and the bands are of uniform width, regardless of the direction of the
*   vector.
*
* CHANGES
*   Oct 1994    : adapted from pigment by [CY]
*   Jul 1999    : adapted from gradient by [JV]
*
******************************************************************************/

/* MUY IMPORTANTE! This function expects the Vals.Gradient vector in TPat to
 * be normalized (ie, unit length).  If the vector is greater or less than
 * unit length, there will be a scaling effect on the pattern.  I could have
 * divided by the magnitude of the vector, but since this function will be
 * called more than once by the renderer, it is more appropriate for the
 * parser to normalize the vector (where it need be done only once). */

static DBL bands (VECTOR EPoint, TPATTERN *TPat)
{
  register DBL value;

  value = EPoint[X]*TPat->Vals.Gradient[X]
        + EPoint[Y]*TPat->Vals.Gradient[Y]+EPoint[Z]*TPat->Vals.Gradient[Z];

  return(value-floor(value));
}

/*****************************************************************************
*
* FUNCTION
*
*   sheet
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   
* OUTPUT
*   
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*   
* AUTHOR
*
*   John VanSickle, based on the gradient code
*   
* DESCRIPTION
*
*   Takes the dot product of the evaluated point and a specified vector.
*   If the product is <0, 0 is returned.  If the product is >1, 1 is returned.
*   Otherwise the product is returned.  This is basically a non-repeating,
*   non-mirrored gradient, but like the banded pattern above, the width of
*   the band is independent of direction of the vector.
*
* CHANGES
*   Oct 1994    : adapted from pigment by [CY]
*   Jul 1999    : adapted from gradient by [JV]
*
******************************************************************************/

/* MUY IMPORTANTE! This function expects the Vals.Gradient vector in TPat to
 * be normalized (ie, unit length).  If the vector is greater or less than
 * unit length, there will be a scaling effect on the pattern.  I could have
 * divided by the magnitude of the vector, but since this function will be
 * called more than once by the renderer, it is more appropriate for the
 * parser to normalize the vector (where it need be done only once). */

static DBL sheet (VECTOR EPoint, TPATTERN *TPat)
{
  register DBL value;

  value = EPoint[X]*TPat->Vals.Gradient[X]
        + EPoint[Y]*TPat->Vals.Gradient[Y]+EPoint[Z]*TPat->Vals.Gradient[Z];

  return(value>1.0 ? 1.0 : (value<0.0? 0.0:value));
}

/* JG2 end */

/* JG8 start */

/*****************************************************************************
*
* FUNCTION
*
*   voronoi
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   J. Grimbert, based on code by Jim McElhiney
*
* DESCRIPTION
*
*   "voronoi":
*
*   New colour function by J. Grimbert
*     jgrimbert@atos-group.com
*
*   Assigns a pseudorandom point to each unit cube
*   The nearest pseudorandom point to the evaluated point is
*   selected, and a distance from it is returned.  It creates
*   regions of increasing value, just like boxed but with 
*   a more random placement of the cells.
*   In fact, its a normalized crackle... sort of
*
*   All "voronoi" specific source code and examples are in the public domain.
*
******************************************************************************/

static DBL voronoi (VECTOR EPoint)
{
  int    i;
  long   thisseed;
  DBL    sum, minsum, minsum2, tf;
  VECTOR sv, tv, dv, t1, add;

  static int cvc;
  static long lastseed = 0x80000000;
  static VECTOR cv[81];

  Assign_Vector(tv,EPoint);

  /*
   * Check to see if the input point is in the same unit cube as the last
   * call to this function, to use cache of cubelets for speed.
   */

  thisseed = PickInCube(tv, t1);

  if (thisseed != lastseed)
  {
    /*
     * No, not same unit cube.  Calculate the random points for this new
     * cube and its 80 neighbours which differ in any axis by 1 or 2.
     * Why distance of 2?  If there is 1 point in each cube, located
     * randomly, it is possible for the closest random point to be in the
     * cube 2 over, or the one two over and one up.  It is NOT possible
     * for it to be two over and two up.  Picture a 3x3x3 cube with 9 more
     * cubes glued onto each face.
     */

    /* Now store a points for this cube and each of the 80 neighbour cubes. */

    cvc = 0;

    for (add[X] = -2.0; add[X] < 2.5; add[X] +=1.0)
    {
      for (add[Y] = -2.0; add[Y] < 2.5; add[Y] += 1.0)
      {
        for (add[Z] = -2.0; add[Z] < 2.5; add[Z] += 1.0)
        {
          /* For each cubelet in a 5x5 cube. */

          if ((fabs(add[X])>1.5)+(fabs(add[Y])>1.5)+(fabs(add[Z])>1.5) <= 1.0)
          {
            /* Yes, it's within a 3d knight move away. */

            VAdd(sv, tv, add);

            PickInCube(sv, t1);

            cv[cvc][X] = t1[X];
            cv[cvc][Y] = t1[Y];
            cv[cvc][Z] = t1[Z];
            cvc++;
          }
        }
      }
    }

    lastseed = thisseed;
  }

  /*
   * Find the 2 points with the 2 shortest distances from the input point.
   * Loop invariant:  minsum is shortest dist, minsum2 is 2nd shortest
   */

  /* Set up the loop so the invariant is true:  minsum <= minsum2 */

  VSub(dv, cv[0], tv);  minsum  = VSumSqr(dv);
  VSub(dv, cv[1], tv);  minsum2 = VSumSqr(dv);

  if (minsum2 < minsum)
  {
    tf = minsum; minsum = minsum2; minsum2 = tf;
  }

  /* Loop for the 81 cubelets to find closest and 2nd closest. */

  for (i = 2; i < cvc; i++)
  {
    VSub(dv, cv[i], tv);

    sum = VSumSqr(dv);

    if (sum < minsum)
    {
      minsum2 = minsum;
      minsum = sum;
    }
    else
    {
      if (sum < minsum2)
      {
        minsum2 = sum;
      }
    }
  }

  /* voronoi value is 1. at the voronoi boundary, 0. at a node */

  tf= (minsum / minsum2); /* minsum is know <= minsum2 and minsum2 != 0 */


  return min(tf, 1.);
}
/* Carrelage */


/*****************************************************************************
*
* FUNCTIONS
*
*   related to "Carrelage"
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   New colour function by J. Grimbert
*     jerome.grimbert@atosorigin.com
*
*   Fill the X-Z plane with repeating pattern
*   When there is only one kind of form, 1 is the outside, 0 is most inside
*   When there is two kind of form, 0 & 1 are the most inside, 1/2 is
*    the outside
*
******************************************************************************/

static DBL car_carre  (VECTOR EPoint)         
{
/* 
** Carrelage carre classique
** On se ramene dans le carre [0,1][0,1]
** la valeur est le double de la distance(1) depuis 1/2,1/2
*/
DBL x,z;
x = fabs(EPoint[X]);
x -= floor(x);
x = 2*fabs( x-0.5 );

z = fabs(EPoint[Z]);
z -= floor(z);
z = 2*fabs( z-0.5 );

return max(x,z);
}

static DBL car_hexagone  (VECTOR EPoint)         
{
/* 
** Carrelage avec un hexagone classique
*/
DBL x,z;
DBL dist1,dist2;
DBL answer;

x=EPoint[X];
z=EPoint[Z];
 x += 0.5;
 x -= 3.0*floor(x/3.0);
 z -= SQR3*floor(z/(SQR3));
 /* x,z is in { [0.0, 3.0 [, [0.0, SQR3 [ } 
 ** but there is some symetry to simplify the testing
 */
 if (z > SQR3_2 )
 {
    z = SQR3 - z; 
 } 
 /* 
 ** Now only [0,3[,[0,SQR3/2[
 */
 if (x > 1.5)
 {
    x -= 1.5 ; /* translate */
    z = SQR3_2 -z; /* mirror */
 }
 /*
 ** And now, it is even simpler :  [0,1.5],[0,SQR3/2]
 ** on the bottom left corner, part of some other hexagone
 ** on the top right corner, center of the hexagone
 */
 if ((SQR3*x+z)<SQR3_2)
 {
    x = 0.5 - x;
    z = SQR3_2 -z; /* mirror */
 }
 if (x > 1.0)
 {
    x = 2.0 -x; /* mirror */
 }
   /* Hexagone */
    dist1 = 1.0 - ( z / SQR3_2 );
    dist2 = 1.0 - (((SQR3 * x + z - SQR3_2) ) / SQR3 );
    answer = max(dist1,dist2);
    answer = min(1.0,answer);
    answer = max(0.0,answer);
 return answer;
}

static DBL car_triangle  (VECTOR EPoint)         
{
DBL x,z;
DBL slop1;
DBL dist1,dist2;
int delta;
x=EPoint[X];
z=EPoint[Z];
 delta = 0;
 x -= floor(x);
 z -= SQR3*floor(z/SQR3);
 /* x,z is in { [0.0, 1.0 [, [0.0, SQR3 [ } 
 ** but there is some symetry to simplify the testing
 */
 if (z > SQR3_2 )
 {
    z = SQR3 - z; /* mirror */
    delta = 1 - delta;
 } 
 if (x > 0.5)
 {
    x = 1.0 - x; /* mirror */
 }
 if (x != 0.0)
 {
    slop1 = z/x;
    if (slop1>SQR3)
    { 
      z = SQR3_2 - z;
      x = 0.5 -x;
      delta = 1 - delta;
    }
 }
 else
 {
   z = SQR3_2 -z;
   x = 0.5;
 }
 dist1 = 1.0 - (z * 2 * SQR3 );
 dist2 = 1.0 - ((SQR3 * x - z) * SQR3 );
 return delta+(0.5-delta)*max(dist1,dist2);
}

static DBL car_lozange  (VECTOR EPoint)         
{
DBL x,z;
DBL slop1;
DBL dist1,dist2;

x=EPoint[X];
z=EPoint[Z];

 x -= floor(x);
 z -= SQR3*floor(z/SQR3);
 /* x,z is in { [0.0, 1.0 [, [0.0, SQR3 [ } 
 ** There is some mirror to reduce the problem
 */
 if ( z > SQR3_2 )
 {
    z -= SQR3_2;
    x += 0.5;
 }
 if ( (2.0*z) > SQR3_2 )
 {
    z = SQR3_2 - z;
    x = 1.5 - x;
 }
 if (x > 0.75)
 {
    x -= 1.0;
 }
 if (x != 0.0)
 {
    slop1 = z/x;
    if (slop1>SQR3)
    { 
      z = SQR3_2 - z;
      x = 0.5 -x;
    }
 }
 dist1 = 1.0 - (z * 4.0 * SQR3 / 3.0 );
 dist2 = 1.0 - (fabs(SQR3 * x - z) * SQR3 *2.0 / 3.0);
 return max(dist1,dist2);
}
 
static DBL car_rombhe  (VECTOR EPoint)         
{
DBL x,z;
DBL answer;
DBL slop1;
DBL dist1,dist2;
int delta;
x=EPoint[X];
z=EPoint[Z];
delta = 0;
 x += 0.5;
 x -= 3.0*floor(x/3.0);
 z -= SQR3*floor(z/SQR3);
 /* x,z is in { [0.0, 3.0 [, [0.0, SQR3 [ } 
 ** There is some mirror to reduce the problem
 */
 if ( z > SQR3_2 )
 {
    z = SQR3 -z; /* mirror */
    delta = 2 - delta;
 }
 if (x > 1.5)
 {
    x -= 1.5 ; /* translate */
    z = SQR3_2 -z; /* mirror */
    delta = 2 - delta;
 } 
 /* Now in [0,1.5],[0,SQR3/2] 
 ** from left to right
 ** part of a horizontal (z=0)
 ** half a vertical 
 ** part of a horizontal 
 */
 if (x < 0.5)
 {
   /* mirrror */
   x = 1.0 - x;
   delta = 2 - delta;
 }
 /* 
 ** Let shift the [0.5,1.5],[0,SQR3/2] to [0,1]....
 */
 x -= 0.5;
 if (x != 0.0)
 {
    slop1 = z/x;
    if (slop1>SQR3)
    { /* rotate the vertical to match the horizontal on the right */
      dist1 = ( x / 2.0 ) + ( z * SQR3_2 ); 
      dist2 = ( z / 2.0 ) - ( x * SQR3_2 ); 
      z = dist2;
      x = dist1;
      delta = 1;
    }
 }
 else
 {
    /* rotate the vertical to match the horizontal on the right */
     dist1 = ( x / 2.0 ) + ( z * SQR3_2 ); 
     dist2 = ( z / 2.0 ) - ( x * SQR3_2 ); 
     z = dist2;
     x = dist1;
     delta = 1;
 }
 /* It may be similar to lozange (in fact, IT IS !), now */

 if ( (2.0*z) > SQR3_2 )
 {
    z = SQR3_2 - z;
    x = 1.5 - x;
 }
 if (x > 0.75)
 {
    x -= 1.0;
 }
 if (x != 0.0)
 {
    slop1 = z/x;
    if (slop1>SQR3)
    { 
      z = SQR3_2 - z;
      x = 0.5 -x;
    }
 }
 dist1 = 1.0 - (z * 4.0 * SQR3 / 3.0 );
 dist2 = 1.0 - (fabs(SQR3 * x - z) * SQR3 *2.0 / 3.0);
 answer = max(dist1,dist2);
 answer /= 3.0;
 if (delta == 2)
 {
 answer = 1 - answer;
 }
 else
 {
 answer += delta/3.0;
 }
 answer = min(1.0,answer);
 answer = max(0.0,answer);
 return answer;
}

static DBL car_rectangle  (VECTOR EPoint)         
{
/*
** Carrelage avec deux rectangles (2x1)
** On se ramene dans le carre [0,4][0,4]
** puis on divise ce carre en 16 cas
**
**  +-----+--+  +
**  |     |  |  |
**  +--+--+  +--+
**     |  |  |
**  +--+  +--+--+
**  |  |  |     |
**  +  +--+--+--+
**  |  |     |  |
**  +--+-----+  +
*/
DBL x,z;
DBL delta;
DBL answer;
int valueX,valueZ;
x=EPoint[X];
z=EPoint[Z];                                                                    
x -= 4.0*floor(x/4.0);
z -= 4.0*floor(z/4.0);
  valueX = (int)x;
  valueZ = (int)z;
  delta = 1.0;
  switch((valueX+valueZ*4))                                                       
  {
  case 0:
  case 4:
     z -= 1.0;
     break;
  case 1:
  case 2:
     x -= 2.0;
     delta = 0.0;
     break;
  case 3:
     x -= 3.0;
     break;
  case 5:
  case 9:
     x -= 1.0;
     z -= 2.0;
     break;
  case 6:
  case 7:
     x -= 3.0;
     z -= 1.0;
     delta = 0.0;
     break;
  case 8:
     z -= 2.0;
     delta = 0.0;
     break;
  case 10:
  case 14:
     x -= 2.0;
     z -= 3.0;
     break;
  case 11:
     x -= 4.0;
     z -= 2.0;
     delta = 0.0;
     break;
  case 12:
  case 13:
     x -= 1.0;
     z -= 3.0;
     delta = 0.0;
     break;
  case 15:
     x -= 3.0;
     z -= 4.0;
     break;
  }
  if (delta == 1.0)
  {
    x = 2*fabs( x-0.5 );
    z = 2*( max(fabs( z),0.5) -0.5 );
  }
  else 
  {
     x = 2*( max(fabs( x),0.5) -0.5 );
     z = 2*fabs( z -0.5 );
  }
  answer = fabs(max(x,z)/2.0 - delta);
return answer;
}

static DBL car_ortho_carre  (VECTOR EPoint)         
{
/*
** Carrelage a base d'un carre et d'un octogone 
*/
DBL answer;
DBL x,z;
DBL dist1,dist2;
z=EPoint[Z];                                                                    
z -= (SQR2+1.0)*floor(z/(SQR2+1.0));
z -= SQR2_2+0.5;
z = fabs(z);
x=EPoint[X];
x -= (SQR2+1.0)*floor(x/(SQR2+1.0));
x -= SQR2_2+0.5;
x = fabs(x);
if (z > x)
{
   answer = x;
   x = z;
   z = answer;
}
if ( (x+z) < SQR2_2)
{
  /* Carrelage carre */
  return ( (x+z)/SQR2 );
}
dist1 = 0.5 + z  ;
if (dist1 > 1.0) 
{
   dist1 = 1.0;
}
dist2 = (z+x)/SQR2;

return min(dist1,dist2);
}
static DBL car_carre_triangle  (VECTOR EPoint)         
{
DBL x,z;
DBL slop1;
DBL dist1,dist2;
int delta;
x=EPoint[X];
z=EPoint[Z];
delta = 0;
 x -= floor(x);
 z -= (2.0+SQR3)*floor(z/(SQR3+2.0));
 /* x,z is in { [0.0, 1.0 [, [0.0, 2+SQR3 [ } 
 ** but there is some symetry to simplify the testing
 */
 if (z > SQR3_2+1.0 )
 {
    z -= SQR3_2+1.0;
    x += (x>0.5)?-0.5:0.5;
 } 
 if (x > 0.5)
 {
    x = 1.0 - x; /* mirror */
 }
 z -= 1.0;
 if (z > 0.0)
 { /* triangle */
 if (x != 0.0)
 {
    slop1 = z/x;
    if (slop1>SQR3)
    { 
      z = SQR3_2 - z;
      x = 0.5 -x;
      delta = 1 - delta;
    }
 }
 else
 {
   z = SQR3_2 -z;
   x = 0.5;
   delta = 1 - delta;
 }
 dist1 = 1.0 - (2 * z * SQR3 );
 dist2 = 1.0 - ((SQR3 * x - z) * SQR3 );
 return (delta+max(dist1,dist2))/3.0;
 }
 else
 { /* carre */
 if (z < -0.5)
 { 
    z = -1.0 -z;
 }
 if (x > 0.5)
 {
    x = 1.0 - x;
 }
 dist1 = 2 + 2 * SQR3 * fabs( x ) ;
 dist2 = 2 + 2 * SQR3 * fabs( z ) ;
 dist1 = min(dist1,3.0);
 dist2 = min(dist2,3.0);
 return min(dist1,dist2)/3.0;
 }
}
static DBL car_hexa_triangle  (VECTOR EPoint)         
{
/* 
** Carrelage avec un hexagone et deux triangles
*/
DBL x,z;
DBL dist1,dist2;
DBL answer=0;
int delta;
x=EPoint[X];
z=EPoint[Z];
delta = 0;
 x += 0.5;
 x -= 2.0*floor(x/2.0);
 z -= 2.0*SQR3*floor(z/(SQR3*2.0));
 /* x,z is in { [0.0, 2.0 [, [0.0, 2*SQR3 [ } 
 ** but there is some symetry to simplify the testing
 */
 if (z > SQR3 )
 {
    z -= SQR3; 
    x += (x<1.0)?1.0:-1.0;
 } 
 /* 
 ** Now only [0,2[,[0,SQR3[
 */
 if (z > SQR3_2 )
 {
    z = SQR3 - z; /* mirror */
    delta = 1 - delta;
 } 
 
 if (x > 1.0)
 {
    x = 2.0 - x; /* mirror */
 }
 /*
 ** And now, it is even simpler :  [0,1],[0,SQR3/2]
 ** on the bottom left corner, part of the triangle
 ** on the top right corner, center of the hexagone
 */
 if ((SQR3*x+z)<SQR3_2)
 {
    /* Triangle */
    dist1 = 1.0 - (z * 2 * SQR3 );
    dist2 = 1.0 + ((SQR3 * x + z) - SQR3_2 ) * SQR3; /* really substracting */
    answer = (delta + max(dist1,dist2))/3.0;
 }
 else
 {
   /* Hexagone */
    dist1 = 2 + 2* (z * SQR3 );
    dist2 = 2 + 2* ((SQR3 * x + z - SQR3_2) ) * SQR3_2 ;
    answer = min(dist1,dist2);
    answer /= 3.0;
    answer = min(answer,1.0);
 }
 return answer;
}
static DBL car_carre_decale  (VECTOR EPoint)         
{
/*
** Carrelage a carre, avec un decalage d'un demi cote par ligne 
** On se ramene dans le rectangle [0,1][0,2]
** On deplace ensuite x,[1,2] sur [0,1][0,1] , avec x' = x+1/2
** la valeur est le double de la distance(1) depuis 1/2,1/2
*/
DBL x,z;
z = EPoint[Z];
z -= 2.0*floor(z/2.0);
x = EPoint[X];
if (z > 1.0)
{
  x += 0.5;
  z -= 1.0;
}
x -= floor(x);
x = 2*fabs( x-0.5 );
z = 2*fabs( z-0.5 );

return max(x,z);
}

static DBL car_carre_rectangle  (VECTOR EPoint)         
{
/*
** Carrelage avec un carre central et quatre rectangles (2x1)
** qui tournent autour du carre
** On se ramene dans le carre [0,3][0,3]
** puis on divise ce carre en 9 cas
**
**  +-----+--+
**  |     |  |
**  +--+--+  |
**  |  |  |  |
**  |  +--+--+
**  |  |     |
**  +--+-----+
*/
DBL x,z;
DBL delta;
int valueX,valueZ;
  x = EPoint[X];
  x -= 3.0*floor(x/3.0);
  z = EPoint[Z];
  z -= 3.0*floor(z/3.0);
  valueX = (int)x;
  valueZ = (int)z;
  delta = 3.0;
  switch((valueX+valueZ*3))
  {
  case 0:
  case 3:
    z -= 1.0;
    break;
  case 1:
  case 2:
    x -= 2.0;
    delta = 1.0;
    break;
  case 4:
    x -= 1.0;
    z -= 1.0;
   delta = 0.0;
    break;
  case 5:
  case 8:
    x -= 2.0;
    z -= 2.0;
    break;
  case 6:
  case 7:
    x -= 1.0;
    z -= 2.0;
    delta = 1.0;
    break;
  }
  if (delta == 1.0)
  {
     x = fabs( x);
     x = 2.0*( max( x,0.5) -0.5 );
     z = 2.0*fabs( z -0.5 );
  }
  if (delta == 3.0)
  {
     x = 2.0*fabs( x-0.5 );
     z = fabs(z);
     z = 2.0*( max(z,0.5) - 0.5 );
     return (delta-(max(x,z)))/3.0 ;
  }
  if (delta == 0.0)
  {
     x = 2.0*fabs( x-0.5 );
     z = 2.0*fabs( z-0.5 );
  }
  return ((max(x,z))+delta)/3.0 ;
}
static DBL car_rectangle_carre  (VECTOR EPoint)         
{
/*
** Carrelage avec un carre central et quatre rectangles (2x1)
** qui tournent autour du carre, dans les deux sens
** On se ramene dans le carre [0,6][0,6], on plie en quatre
** puis on divise ce carre en 9 cas
**
**  +-----+--+
**  |     |  |
**  +--+--+  |
**  |  |  |  |
**  |  +--+--+
**  |  |     |
**  +--+--+--+
*/
DBL x,z;
DBL delta;
int valueX,valueZ;
  x = EPoint[X];
  x -= 6.0*floor(x/6.0);
  x -= 3.0;
  x = fabs(x);
  z = EPoint[Z];
  z -= 6.0*floor(z/6.0);
  z -= 3.0;
  z = fabs(z);
  valueX = (int)x;
  valueZ = (int)z;
  delta = 3.0;
  switch((valueX+valueZ*3))
  {
  case 0:
  case 3:
    z -= 1.0;
    break;
  case 1:
  case 2:
    x -= 2.0;
    delta = 1.0;
    break;
  case 4:
    x -= 1.0;
    z -= 1.0;
   delta = 0.0;
    break;
  case 5:
  case 8:
    x -= 2.0;
    z -= 2.0;
    break;
  case 6:
  case 7:
    x -= 1.0;
    z -= 2.0;
    delta = 1.0;
    break;
  }
  if (delta == 1.0)
  {
     x = fabs( x);
     x = 2.0*( max(x,0.5) -0.5 );
     z = 2.0*fabs( z -0.5 );
  }
  if (delta == 3.0)
  {
     x = 2.0*fabs( x-0.5 );
     z = fabs(z);
     z = 2.0*( max(z,0.5) - 0.5 );
     return (delta-(max(x,z)))/3.0 ;
  }
  if (delta == 0.0)
  {
     x = 2.0*fabs( x-0.5 );
     z = 2.0*fabs( z-0.5 );
  }
  return ((max(x,z))+delta)/3.0 ;
}
static DBL car_carre_interne  (VECTOR EPoint)         
{
DBL answer=0;
DBL x,z;
DBL dist1,dist2;
int valueX,valueZ;
x=EPoint[X];
x *= SQR2;
x -= 4.0 * floor(x/4.0);
x -= 2.0;
x = fabs(x);
z=EPoint[Z];                                                                    
z *= SQR2;
z -= 4.0 * floor(z/4.0);
z -= 2.0;
z = fabs(z);
valueX=(int)x;
valueZ=(int)z;
switch((valueX+valueZ*2))
{
case 0:
  x -= 0.5;
  x = max(x,0);
  x *= 2.0; 
  z -= 0.5;
  z = max(z,0);
  z *= 2.0;
  answer=max(x,z)/3.0;
  break;
case 1: 
  answer=(3.0-fabs(1.5-x)*2.0)/3.0;
  if (z>0.5)
  {
    dist2=(2.0+SQR2*fabs(x-z))/3.0;
    answer = min(answer,dist2);
  }
  break;
case 2:
  answer=(1.0+fabs(1.5-z)*2.0)/3.0;
  if (x>0.5)
  {
    dist2=(2.0-SQR2*fabs(x-z))/3.0;
    answer = max(answer,dist2);
  }
  break;
case 3:
  if (x > z)
  {
  dist1=(3.0-fabs(1.5-x)*2.0)/3.0;
  dist2=(2.0+SQR2*fabs(z-x))/3.0;
  answer=min(dist1,dist2);
  }
  else
  {
  dist1=(1.0+fabs(1.5-z)*2.0)/3.0;
  dist2=(2.0-SQR2*fabs(x-z))/3.0;
  answer=max(dist1,dist2);
  }
  break;
}
return answer;
}
static DBL car_carre_interne_5  (VECTOR EPoint)         
{
DBL answer=0;
DBL x,z;
DBL dist1,dist2;
int mirX,mirZ;
int valueX,valueZ;
mirX=mirZ=0;
x=EPoint[X];
x *= SQR2;
x -= 4.0 * floor(x/4.0);
x -= 2.0;
mirX = (x < 0)?1:0;
x = fabs(x);
z=EPoint[Z];                                                                    
z *= SQR2;
z -= 4.0 * floor(z/4.0);
z -= 2.0;
mirZ = (z < 0)?2:3;
z = fabs(z);
valueX=(int)x;
valueZ=(int)z;
switch((valueX+valueZ*2))
{
case 0:
  x -= 0.5;
  x = max(x,0);
  x *= 2.0; 
  z -= 0.5;
  z = max(z,0);
  z *= 2.0;
  answer=(5.0 - max(x,z))/5.0;
  break;
case 1: 
  answer=fabs(1.5-x)*2.0;
  if (z>0.5)
  {
    dist2=1.0-SQR2*fabs(x-z);
    answer = max(answer,dist2);
  }
  answer += mirX;
  answer /= 5.0;
  break;
case 2:
  answer=fabs(1.5-z)*2.0;
  if (x>0.5)
  {
    dist2=1.0-SQR2*fabs(x-z);
    answer = max(answer,dist2);
  }
  answer += mirZ;
  answer /= 5.0;
  break;
case 3:
  if (x > z)
  {
  dist1=fabs(1.5-x)*2.0;
  dist2=1.0-SQR2*fabs(z-x);
  answer=max(dist1,dist2) + mirX;
  }
  else
  {
  dist1=fabs(1.5-z)*2.0;
  dist2=1.0-SQR2*fabs(x-z);
  answer=max(dist1,dist2) + mirZ;
  }
  answer /= 5.0;
  break;
}
return answer;
}
static DBL car_carre_double  (VECTOR EPoint)                         
{
/*
** Carrelage avec un carre (1x1) et un carre (2x2)
** qui tournent autour du carre (1x1) (et reciproquement)
** On se ramene dans le carre [0,5][0,5]
** puis on divise ce carre en 25 cas
**
**  +--+     +--+--+
**     |     |  |  
**     +--+--+--+   
**     |  |     |  
**  +--+--+     +--+
**  |     |     |  |
**  +     +--+--+--+
**  |     |  |     |
**  +--+--+--+     +
**  |  |     |     |
**  +--+     +--+--+
*/
DBL x,z;
DBL delta;
int valueX,valueZ;
  x = EPoint[X];
  x -= 5.0*floor(x/5.0);
  z = EPoint[Z];
  z -= 5.0*floor(z/5.0);
  valueX = (int)x;
  valueZ = (int)z;
  delta = 1.0;
  switch((valueX+valueZ*5))
  {
     case 0:
       delta = 0.0;
       break;
     case 1:
     case 2:
       x -= 2.0;
       break;
     case 3:
     case 4:
     case 8:
     case 9:
       x -= 4.0;
       z -= 1.0;
       break;
     case 5 :
     case 6 :
     case 10:
     case 11:
       z -= 2.0;
       x -= 1.0;
       break;
     case 7:
       delta = 0.0;
       x -= 2.0;
       z -= 1.0;
       break;
     case 12:
     case 13:
     case 17:
     case 18:
       x -= 3.0;
       z -= 3.0;
       break;
     case 14:
       x -= 4.0;
       z -= 2.0;
       delta = 0.0;
       break;
     case 15:
     case 20:
       z -= 4.0;
       break;
     case 16:
       x -= 1.0;
       z -= 3.0;
       delta = 0.0;
       break;
     case 21:
     case 22:
       x -= 2.0;
       z -= 5.0;
       break;
     case 23:
       x -= 3.0;
       z -= 4.0;
       delta = 0.0;
       break;
     case 24:
     case 19:
       x -= 5.0;
       z -= 4.0;
       break;
  }
  if (delta)
  {
     x = fabs(x);
     x = 2*( max(x,0.5) -0.5 );
     z = fabs(z);
     z = 2*( max(z,0.5) -0.5 );
  }
  else
  {
    x = 2*fabs( x-0.5 );
    z = 2*fabs( z-0.5 );
  }
  return fabs((max(x,z))/2.0-delta);
}

static DBL car_hexa_carre_triangle  (VECTOR EPoint)
{
/* 
** Carrelage avec un hexagone, carres et triangles
*/
DBL x,z;
DBL dist1,dist2;
DBL answer=0;

x=EPoint[X];
z=EPoint[Z];
 x += 0.5;
 x -= (3.0+SQR3)*floor(x/(3.0+SQR3));
 z -= 2.0*(1.0+SQR3)*floor(z/((SQR3+1.0)*2.0));
 /* x,z is in { [0.0, 3.0+SQR3 [, [0.0, 2*(1+SQR3) [ } 
 ** but there is some symetry to simplify the testing
 */
 if (x > (SQR3_2+1.5))
 {
    x -= (SQR3_2+1.5);
    z += (z < SQR3+1.0)?(SQR3_2+0.5):(-1*(SQR3_2+0.5));
 }
 if (z > (SQR3+1.0) )
 {
    z -= SQR3+1.0; 
 } 
 /* 
 ** Now only [0, SQR3/2+1.5 ], [0,SQR3+1.0]
 */
 if (z > (0.5+SQR3_2 ) )
 {
    z = SQR3+1.0 - z; /* mirror */
 } 
 
 if (x > (1.0+SQR3_2) )
 {
    x = SQR3+2.0 - x; /* mirror */
 }
 /*
 ** And now, it is even simpler :  [0,1+(SQR3/2)],[0,1+(SQR3/2)]
 ** on the top right corner, center of the hexagone
 ** on the bottom right, middle of square, a triangle on the left
 ** on the top left corner, half a triangle
 ** bottom
 */
 if (((SQR3*x+z)<(SQR3_2+2.0))&&(x < 0.5+SQR3_2))
 {
    /* Triangle or square */
    /* rotate in the lower part */
     z -= 0.5 + SQR3_2 ;
     x -= 1.0 + SQR3_2 ;
     dist1 = ( x / 2.0 ) - ( z * SQR3_2 ); 
     dist2 = ( z / 2.0 ) + ( x * SQR3_2 ); 
     z = dist2;
     x = dist1;
     z += 0.5 + SQR3_2 ;
     x += 1.0 + SQR3_2 ;
    if (z < 0)
    {
       z *= -1;
    }
 if (x > (1.0+SQR3_2) )
 {
    x = SQR3+2.0 - x; /* mirror */
 }
 }
 if ((!((SQR3*x+z)<(SQR3_2+2.0)))&&(!(z < 0.5)))
 {
   /* Hexagone */
    z -= 0.5;
    x -= SQR3_2;
    z= fabs(z);
    x= fabs(x);
    dist1 = 2* z * SQR3 ;
    dist2 = ((SQR3 * x + z ) ) * SQR3 - 1.5;
    answer = min(dist1,dist2);
    answer += 2.0;
    answer = min(answer,3.0);
    answer /= 3.0;
 return answer;
 }
 if (z < 0.5 )
 {
  if ( x > SQR3_2+0.5)
 {
   /* carre */
   x -= SQR3_2+0.5;
   z -= 0.5;
   dist1 = 1.0 + z*2 * SQR3;
   dist2 = 1.0 - x*2*SQR3;
   dist1 = max(dist1,0);
   dist2 = max(dist2,0);
   answer = max(dist1,dist2)/3.0;
 }
 else
 {
   /* triangle */
  x -= SQR3_2 + 0.5;
  x = fabs(x);
  dist2 = ((SQR3 * z + x) ) * SQR3 +0.5; 
  dist1 = 2.0 - (x * 2.0 * SQR3 );
  answer = (max(dist1,dist2))/3.0;
 }
 }
 return answer;
}
static DBL car_hexa_carre_triangle_6  (VECTOR EPoint)
{
/* 
** Carrelage avec un hexagone, carres et triangles
*/
DBL x,z;
DBL dist1,dist2;
DBL answer=0;
int mirX=0;
int mirZ=0;
int rota=0;
x=EPoint[X];
z=EPoint[Z];
 x += 0.5;
 x -= (3.0+SQR3)*floor(x/(3.0+SQR3));
 z -= 2.0*(1.0+SQR3)*floor(z/((SQR3+1.0)*2.0));
 /* x,z is in { [0.0, 3.0+SQR3 [, [0.0, 2*(1+SQR3) [ } 
 ** but there is some symetry to simplify the testing
 */
 if (x > (SQR3_2+1.5))
 {
    x -= (SQR3_2+1.5);
    z += (z < SQR3+1.0)?(SQR3_2+0.5):(-1*(SQR3_2+0.5));
 }
 if (z > (SQR3+1.0) )
 {
    z -= SQR3+1.0; 
 } 
 /* 
 ** Now only [0, SQR3/2+1.5 ], [0,SQR3+1.0]
 */
 if (z > (0.5+SQR3_2 ) )
 {
    z = SQR3+1.0 - z; /* mirror */
    mirZ = 1 - mirZ;
 } 
 
 if (x > (1.0+SQR3_2) )
 {
    x = SQR3+2.0 - x; /* mirror */
    mirX = 1 - mirX;
 }
 /*
 ** And now, it is even simpler :  [0,1+(SQR3/2)],[0,1+(SQR3/2)]
 ** on the top right corner, center of the hexagone
 ** on the bottom right, middle of square, a triangle on the left
 ** on the top left corner, half a triangle
 ** bottom
 */
 if (((SQR3*x+z)<(SQR3_2+2.0))&&(x < 0.5+SQR3_2))
 {
    /* Triangle or square */
    /* rotate in the lower part */
     z -= 0.5 + SQR3_2 ;
     x -= 1.0 + SQR3_2 ;
     dist1 = ( x / 2.0 ) - ( z * SQR3_2 ); 
     dist2 = ( z / 2.0 ) + ( x * SQR3_2 ); 
     z = dist2;
     x = dist1;
     z += 0.5 + SQR3_2 ;
     x += 1.0 + SQR3_2 ;
     rota = 1 - rota;
    if (z < 0)
    {
       z *= -1;
    }
   if (x > (1.0+SQR3_2) )
   {
      x = SQR3+2.0 - x; /* mirror */
      mirX = 1 - mirX;
      mirZ = 1 - mirZ;
   }
 }
 if ((!((SQR3*x+z)<(SQR3_2+2.0)))&&(!(z < 0.5)))
 {
   /* Hexagone */
    z -= 0.5;
    x -= SQR3_2;
    z= fabs(z);
    x= fabs(x);
    dist1 = 2* z * SQR3 ;
    dist2 = ((SQR3 * x + z ) ) * SQR3 - 1.5;
    answer = 5.0 + min(dist1,dist2);
    answer = min(answer,6.0);
    answer /= 6.0;
 return answer;
 }
 if (z < 0.5 )
 {
  if ( x > SQR3_2+0.5)
 {
   /* carre */
   x -= SQR3_2+0.5;
   z -= 0.5;
   dist1 = 1.0 + z*2 * SQR3;
   dist2 = 1.0 - x*2*SQR3;
   dist1 = max(dist1,0.0);
   dist2 = max(dist2,0.0);
   answer = (max(dist1,dist2) +  rota * (1.000001 + ((mirX + mirZ) % 2)))/6.0;
 }
 else
 {
   /* triangle */
  x -= SQR3_2 + 0.5;
  x = fabs(x);
  dist2 = ((SQR3 * z + x) ) * SQR3 +0.5; 
  dist1 = 2.0 - (x * 2.0 * SQR3 );
  answer = max(dist1,dist2);
  if ((rota + mirX)%2) 
  {
     answer = 2.0 + answer;
  }
  else
  {
     answer = 3.0 + answer;
  }
  answer /= 6.0;
 }
 }
 return answer;
}
static DBL car_rectangle_paire  (VECTOR EPoint)
{

/*
** Carrelage avec deux rectangles (2x1)
** On se ramene dans le carre [0,4][0,4]
** puis on divise ce carre en 16 cas
**
**  +-----+--+--+
**  |     |  |  |
**  +-----+  +  +
**  |     |  |  |
**  +-----+--+--+
**  |  |  |     |
**  |  |  +-----+
**  |  |  |     |
**  +--+--+-----+
*/
DBL x,z;
DBL delta;
DBL answer;
int valueX,valueZ;
x=EPoint[X];
z=EPoint[Z];                                                                    
x -= 4.0*floor(x/4.0);
z -= 4.0*floor(z/4.0);
  valueX = (int)x;
  valueZ = (int)z;
  delta = 1.0;
  switch((valueX+valueZ*4))                                                       
  {
  case 0:
  case 4:
     z -= 1.0;
     break;
  case 1:
  case 5:
     x -= 1.0;
     z -= 1.0;
     break;
  case 2:
  case 3:
     x -= 3.0;
     delta = 0.0;
     break;
  case 6:
  case 7:
     x -= 3.0;
     z -= 1.0;
     delta = 0.0;
     break;
  case 8:
  case 9:
     x -= 1.0;
     z -= 2.0;
     delta = 0.0;
     break;
  case 12:
  case 13:
     x -= 1.0;
     z -= 3.0;
     delta = 0.0;
     break;
  case 10:
  case 14:
     x -= 2.0;
     z -= 3.0;
     break;
  case 11:
  case 15:
     x -= 3.0;
     z -= 3.0;
     break;
  }
  if (delta == 1.0)
  {
    x = 2*fabs( x-0.5 );
    z = 2*( max(fabs( z),0.5) -0.5 );
  }
  else 
  {
     x = 2*( max(fabs( x),0.5) -0.5 );
     z = 2*fabs( z -0.5 );
  }
  answer = fabs(max(x,z)/2.0 - delta);
return answer;
}

static DBL car_hexa_tri_droite  (VECTOR EPoint)         
{
DBL x,z;
DBL answer;
DBL slop1;
DBL dist1,dist2;
int zzof;
int delta;
x=EPoint[X];
z=EPoint[Z];
 /* First, resume to a simple pattern */
 zzof = z / SQR3_2;
 zzof /= 3;
 if (z < 0)
 {
   zzof -= 1;
 }
 x += zzof / 2.0; /* right handed */
 z -= 3*SQR3_2*floor(z/(3*SQR3_2));
 x += 7.0;
 x -= 7.0 *floor(x/7.0);
 if ((x > 4.5) && (z > SQR3))
 {
    x -= 4.5;
    z -= SQR3_2;
 }
 if ((x > 5.0) && (z < SQR3_2))
 {
    x -= 5;
    z += SQR3;
 }
 if ((x > 2.5) && (z < SQR3))
 {
    x -= 2.5;
    z += SQR3_2;
 }
 delta = 0;
 zzof = z /SQR3_2;
 if ( zzof == 2) 
 {
   zzof = 1;
   z = 2 * SQR3 - z;
   delta = 1 - delta;
 }
if ((!zzof) || (x > 2.0) ||
 (( z + SQR3*x < SQR3) || ( SQR3*x -z > SQR3)) 
 )
{
 /* triangle */
 x -= 1.0 *floor(x/1.0);
 z -= SQR3*floor(z/SQR3);
 /* x,z is in { [0.0, 1.0 [, [0.0, SQR3 [ } 
 ** but there is some symetry to simplify the testing
 */
 if (z > SQR3_2 )
 {
    z = SQR3 - z; /* mirror */
    delta = 1 - delta;
 } 
 if (x > 0.5)
 {
    x = 1.0 - x; /* mirror */
 }
 if (x != 0.0)
 {
    slop1 = z/x;
    if (slop1>SQR3)
    { 
      z = SQR3_2 - z;
      x = 0.5 -x;
      delta = 1 - delta;
    }
 }
 else
 {
   z = SQR3_2 -z;
   x = 0.5;
   delta = 1 - delta;
 }
 dist1 = 1.0 - (z * 2 * SQR3 );
 dist2 = 1.0 - ((SQR3 * x - z) * SQR3 );
 return (delta+max(dist1,dist2))/3.0;
}
 else
 { /* hexagone */
 z -= SQR3_2;
 if (x > 1.0)
 {
    x = 2.0 - x; /* mirror */
 }
   /* Hexagone */
    dist1 = 2 + 2* (z * SQR3 );
    dist2 = 2 + 2* ((SQR3 * x + z - SQR3_2) ) * SQR3_2 ;
    answer = min(dist1,dist2);
    answer /= 3.0;
    answer = min(answer,1.0);
return answer;
 }
}
static DBL car_hexa_tri_gauche  (VECTOR EPoint)         
{
DBL x,z;
DBL slop1;
DBL dist1,dist2;
DBL answer;
int zzof;
int delta;
x=EPoint[X];
z=EPoint[Z];
 /* First, resume to a simple pattern */
 zzof = z / SQR3_2;
 zzof /= 3;
 if (z < 0)
 {
   zzof -= 1;
 }
 x -= zzof / 2.0; /* left handed */
 z -= 3*SQR3_2*floor(z/(3*SQR3_2));
 x += 7.0;
 x -= 7.0 *floor(x/7.0);
 if ((x > 2.0) && (z < SQR3_2))
 {
    x -= 2.0;
    z += SQR3;
 }
 if ((x > 4.5) && (z < SQR3))
 {
    x -= 4.5;
    z += SQR3_2;
 }
 if ((x > 2.5) && (z > SQR3))
 {
    x -= 2.5;
    z -= SQR3_2;
 }
 delta = 0;
 zzof = z /SQR3_2;
 if ( zzof == 2) 
 {
   zzof = 1;
   z = 2 * SQR3 - z;
   delta = 1 - delta;
 }
if ((!zzof) || (x > 2.0) ||
 (( z + SQR3*x < SQR3) || ( SQR3*x -z > SQR3)) 
 )
{
 /* triangle */
 x -= 1.0 *floor(x/1.0);
 z -= SQR3*floor(z/SQR3);
 /* x,z is in { [0.0, 1.0 [, [0.0, SQR3 [ } 
 ** but there is some symetry to simplify the testing
 */
 if (z > SQR3_2 )
 {
    z = SQR3 - z; /* mirror */
    delta = 1 - delta;
 } 
 if (x > 0.5)
 {
    x = 1.0 - x; /* mirror */
 }
 if (x != 0.0)
 {
    slop1 = z/x;
    if (slop1>SQR3)
    { 
      z = SQR3_2 - z;
      x = 0.5 -x;
      delta = 1 - delta;
    }
 }
 else
 {
   z = SQR3_2 -z;
   x = 0.5;
   delta = 1 - delta;
 }
 dist1 = 1.0 - (z * 2 * SQR3 );
 dist2 = 1.0 - ((SQR3 * x - z) * SQR3 );
 return (delta+max(dist1,dist2))/3.0;
}
 else
 { /* hexagone */
 z -= SQR3_2;
 if (x > 1.0)
 {
    x = 2.0 - x; /* mirror */
 }
   /* Hexagone */
    dist1 = 2 + 2* (z * SQR3 );
    dist2 = 2 + 2* ((SQR3 * x + z - SQR3_2) ) * SQR3_2 ;
    answer = min(dist1,dist2);
    answer /= 3.0;
    answer = min(answer,1.0);
return answer;
 }
}

static DBL car_carre_tri  (VECTOR EPoint)         
{
DBL x,z;
DBL slop1;
DBL dist1,dist2;
int delta;
int gamma,beta;
int xflop,zflop;
x=EPoint[X];
z=EPoint[Z];
 delta = 0;
 gamma = 0;
 beta = 0;
 x -= (1.0+SQR3)*floor(x/(1.0+SQR3));
 z -= (1.0+SQR3)*floor(z/(1.0+SQR3));
 /* x,z is in { [0.0, SQR3+1.0 [, [0.0, SQR3+1.0 [ } 
 ** but there is some symetry to simplify the testing
 */
 if (z > 0.5 + SQR3_2 )
 {
    z = 1.0 + SQR3 - z; /* mirror */
    delta = 1 - delta;
 } 
 if (x > 0.5 + SQR3_2)
 {
    x = 1.0 + SQR3 - x; /* mirror */
    delta = 1 - delta;
    beta = 1;
 }
 /* x,z is in { [0.0, (SQR3+1)/2 ], [0.0, (SQR3+1)/2 ] }
 ** but there is still a symetry and a rotation 
 */
 xflop = ( 2.0 * x + (SQR3 -1.0)* 2.0 * z / ( SQR3 + 1.0) > (SQR3)) ? 1: 0;
 zflop = ( 2.0 * z * SQR3 + (SQR3 -3.0)* 2.0 * x / ( SQR3 + 1.0) > (SQR3)) ? 1: 0;
 switch (xflop + 2*zflop)
 {
  case 0: /* do nothing */
    gamma = 2 * beta;
    break;
  case 1: /* rotate clockwise */
    gamma = beta ? 1 + 2 * delta : 3 - 2 *delta;
    slop1 = x;
    x = z;
    z = SQR3_2 + 0.5 - slop1;
    break;
  case 2: /* rotate normal */
    gamma = beta ? 3 - 2 * delta : 1 + 2 *delta;
    slop1 = z;
    z = x;
    x = SQR3_2 +0.5 - slop1;
    break;
  case 3: /* symetry */
    gamma = beta ? 0: 2 ;
    x = SQR3_2+0.5 -x;
    z = SQR3_2+0.5 -z;
    break;
 }
 
 if (x == 0.0)
 {
   z = 0.0;
   x = 0.0;
 }
 slop1 = (z * 2 * SQR3) - SQR3 + x * 2;
 if (slop1 < 0)
 {
   /* triangle */
   
   dist1 = -1.5 / SQR3 * slop1;
   dist2 = 2.0 * x * SQR3;
   slop1 = min(dist1,dist2);
   return (gamma+1.0-slop1)/6.0;
 }
 else 
 {
   /* square */
   slop1 *= 1.5 / SQR3;
   slop1 = min(slop1,1);
   if (delta)
   {
      slop1 += 5.0;
      slop1 /= 6.0;
      slop1 = max(slop1,5.0/6.0);
      slop1 = min(slop1,1.0);
   }
   else
   {
      slop1 *= -1;
      slop1 += 5.0;
      slop1 /= 6.0;
      slop1 = min(slop1,5.0/6.0);
      slop1 = max(slop1,4.000001/6.0); /* tricky, representation of 4/6 is not exact */
   }
   return slop1;
 }
}

static DBL car_dodeca_tri  (VECTOR EPoint)         
{
DBL x,z;
DBL dist1,dist2,dist3,dist4,ret_value;
DBL tmpx,tmpz;
int toggle=0; /* switched each time a triangle get toggled */
x=EPoint[X];
z=EPoint[Z];
 x -= (2.0+SQR3)*floor(x/(2.0+SQR3));
 z -= (3.0+2.0*SQR3)*floor(z/(3.0+2.0*SQR3));
 /* x,z is in { [0.0, SQR3+2.0 [, [0.0, 2*SQR3+3.0 [ } 
 ** but there is some symetry to simplify the testing
 */
 if (z > SQR3+1.5)
 {
   /* translate */
   z -= SQR3+1.5;
   x += (x<(1.0+SQR3_2) ? 1.0: -1.0)*(1.0+SQR3_2);
 }
 if (x > 1.0+SQR3_2)
 {
   x = 2.0 + SQR3 -x;
 }
 if (z > 1.0+SQR3_2)
 {
   z = 2.0 + SQR3 -z;
   toggle = 1 - toggle;
 }
 dist2 = x - SQR3_2 - 0.5 + z * SQR3;
 if (dist2 < 0.0)
 {
   tmpx = x;
   tmpz = z;
   x = (1.0+SQR3)/4.0 + 0.5 * tmpx - SQR3_2 * tmpz ;
   z = (3.0+SQR3)/4.0 - SQR3_2 * tmpx - 0.5 * tmpz ;
   dist2 *= -1.0;
 }
 dist1 = (z * 3.0 ); /* from the bottom line */
 dist3 =  z - SQR3_2 - 0.5 + x * SQR3; 
 dist3 *= SQR3;
 dist2 *= SQR3;
 dist4 = (x * 3.0 ); /* from the vertical line */
 if (dist3 < 0.0)
 {
    ret_value = max(dist3,-1.0);
    ret_value += 1+toggle;
    ret_value /= 3.0;
 }
 else 
 {
    /* dodecagone */
    ret_value = min(dist1,dist2);
    ret_value = min(ret_value,dist3);
    ret_value = min(ret_value,dist4);
    ret_value = min(ret_value,1.0);
    ret_value = 2.0 + ret_value;
    ret_value /= 3.0;
 }
 return ret_value;
}
static DBL car_dodeca_hex  (VECTOR EPoint)         
{
DBL x,z;
DBL dist1,dist2,dist3,dist4,ret_value;
DBL dist5,dist6,dist7;
DBL tmpx,tmpz;
x=EPoint[X];
z=EPoint[Z];
 x -= (3.0+SQR3)*floor(x/(3.0+SQR3));
 z -= (3.0+3.0*SQR3)*floor(z/(3.0+3.0*SQR3));
 /* x,z is in { [0.0, SQR3+3.0 [, [0.0, 3*SQR3+3.0 [ } 
 ** but there is some symetry to simplify the testing
 */
 if (z > 1.5*SQR3+1.5)
 {
   /* translate */
   z -= 1.5*SQR3+1.5;
   x += (x<(1.5+SQR3_2) ? 1.0: -1.0)*(1.5+SQR3_2);
 }
 if (x > 1.5+SQR3_2)
 {
   x = 3.0 + SQR3 -x;
 }
 if (z > 1.0+SQR3)
 {
   z = 2.0 + 2.0*SQR3 -z;
 }
 dist2 = x - SQR3_2 - 1.5 + z * SQR3;
 if (dist2 < 0.0)
 {
   tmpx = x;
   tmpz = z;
   x = (3.0+SQR3)/4.0 + 0.5 * tmpx - SQR3_2 * tmpz ;
   z = (3.0+3.0*SQR3)/4.0 - SQR3_2 * tmpx - 0.5 * tmpz ;
 }
 dist2 = x - SQR3_2 - 2.5 + z * SQR3;
 dist1 = (z * 2.0 ) - SQR3; /* from the bottom line */
 dist3 = z - 1.5*SQR3 - 0.5 + x * SQR3; 
 dist4 = ((x-0.5) * 2.0 ); /* from the vertical line */
 if ((dist2 >= 0.0)
   &&(dist1 >= 0.0)
   &&(dist3 >= 0.0)
   &&(dist4 >= 0.0)
    )
 {
    /* dodecagone */
    ret_value = min(dist1,dist2);
    ret_value = min(ret_value,dist3);
    ret_value = min(ret_value,dist4);
    ret_value = min(ret_value,1.0);
    ret_value = 2.0 + ret_value;
 }
 else
 {
    dist5 = 2*z - 2*SQR3 - 1.0;
    if (dist5 >= 0)
    {
       dist4 *= -1.0;
       ret_value = min(dist5,dist4);
       ret_value = 2.0 - ret_value;
    }
    else
    {
       dist6 = SQR3 * x - z -SQR3_2 +0.5;
       dist7 = dist6 - 2.0;
       switch((dist6 >= 0?0:1)+(dist7 >= 0.0 ?2:0))
       {
          case 1: /* left hexagone */
             dist5 *= -1.0;
             dist6 *= -1.0;
             dist3 *= -1.0;
             ret_value = min(dist6,dist3);
             ret_value = min(ret_value,dist5);
             ret_value = min(ret_value,1.0);
             ret_value = 1.0 - ret_value;
             break;
          case 2: /* bottom hexagone */
             dist1 *= -1.0;
             ret_value = min(dist7,dist1);
             ret_value = min(ret_value,1.0);
             ret_value = 1.0 - ret_value;
             break;
          case 0:
          default: /* slanted square */
             dist2 *= -1.0;
             dist7 *= -1.0;
             ret_value = min(dist6,dist2);
             ret_value = min(ret_value,dist7);
             ret_value = 2.0 - ret_value;
             break;
       }
    }
 }
 ret_value /= 3.0;
 return ret_value;
}
static DBL car_dodeca_hex_5  (VECTOR EPoint)         
{
DBL x,z;
DBL dist1,dist2,dist3,dist4,ret_value;
DBL dist5,dist6,dist7;
DBL tmpx,tmpz;
int mirX,mirZ,rota;
mirX = mirZ = rota = 0;
x=EPoint[X];
z=EPoint[Z];
 x -= (3.0+SQR3)*floor(x/(3.0+SQR3));
 z -= (3.0+3.0*SQR3)*floor(z/(3.0+3.0*SQR3));
 /* x,z is in { [0.0, SQR3+3.0 [, [0.0, 3*SQR3+3.0 [ } 
 ** but there is some symetry to simplify the testing
 */
 if (z > 1.5*SQR3+1.5)
 {
   /* translate */
   z -= 1.5*SQR3+1.5;
   x += (x<(1.5+SQR3_2) ? 1.0: -1.0)*(1.5+SQR3_2);
 }
 if (x > 1.5+SQR3_2)
 {
   x = 3.0 + SQR3 -x;
   mirX = 1 - mirX;
 }
 if (z > 1.0+SQR3)
 {
   z = 2.0 + 2.0*SQR3 -z;
   mirZ = 1 - mirZ;
 }
 dist2 = x - SQR3_2 - 1.5 + z * SQR3;
 if (dist2 < 0.0)
 {
   tmpx = x;
   tmpz = z;
   x = (3.0+SQR3)/4.0 + 0.5 * tmpx - SQR3_2 * tmpz ;
   z = (3.0+3.0*SQR3)/4.0 - SQR3_2 * tmpx - 0.5 * tmpz ;
 }
 dist2 = x - SQR3_2 - 2.5 + z * SQR3;
 dist1 = (z * 2.0 ) - SQR3; /* from the bottom line */
 dist3 = z - 1.5*SQR3 - 0.5 + x * SQR3; 
 dist4 = ((x-0.5) * 2.0 ); /* from the vertical line */
 if ((dist2 >= 0.0)
   &&(dist1 >= 0.0)
   &&(dist3 >= 0.0)
   &&(dist4 >= 0.0)
    )
 {
    /* dodecagone */
    ret_value = min(dist1,dist2);
    ret_value = min(ret_value,dist3);
    ret_value = min(ret_value,dist4);
    ret_value = min(ret_value,1.0);
    ret_value = 4.0 + ret_value;
 }
 else
 {
    dist5 = 2*z - 2*SQR3 - 1.0;
    if (dist5 >= 0)
    {
       dist4 *= -1.0;
       ret_value = min(dist5,dist4);
       ret_value = 2.0 - ret_value;
    }
    else
    {
       dist6 = SQR3 * x - z -SQR3_2 +0.5;
       dist7 = dist6 - 2.0;
       switch((dist6 >= 0?0:1)+(dist7 >= 0.0 ?2:0))
       {
          case 1: /* left hexagone */
             dist5 *= -1.0;
             dist6 *= -1.0;
             dist3 *= -1.0;
             ret_value = min(dist6,dist3);
             ret_value = min(ret_value,dist5);
             ret_value = min(ret_value,1.0);
             ret_value = 1.0 - ret_value;
             break;
          case 2: /* bottom hexagone */
             dist1 *= -1.0;
             ret_value = min(dist7,dist1);
             ret_value = min(ret_value,1.0);
             ret_value = 1.0 - ret_value;
             break;
          case 0:
          default: /* slanted square */
             dist2 *= -1.0;
             dist7 *= -1.0;
             ret_value = min(dist6,dist2);
             ret_value = min(ret_value,dist7);
             ret_value = (mirZ + mirX) %2 + 3.0 - ret_value;
             break;
       }
    }
 }
 ret_value /= 5.0;
 return ret_value;
}
static unsigned char digone[]={ 0x0E, 0x0B };
static unsigned char trigone[][6]={
{ 0x0D, 0x05, 0x07 , 0x0D, 5,7 },
{ 0x0E, 0x0D, 0x16, 0x49, 7, 0x0B }
};
static unsigned char tetragone[][16]={
{ 0x0D, 5, 5, 7, 0x0D, 5, 5, 7 ,
 0x0D, 5, 5, 7, 0x0D, 5, 5, 7 },
{ 0x0E, 0x0D, 5, 0x16, 0x49, 5, 7, 0x0B ,
 0x0E, 0x0D, 5, 0x16, 0x49, 5, 7, 0x0B },
{ 0x0D, 0x34, 0x7, 0x0B, 0x07, 0x0B, 0x0D, 0x34,
 0x0D, 0x34, 0x7, 0x0B, 0x07, 0x0B, 0x0D, 0x34},
{ 0x0C, 0x06, 0x0C, 0x06, 0x09, 0x03, 0x09, 0x03 ,
 0x0C, 0x06, 0x0C, 0x06, 0x09, 0x03, 0x09, 0x03 },
{ 7, 0x2C, 7, 0x2C, 0x0D, 0x83, 0x0D, 0x83, 
  0x2C, 7, 0x2C,7, 0x83, 0x0D, 0x83, 0x0D  }
};
static unsigned char pentagone[][10]={
{ 0x0D, 5,5,5,7 , 0x0D, 5,5,5,7},
{ 0x0E, 0x0D,5,5,0x16, 0x49, 5,5,7,0x0B},
{ 7, 0x0C, 6, 0x0C, 0x24, 0x0D, 0x81, 3, 9, 3 },
{ 0x0D, 5, 0x34, 7, 0x0B, 15,15,15,15,15 },
{ 0x0D, 0xF0, 7, 0x0B, 0x0E , 15,15,15,15,15 },
{ 0x0E, 0x0A, 0x49, 5, 7, 0x0D, 5, 0x16, 0x0A, 0x0B },
{ 0x0B, 0x0E, 0x49, 0x16, 0x49, 7, 0x0D, 0x16, 0x49, 0x16 },
{ 0x2C, 5, 0x16, 0x0B, 0x0E, 0x0B, 0x0E, 0x49, 5, 0x83 },
{ 0x0E, 0x49, 5, 0x16, 0x0B, 0x0E, 0x49, 5, 0x16, 0x0B },
{ 0x0E, 0x0D, 0xC1, 0x16, 0x0B, 0x49, 0x34, 7, 0x0B, 0x0E },
{ 7, 0x2C, 5, 7, 0x2C, 0x0D, 0x83, 0x0D, 5, 0x83 },
{ 0x0D, 0x34, 7, 0x0D, 0xC1, 7, 0x0A, 0x0B, 0x0E, 0x0A}
};

static unsigned char hexagone[][12]={
{ 0x0D, 5, 5, 5, 5, 7, 0x0D, 5, 5, 5, 5, 7},
{ 0x2C, 5, 5, 5, 7, 0x0E, 0x0B, 0x0D, 5, 5, 5, 0x83 },
{ 0x0D, 0x34, 5, 5, 7, 0x0E, 7, 0x0B, 0x0D, 5, 5, 0xC1 },
{ 0x0D, 5, 0x34, 5, 7, 0x0E, 5, 7, 0x0B, 0x0D, 5, 0xC1 },
{ 0x0D, 0x83, 0x2C, 5, 5, 7 , 15,15,15,15,15,15},

{ 0x0C, 0x24, 5, 7, 0x0C, 6, 9, 3, 0x0D, 5, 0x81, 3 },
{ 7, 0x0C, 6, 0x0D, 0x14, 0x24, 0x0D, 0x81, 0x41, 7, 9, 3},
{ 0x0D, 5, 0x83, 0x2C, 5, 7, 15,15,15,15,15,15},
{ 7, 0x0C, 0x24, 7, 0x0C, 0x24, 0x0D, 0x81, 3, 0x0D, 0x81, 3},
{ 0x0C, 4, 6, 0x0C, 4, 6, 9, 1, 3, 9, 1, 3 },

{ 0x0E, 0x0D, 0x14, 6, 0x48, 6, 9, 0x12, 9, 0x41, 7, 0x0B},
{ 0x0D, 5, 0x16, 0x49, 0x16, 0x0B, 15,15,15,15,15,15},
{ 0x0D, 0x34, 5, 0x16, 0x0E, 0x0B, 0x0E, 0x0B, 0x49, 5, 0xC1, 7},
{ 0x0D, 0x34, 7, 0x16, 0x0A, 0x0D, 0x0A, 0x49, 7, 0xC1, 7, 0x0D},
{ 0x0D, 0x14, 0x82, 0x0E, 9, 3, 15,15,15,15,15,15},

{ 0x0E, 0x0D, 0xD0, 6, 9, 3, 15,15,15,15,15,15},
{ 0x0D, 0xF0, 5, 7, 0x0E, 0x0B, 0x0E, 0x0B, 0x0D, 5, 0xF0, 7},
{ 0x0E, 0x0B, 0x0D, 5, 5, 0x92, 0x68, 5, 5, 7, 0x0E, 0x0B},
{ 0x0E, 0x49, 5, 0x16, 0x0A, 0x0B, 15,15,15,15,15,15},
{ 0x0B, 0x2C, 5, 5, 0x83, 0x0E, 15,15,15,15,15,15},

{ 0x0B, 0x0A, 0x2C, 5, 5, 7, 15,15,15,15,15,15},
{ 0x2C, 5, 5, 0x16, 0x0B, 0x0E, 0x0E, 0x0B, 5, 0x83, 0x49, 5},
{ 0x0A, 0x0D, 0xC1, 0x16, 0x49, 0x34, 7, 0x0A, 0x0E, 0x0B, 0x0E, 0x0B},
{ 0x0D, 0x16, 0x0D, 5, 0x92, 0x0B, 0x0E, 0x68, 5, 7, 0x49, 7},
{ 0x0D, 0x16, 0x0D, 0x16, 0x16, 0x49, 0x16, 0x49, 0x49, 7, 0x49, 7},

{ 0x0C, 6, 0x0C, 6, 9, 0x12, 9, 0x12, 7, 0x49, 7, 0x49 },
{ 0x2C, 7, 0x2C, 7, 0x92, 0x0D, 0x92, 0x0D, 0x49, 7, 0x49, 7},
{ 0x0E, 0x49, 5, 0x16, 0x49, 7, 15,15,15,15,15,15},
{ 0x0D, 0x16, 0x0B, 0x0E, 0x49, 7, 0x16, 0x49, 5, 0x83, 0x2C, 5},
{ 7, 0x0E, 0x49, 0x34, 7, 0x49, 0x0D, 0xC1, 0x16, 0x0B, 0x0D, 0x16},

{ 0x2C, 5, 0xC1, 7, 0x0E, 0x0B, 0x0B, 0x0D, 0x34, 5, 0x83, 0x0E },
{ 0x0E, 0x0B, 0x0D, 0x34, 0xC1, 7, 15,15,15,15,15,15},
{ 0x0D, 5, 0x16, 0x2C, 7, 0x0A, 0x0A, 0x0D, 0x83, 0x49, 5, 7 },
{ 0x2C, 0xC1, 5, 7, 0x0E, 0x0B, 0x0E, 0x0B, 0x0D, 5, 0x34, 0x83 },
{ 0x0A, 0x0D, 5, 0x92, 0x68, 5, 7, 0x0A, 0x0B, 0x0E, 0x0E, 0x0B}
};
static DBL tetragonal  (VECTOR EPoint, TPATTERN *TPat)
{
   unsigned char how;   
   long xv,zv;
   DBL return_value=0;
   DBL value;
   DBL value1;
   DBL value2;
   DBL x,y;
   int lng=0;
   zv = floor(y=EPoint[Z]); 
   xv = floor(x=EPoint[X]); 
   switch(TPat->Vals.Strata.Number)
   {
      case 6:
         switch(TPat->Vals.Strata.Variant)
         {
             case 0:
             case 1:
             case 2:
             case 3:
             case 5:
             case 6:
             case 8:
             case 9:
             xv %= 6; if (xv < 0) { xv += 6 ;}
             zv &= 0x01;
             lng = 6;
             break;
             case 4:
             case 7:
             case 19:
             case 20:
             lng = 0;
             zv %= 6; if (zv <0) { zv += 6;}
             xv += 5*zv;
             xv %= 6; if (xv <0) { xv += 6;  }                              
             break;
             case 11:
             case 18:
             case 27:
             lng = 0;
             zv %= 6; if (zv <0) { zv += 6;}
             xv += zv;
             xv %= 6; if (xv <0) { xv += 6;  }                              
             break;
             case 10:
             case 12:
             case 21:
             case 22:
             case 24:
             case 25:
             case 26:
             lng = 4;
             xv &= 0x03;
             zv %= 3; if (zv<0) { zv += 3;}
             break;
             case 13:
             case 32:
             lng = 3;
             zv &= 0x03;
             xv %= 3; if (xv < 0) { xv += 3; }
             break;
             case 14:
             lng = 3;
             zv %= 6; if (zv < 0) { zv += 6; }
             xv += 2 * (zv/2);
             zv &= 0x01;
             xv %= 3; if (xv < 0) { xv += 3; }
             break;
             case 15:
             lng = 2;
             xv %= 6; if (xv < 0) { xv+= 6; }
             zv += (xv/2);
             xv &= 0x01;
             zv %= 3; if (zv<0) { zv += 3;}
             break;
             case 16:
             case 17:
             lng = 6;
             zv %= 12; if (zv <0) { zv+=12; }
             xv += zv/2;
             zv &= 0x01;
             xv %= 6; if (xv < 0) { xv+= 6; }
             break;
             case 23:
             case 28:
             lng = 6;
             zv %= 12; if (zv <0) { zv+=12; }
             xv += 4* (zv/2);
             zv &= 0x01;
             xv %= 6; if (xv < 0) { xv+= 6; }
             break;
             case 29:
             case 30:
             lng = 6;
             zv &= 0x03;
             xv += 3* (zv/2);
             zv &= 0x01;
             xv %= 6; if (xv < 0) { xv+= 6; }
             break;
             case 31:
             lng = 0;
             zv %= 3; if (zv <0) { zv+=3; }
             xv += 4* zv;
             xv %= 6; if (xv < 0) { xv+= 6; }
             break;
             case 33:
             lng = 0;
             zv %= 12; if (zv < 0) { zv+= 12; }
             xv += 7*zv;
             xv %= 12; if (xv < 0) { xv+= 12; }
             break;
             case 34:
             lng = 4;
             zv %= 6; if (zv<0) { zv+=6;}
             xv += 2 * (zv/3);
             xv &= 0x03;
             zv %= 3; if (zv<0) { zv += 3;}
             break;
         }
         how = hexagone[TPat->Vals.Strata.Variant][xv+zv*lng];
         break;
      case 5:
         switch(TPat->Vals.Strata.Variant)
         {
             case 0:
             case 1:
             xv %= 5; if (xv <0) { xv += 5 ;  }
             zv &= 0x01;
             break;
             case 2:
             case 9:
             zv %= 10; if (zv <0) { zv += 10;}
             xv += 3 * (zv/2);
             xv %= 5; if (xv <0) { xv += 5;  }
             zv &= 0x01;
             break;
             case 10:
             zv %= 10; if (zv <0) { zv += 10;}
             xv += 4*(zv/2);
             xv %= 5; if (xv <0) { xv += 5 ;  }
             zv &= 0x01;
             break;
             case 3:
             zv %= 5; if (zv <0) { zv += 5;}
             xv += 2*zv;
             xv %= 5; if (xv <0) { xv += 5 ;  }
             zv = 0x0;
             break;
             case 4:
             zv %= 5; if (zv <0) { zv += 5;}
             xv += 2 * zv;
             xv %= 5; if (xv <0) { xv += 5;  }
             zv = 0x00;
             break;
             case 5:
             case 6:
             case 8:
             zv %= 10; if (zv <0) { zv += 10;}
             xv += zv;
             xv %= 10; if (xv <0) { xv += 10;  }
             zv = 0x00;
             break;
             case 11:
             zv %= 10; if (zv <0) { zv += 10;}
             xv += 8* zv;
             xv %= 10; if (xv <0) { xv += 10;  }
             zv = 0x00;
             break;
             case 7:
             zv %= 10; if (zv <0) { zv += 10;}
             xv += 3*zv;
             xv %= 10; if (xv <0) { xv += 10;  }
             zv = 0x00;
             break;
         }
         how = pentagone[TPat->Vals.Strata.Variant][xv+zv*5];
         break;
      case 4:
         xv &= 0x03; 
         zv &= 0x03;
         how = tetragone[TPat->Vals.Strata.Variant][xv+zv*4];
         break;
      case 3:
         xv %= 3; if (xv < 0) { xv += 3; }
         zv &= 0x01;
         how = trigone[TPat->Vals.Strata.Variant][xv+zv*3];
         break;
      case 2:
         zv &= 0x01;
         how = digone[zv]; 
         break;
      case 1:
      default:
         how = 0x0F;
      break;
   }
   /* 
   **   5---1---6
   **   |       |
   **   4       2
   **   |       |
   **   8---3---7
   */
   x -= floor(x);
   y -= floor(y);
   switch(TPat->Vals.Strata.Measure_ang)
   { 
   case 2:
   if ((how & 0x16) == 0x16)
   {
      value1 = 2*x;
      value2 = 2 - 2*y;
      value = fabs(sqrt(value1*value1+value2*value2) - 1.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if ((how & 0x2C) == 0x2C)
   {
      value1 = 2 - 2*x;
      value2 = 2 - 2*y;
      value = fabs(sqrt(value1*value1+value2*value2) - 1.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if ((how & 0x49) == 0x49)
   {
      value1 = 2 - 2*x;
      value2 = 2*y;
      value = fabs(sqrt(value1*value1+value2*value2) - 1.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if ((how & 0x83) == 0x83)
   {
      value1 = 2*x;
      value2 = 2*y;
      value = fabs(sqrt(value1*value1+value2*value2) - 1.0);
      return_value = min(value,1.0);
      return return_value;
   }
   break;
   case 1:
   if ((how & 0x16) == 0x16)
   {
      value1 = 2*x;
      value2 = 2*y;
      value = fabs(value1 - value2 + 1.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if ((how & 0x2C) == 0x2C)
   {
      value1 = 2*x;
      value2 = 2 - 2*y;
      value = fabs(value2 - value1 + 1.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if ((how & 0x49) == 0x49)
   {
      value1 = 2*x;
      value2 = 2*y;
      value = fabs(value2 - value1 + 1.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if ((how & 0x83) == 0x83)
   {
      value1 = 2*x;
      value2 = 2 - 2*y;
      value = fabs(value1 - value2 + 1.0);
      return_value = min(value,1.0);
      return return_value;
   }
   break;
   case 3: 
   if ((how & 0x16) == 0x16)
   {
      value1 = 2*x -1;
      value2 = 1 - 2*y;
      return_value = max(value1,value2);
      value1 = 1 - 2*x;
      value2 = 2*y - 1;
      value = min(value1,value2);
      return_value = max(return_value,value);                                   
      return return_value;
   }
   if ((how & 0x2C) == 0x2C)
   {
      value1 = 1 - 2*x;
      value2 = 1 - 2*y;
      return_value = max(value1,value2);
      value1 = 2*x - 1;
      value2 = 2*y - 1;
      value = min(value1,value2);
      return_value = max(return_value,value);                                   
      return return_value;
   }
   if ((how & 0x49) == 0x49)
   {
      value1 = 1 - 2*x;
      value2 = 2*y - 1;
      return_value = max(value1,value2);
      value1 = 2*x - 1;
      value2 = 1 - 2*y;
      value = min(value1,value2);
      return_value = max(return_value,value);                                   
      return return_value;
   }
   if ((how & 0x83) == 0x83)
   {
      value1 = 2*x -1;
      value2 = 2*y -1;
      return_value = max(value1,value2);
      value1 = 1 - 2*x;
      value2 = 1 - 2*y;
      value = min(value1,value2);
      return_value = max(return_value,value);                                   
      return return_value;
   }
   break;
   default:
   case 0:
      break;
   }
   if (how & 0x01)
   {
      value = 2*y - 1;
      return_value = max(return_value,value);
   }
   if (how & 0x02)
   {
      value = 2*x -1;
      return_value = max(return_value,value);
   }
   if (how & 0x04)
   {
      value = 1 - 2*y;
      return_value = max(return_value,value);
   }
   if (how & 0x08)
   {
      value = 1 - 2*x;
      return_value = max(return_value,value);
   }
   switch(TPat->Vals.Strata.Measure_int)
   { 
   case 2:
   
   if (how & 0x40)
   {
      value1 = 2 - 2*x;
      value2 = 2*y;
      value = 1.0- sqrt(value1*value1+value2*value2);
      return_value = max(return_value,value);
   }
   if (how & 0x80)
   {
      value1 = 2*x;
      value2 = 2*y;
      value = 1.0- sqrt(value1*value1+value2*value2);
      return_value = max(return_value,value);
   }
   if (how & 0x10)
   {
      value1 = 2*x;
      value2 = 2 - 2*y;
      value = 1.0- sqrt(value1*value1+value2*value2);
      return_value = max(return_value,value);
   }
   if (how & 0x20)
   {
      value1 = 2 - 2*x;
      value2 = 2 - 2*y;
      value = 1.0- sqrt(value1*value1+value2*value2);
      return_value = max(return_value,value);
   }
   break;
   case 1:
   if (how & 0x40)
   {
      value1 = 2 - 2*x;
      value2 = 2*y;
      value = 1.0- (value1+value2);
      return_value = max(return_value,value);
   }
   if (how & 0x80)
   {
      value1 = 2*x;
      value2 = 2*y;
      value = 1.0- (value1+value2);
      return_value = max(return_value,value);
   }
   if (how & 0x10)
   {
      value1 = 2*x;
      value2 = 2 - 2*y;
      value = 1.0- (value1+value2);
      return_value = max(return_value,value);
   }
   if (how & 0x20)
   {
      value1 = 2 - 2*x;
      value2 = 2 - 2*y;
      value = 1.0- (value1+value2);
      return_value = max(return_value,value);
   }
   break;
   default:
   case 0:
   if (how & 0x10)
   {
      value1 = 1 - 2*x;
      value2 = 2*y - 1;
      value = min(value1,value2);
      return_value = max(return_value,value);
   }
   if (how & 0x20)
   {
      value1 = 2*x - 1;
      value2 = 2*y - 1;
      value = min(value1,value2);
      return_value = max(return_value,value);
   }
   if (how & 0x40)
   {
      value1 = 2*x - 1;
      value2 = 1 - 2*y;
      value = min(value1,value2);
      return_value = max(return_value,value);
   }
   if (how & 0x80)
   {
      value1 = 1 - 2*x;
      value2 = 1 - 2*y;
      value = min(value1,value2);
      return_value = max(return_value,value);
   }
      break;
   }
   switch(TPat->Vals.Strata.Measure_ext)
   { 
   case 2:
      value1 = 2*x - 1;
      value2 = 2*y - 1;
   if ((((how & 0x06) == 0x06)&&(value1>=0.0)&&(value2<=0.0))
     ||(((how & 0x0C) == 0x0C)&&(value1<=0.0)&&(value2<=0.0))
     ||(((how & 0x09) == 0x09)&&(value1<=0.0)&&(value2>=0.0))
     ||(((how & 0x03) == 0x03)&&(value1>=0.0)&&(value2>=0.0))
      )
   {
      value = sqrt(value1*value1+value2*value2);
      value = min(value,1.0);
      return_value = max(return_value,value);
   }
   break;
   case 1:
   if ((how & 0x06) == 0x06)
   {
      value1 = 2 - 2*x;
      value2 = 2*y;
      value = 2.0- (value1+value2);
      value = min(value,1.0);
      return_value = max(return_value,value);
   }
   if ((how & 0x0C) == 0x0C)
   {
      value1 = 2*x;
      value2 = 2*y;
      value = 2.0- (value1+value2);
      value = min(value,1.0);
      return_value = max(return_value,value);
   }
   if ((how & 0x09) == 0x09)
   {
      value1 = 2*x;
      value2 = 2 - 2*y;
      value = 2.0- (value1+value2);
      value = min(value,1.0);
      return_value = max(return_value,value);
   }
   if ((how & 0x03) == 0x03)
   {
      value1 = 2 - 2*x;
      value2 = 2 - 2*y;
      value = 2.0- (value1+value2);
      value = min(value,1.0);
      return_value = max(return_value,value);
   }
   break;
   default:
   case 0:
      break;
   }
   return return_value;
}
static unsigned short tritrigone[][6]={
{0x215,0x344,0x126,   0x126,0x344,0x215}};
static unsigned short tritetragone[][8]={
{0x126,0x126,0x144,0x144, 0x126,0x126,0x144,0x144},
{0x611,0x344,0x126,0x423, 0x611,0x344,0x126,0x423},
{5,0,6,3, 0,5,3,6} 
};
static unsigned short tripentagone[][10]={
{0x215,0x244,4,0x144,0x126, 0x126,0x144,4,0x244,0x215},
{0x215,0x244,0x244,0x611,0x611, 0x244,0x244,0x215,0x423,0x423},
{0x146,0x126,0x344,0x611,0x611, 0x344, 0x126,0x146,0x522,0x522},
{5,0,0x244,0x215,3, 3,0x215,0x244,0,5}
};
static unsigned short trihexagone[][12]={
{0x215,0x244,4,4,0x244,0x215, 0x215,0x244,4,4,0x244,0x215},
{0x413,0x522,0x144,4,0x244,0x215, 0x413,0x522,0x144,4,0x244,0x215},
{3,0x126,0x144,0,0x244,0x215, 3,0x126,0x144,0,0x244,0x215},
{0x215,0x244,4,0,6,3, 4,0x244,0x215,3,6,0}, 
{5,0,0x244,0x215,0x245,0x211, 5,0x211,0x245,0x215,0x244,0},
{0x245,0x211,0x215,0x244,0x244,0x211, 0x245,0x211,0x244,0x244,0x215,0x211},

{0x215,0x244,0x244,0x611, 0x522,0x146,0x146,0x522, 0x611,0x244,0x244,0x215},
{5,0,0x244,0x215, 0x146,0x122,0x122,0x146, 0x215,0x244,0,5},
{5,0,6,0x126,0x344,0x211, 5,0x211,0x344,0x126,6,0},
{0x215,0x344,0x122,0x122,0x344,0x215, 0x215,0x344,0x122,0x122,0x344,0x215},
{5,0,6,6,0,5, 6,0,5,5,0,6},
{0x691,0x3C4,0x5A2,0x5A2,0x3C4,0x691, 0x5A2,0x3C4,0x691,0x691,0x3C4,0x5A2},
};

static DBL trigonal  (VECTOR EPoint, TPATTERN *TPat)
{
   unsigned short how;   
   long xv,zv;
   DBL return_value=0;
   DBL value;
   DBL value1;
   DBL value2;
   DBL dist1;
   DBL dist2;
   DBL dist3;
   DBL x,z;
   int lng=0;

x=EPoint[X];
z=EPoint[Z];

 xv = floor(x);
 zv = floor(z/SQR3);
 x -= xv;
 z -= SQR3*zv;
 /* x,z is in { [0.0, 1.0 [, [0.0, SQR3 [ } 
 ** There is some mirror to reduce the problem
 */
 zv *= 2;
 xv *= 2;
 if ( z > SQR3_2 )
 {
    z -= SQR3_2;
    if (x>0.5)
    {
    x -= 0.5;
    xv++;
    }
    else 
    {
    x += 0.5;
    xv--;
    }
    zv++;
 }
    if ((x == 0.0)||(z/x>SQR3))
    { 
      z = SQR3_2 - z;
      x = 0.5 -x;
      xv--;
    }
    if ((x == 1.0)||(z/(1.0-x)>SQR3))
    {
      z = SQR3_2 - z;
      x = 1.5 -x;
      xv++;
    }
   switch(TPat->Vals.Strata.Number)
   {
      case 6:
         switch(TPat->Vals.Strata.Variant)
         {
             case 0:
             case 1:
             case 9:
             xv += 5*zv;
             zv = 0;
             xv %= 6; if (xv <0) { xv += 6;}
             lng = 0;
             break;
             case 2:
             case 10:
             case 11:
             zv &= 0x01;
             xv += 3*zv;
             xv %= 6; if (xv <0) { xv += 6;}
             lng = 0;
             break;
             case 3:
             xv += 14*((zv%6+((zv%6)<0?6:0))/2);
             xv %= 6; if (xv <0) { xv += 6;}
             lng = 6;
             zv &= 0x01;
             break;
             case 4:
             case 8:
             xv += 8*((zv%6+((zv%6)<0?6:0))/2);
             xv %= 6; if (xv <0) { xv += 6;}
             lng = 6;
             zv &= 0x01;
             break; 
             case 5:
             xv %= 6; if (xv <0) { xv += 6;}
             lng = 6;
             zv &= 0x01;
             break; 
             case 6:
             case 7:
             xv -= ((zv%12+((zv%12)<0?12:0))/3); 
             xv &= 3;
             lng = 4;
             zv %= 3; if (zv <0) { zv +=3;}
             break; 
         }
         how = trihexagone[TPat->Vals.Strata.Variant][xv+zv*lng];
         break;
      case 5:
         switch(TPat->Vals.Strata.Variant)
         {
             case 0:
             case 1:
             case 2:
             zv &= 0x01;
             xv += 5*zv;
             xv %= 10; if (xv <0) { xv += 10 ;  }
             zv = 0x00;
             break;
             case 3:
             zv %= 10; if (zv <0) { zv += 10; }
             xv += 3*zv;
             xv %= 10; if (xv <0) { xv += 10 ;  }
             zv = 0x00;
             break;
         }
         how = tripentagone[TPat->Vals.Strata.Variant][xv];
         break;
      case 4:
         zv &= 0x03;
         xv += zv;
         xv &= 0x03; 
         zv &= 0x01;
         how = tritetragone[TPat->Vals.Strata.Variant][xv+zv*4];
         break;
      case 3:
         zv &= 0x01;
         xv += 3*zv;
         xv %= 6; if (xv < 0) { xv += 6; }
         zv = 0x00;
         how = tritrigone[TPat->Vals.Strata.Variant][xv];
         break;
      case 2:
         how = 0x166; 
         break;
      case 1:
      default:
         how = 0x07;
      break;
   }

   /* 
   **     / \
   **    1   \
   **   /     2
   **  /       \
   **  -----3-----
   ** *3/2
   */
   if (how & 0x01)
   {
      dist1 = 1.0 - (fabs(SQR3 * x - z) * SQR3 );
      return_value = max(return_value,dist1);
   }
   if (how & 0x02)
   {
      dist2 = 1.0 - (fabs(SQR3 * (1.0-x) - z) * SQR3  );
      return_value = max(return_value,dist2);
   }
   if (how & 0x04)
   {
      dist3 = 1.0 - (z * 2.0 * SQR3 );
      return_value = max(return_value,dist3);
   }
 switch(TPat->Vals.Strata.Measure_int) 
 {
 case 1:
      dist1 = (1.0 - (fabs(SQR3 * z + x) ));
      dist2 = (1.0 - (fabs(SQR3 * z - x + 1.0) ));
      dist3 = (1.0 - (x * 2.0 ));
   if (((how & 0x83) == 0x00)&&(dist1<0)&&(dist2<0)) 
   {
      value1 = (3.0 / 2.0 *(fabs(SQR3 * z + x) ) - 2.0);
      value2 = (3.0 / 2.0 *(fabs(SQR3 * z - x + 1.0) ) - 2.0);
      value =  min(value1,value2);
      return_value = max(return_value,value);
   }
   if (((how & 0x85) == 0x00)&&(dist1>0)&&(dist3>0))
   {
      value1 = (1.0 - 3.0 / 2.0 * (fabs(SQR3 * z + x) ));
      value2 = (1.0 - (x * 3.0 ));
      value =  min(value1,value2);
      return_value = max(return_value,value);
   }
   if (((how & 0x86) == 0x00)&&(dist3<0)&&(dist2>0))
   {
      value1 = (1.0 - 3.0 / 2.0 *(fabs(SQR3 * z - x + 1.0) ));
      value2 = ((x * 3.0 ) - 2.0);
      value =  min(value1,value2);
      return_value = max(return_value,value);
   }
   break; 
 case 2:
   if ((how & 0x83) == 0x00)
   {
      dist1 = x - 0.5;
      dist2 = z - SQR3_2;
      dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2))*3.0 );
      return_value = max(return_value,dist3);
   }
   if ((how & 0x85) == 0x00)
   {
      dist1 = x;
      dist2 = z;
      dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2)) *3.0);
      return_value = max(return_value,dist3);
   }
   if ((how & 0x86) == 0x00)
   {
      dist1 = x - 1.0;
      dist2 = z ;
      dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2)) *3.0);
      return_value = max(return_value,dist3);
   }
   break; 
 case 0:
   if ((how & 0x83) == 0x00)
   {
      dist3 = 1.0 - ((SQR3_2 - z) * 2.0 * SQR3  );
      return_value = max(return_value,dist3);
   }
   if ((how & 0x85) == 0x00)
   {
      dist2 = 1.0 - (fabs(SQR3 * x + z) * SQR3 );
      return_value = max(return_value,dist2);
   }
   if ((how & 0x86) == 0x00)
   {
      dist1 = 1.0 - (fabs(SQR3 * (x -1.0) - z) * SQR3 );
      return_value = max(return_value,dist1);
   }
   break; 
 }
   switch(TPat->Vals.Strata.Measure_ext)
   { 
   case 2:
      dist1 = (1.0 - (fabs(SQR3 * z + x) ));
      dist2 = (1.0 - (fabs(SQR3 * z - x + 1.0) ));
      dist3 = (1.0 - (x * 2.0 ));
   if ((((how & 0x03) == 0x03)&&(dist1<=0.0)&&(dist2<=0.0))
     ||(((how & 0x06) == 0x06)&&(dist2>=0.0)&&(dist3<=0.0))
     ||(((how & 0x05) == 0x05)&&(dist1>=0.0)&&(dist3>=0.0))
      )
   {
      value1 = x - 0.5;
      value2 = z - SQR3_2/3.0;
      value = 2 * SQR3 *sqrt(value1*value1+value2*value2);
      return_value = min(1.0,value);
   }
   break;
   case 1:
      /* dist1 = (1.0 - (fabs(SQR3 * z + x) )); */
      dist1 = (1.0 - (fabs(SQR3 * x - z)*SQR3 ));
      /* dist2 = (1.0 - (fabs(SQR3 * z - x + 1.0) )); */
      dist2 = (1.0 - (fabs(SQR3 * (1.0 -x ) -z ) *SQR3));
      /* dist3 = (1.0 - (x * 2.0 )); */
      dist3 = (1.0 - (z * 2.0 * SQR3));
      value1 = (x - 0.5);
      value2 = (z - SQR3_2/3.0);
   if  (((how & 0x03) == 0x03)&&(dist1>=0.0)&&(dist2>=0.0))
   {
      value = fabs(value2 * 2.0 * SQR3 );
      return_value = min(1.0,value);
   }
   if  (((how & 0x06) == 0x06)&&(dist2>=0.0)&&(dist3>=0.0))
   {
      value = fabs(SQR3 * value1 - value2) * SQR3 ;
      return_value = min(1.0,value);
   }
   if  (((how & 0x05) == 0x05)&&(dist1>=0.0)&&(dist3>=0.0))
   {
      value = fabs(SQR3 * value1 + value2) * SQR3  ;
      return_value = min(1.0,value);
   }
   break;
   case 0:
   default:
      break;
  }
      dist1 = (1.0 - (fabs(SQR3 * z + x) ));
      dist2 = (1.0 - (fabs(SQR3 * z - x + 1.0) ));
      dist3 = (1.0 - (x * 2.0 ));                                               
   switch(TPat->Vals.Strata.Measure_ang)
   { 
   case 2:
   if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
   {
      value1 = x;
      value2 = z;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x140) == 0x140)&&(dist1>0)&&(dist3>0))
   {
      value1 = x - 0.5;
      value2 = z- SQR3_2;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x210) == 0x210)&&(dist1<0)&&(dist2<0))
   {
      value1 = x - 1.0;
      value2 = z;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x240) == 0x240)&&(dist3<0)&&(dist2>0))
   {
      value1 = x - 0.5;
      value2 = z - SQR3_2;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x410) == 0x410)&&(dist1>0)&&(dist3>0))
   {
      value1 = x - 1.0;
      value2 = z;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x420) == 0x420)&&(dist3<0)&&(dist2>0))
   {
      value1 = x;
      value2 = z;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   break;
   case 1:
   if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
   {
      value = -dist1 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x140) == 0x140)&&(dist1>0)&&(dist3>0))
   {
      value = dist1 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x210) == 0x210)&&(dist1<0)&&(dist2<0))
   {
      value = -dist2 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x240) == 0x240)&&(dist3<0)&&(dist2>0))
   {
      value = dist2 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x410) == 0x410)&&(dist1>0)&&(dist3>0))
   {
      value = dist3 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x420) == 0x420)&&(dist2>0)&&(dist3<0))
   {
      value = -dist3 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
   {
      value = -dist1 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   break;
   default:
   case 0:
      break;
   }
   return return_value;
}

/*
** open face (3b)
** special (1b)
** nexus vertex/open face (3b)
** unused (1b)
** close face (3b)
*/
static unsigned short hexmonogone[][6]={
{0x691,0x3C4,0x5A2,0x5A2,0x3C4,0x691 }
};
static unsigned short hexdigone[][12]={
{0x691,0x3C4,0x1A2,0x1A2,0x3C4,0x691,0x5A2,0x1C4,0x6080,0x6080,0x1C4,0x5A2 }
};
static unsigned short hextrigone[][36]={
{0x691,0x3C4,0x5A2,  0x5A2,0x3C4, 0x691,
 0x5A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x691,0x3C4,0x5A2 },

{0x691,0x3C4,0x1A2, 0x1A2,0x3C4,0x691,
 0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
 0x691,0x2C4,0x4080, 0x1080,0x1080,0x491,
 0x5A2,0x3C4,0x291, 0x291,0x3C4,0x5A2,
 0x491,0x1080,0x1080, 0x4080,0x2C4,0x691,
 0x4A2,0x2080,0x2080, 0x4080,0x1C4,0x5A2
 },

{0x691,0x3C4,0x1A2, 0x1A2,0x3C4,0x291, 0x291,0x3C4,0x5A2,
 0x5A2,0x1C4,0x6080,0x6080,0x0C4,0x5080,0x5080,0x2C4,0x691,
 0x5A2,0x1C4,0x6080,0x6080,0x0C4,0x5080,0x5080,0x2C4,0x691,
 0x691,0x3C4,0x1A2, 0x1A2,0x3C4,0x291, 0x291,0x3C4,0x5A2 }
};
static unsigned short hextetragone[][48]={
{0x691,0x3C4,0x5A2,  0x5A2,0x3C4, 0x691,
 0x5A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x691,0x3C4,0x5A2 },

{0x691,0x2C4,0x4080, 0x1080,0x80,0x2080, 
 0x4080,0x2C4,0x691, 0x291,0x3C4,0x1A2,
 0x4080,0x1C4,0x5A2, 0x1A2,0x3C4,0x291,
 0x5A2,0x1C4,0x4080, 0x2080,0x80,0x1080, 
 0x691,0x2C4,0x4080, 0x1080,0x80,0x2080,
 0x4080,0x2C4,0x691, 0x291,0x3C4,0x1A2,
 0x4080,0x1C4,0x5A2, 0x1A2,0x3C4,0x291,
 0x5A2,0x1C4,0x4080, 0x2080,0x80,0x1080
},

{0x691,0x2C4,0x5080,  0x5080,0x0C4,0x6080, 
 0x6080,0x0C4,0x5080, 0x5080,0x2C4,0x691,
 0x5A2,0x3C4,0x291, 0x291,0x3C4,0x1A2,
 0x1A2,0x3C4,0x291, 0x291,0x3C4,0x5A2,
 0x691,0x2C4,0x5080,  0x5080,0x0C4,0x6080, 
 0x6080,0x0C4,0x5080, 0x5080,0x2C4,0x691,
 0x5A2,0x3C4,0x291, 0x291,0x3C4,0x1A2,
 0x1A2,0x3C4,0x291, 0x291,0x3C4,0x5A2,
},
{0x691,0x3C4,0x5A2,  0x4A2,0x3080,0x491,
 0x1A2,0x3C4,0x691,  0x691,0x3C4,0x5A2,
 0x6080,0x1C4,0x5A2, 0x1A2,0x3C4,0x691,
 0x5A2,0x1C4,0x6080, 0x6080,0x1C4,0x5A2,
 0x491,0x3080,0x0A2, 0x5A2,0x1C4,0x6080,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x0A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2
},
{0x691,0x3C4,0x1A2,  0x5A2,0x1C4,0x6080,
 0x5080,0x2C4,0x691, 0x491,0x3080,0x0A2,
 0x291,0x3C4,0x5A2,  0x4A2,0x3080,0x091,
 0x5A2,0x3C4,0x291,  0x691,0x2C4,0x5080,
 0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
 0x1A2,0x3C4,0x691,  0x091,0x3080,0x4A2,
 0x6080,0x1C4,0x5A2, 0x0A2,0x3080,0x491,
 0x5A2,0x1C4,0x6080, 0x6080,0x1C4,0x5A2
},
{0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x491,
 0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
 0x691,0x2C4,0x4080, 0x1080,0x1080,0x491,
 0x5A2,0x3C4,0x291,  0x291,0x3C4,0x5A2,
 0x491,0x1080,0x1080,0x4080,0x2C4,0x691,
 0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2,
 0x491,0x3080,0x0A2, 0x5A2,0x3C4,0x691,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2
},
{0x691,0x3C4,0x5A2,  0x1A2,0x3C4,0x691, 0x091,0x3080,0x0A2, 0x5A2,0x3C4,0x291,
 0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2,0x4A2,0x3080,0x491, 0x691,0x2C4,0x5080,
 0x091,0x3080,0x0A2, 0x5A2,0x3C4,0x291, 0x691,0x3C4,0x5A2,  0x1A2,0x3C4,0x691,
 0x4A2,0x3080,0x491, 0x691,0x2C4,0x5080,0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2
}
};
static unsigned short hexpentagone[][60]={ /* 0 */
{0x691,0x3C4,0x5A2,  0x5A2,0x3C4, 0x691,
 0x5A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x691,0x3C4,0x5A2 },
/* 1 */
{0x6080,0x1C4,0x5A2, 0x1A2,0x3C4,0x691,
 0x5A2,0x1C4,0x6080, 0x6080,0x1C4,0x5A2 ,
 0x491,0x3080,0x0A2, 0x5A2,0x1C4,0x6080,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x0A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x691,0x3C4,0x5A2,  0x4A2,0x3080,0x491,
 0x1A2,0x3C4,0x691,  0x691,0x3C4,0x5A2
}, /* 2 */
{0x691,0x3C4,0x1A2,  0x1A2,0x3C4, 0x691,
 0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
 0x691,0x2C4,0x4080, 0x1080,0x1080,0x491, 
 0x5A2,0x3C4,0x691,  0x091,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x091, 0x691,0x3C4,0x5A2,
 0x491,0x1080,0x1080,0x4080,0x2C4,0x691,
 0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2},
/* 3 */
{0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x491,
 0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
 0x691,0x2C4,0x4080, 0x1080,0x1080,0x491, 
 0x5A2,0x3C4,0x691,  0x091,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x091, 0x691,0x3C4,0x5A2,
 0x491,0x1080,0x1080,0x4080,0x2C4,0x691,
 0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2,
 0x491,0x3080,0x0A2, 0x5A2,0x3C4, 0x691,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2},
/* 4 */
{0x6080,0x1C4,0x5A2, 0x0A2,0x3080,0x491,
 0x5A2,0x1C4,0x6080, 0x6080,0x1C4,0x5A2 ,
 0x491,0x3080,0x0A2, 0x5A2,0x1C4,0x6080,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x0A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x691,0x3C4,0x5A2,  0x4A2,0x3080,0x491,
 0x5A2,0x3C4,0x691,  0x691,0x3C4,0x5A2,
 0x491,0x3080,0x4A2, 0x5A2,0x3C4,0x691,
 0x0A2,0x3080,0x491, 0x491,0x3080,0x4A2,
}, /* 5 */
{0x691,0x3C4,0x1A2,  0x1A2,0x3C4, 0x691,
 0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
 0x491,0x1080,0x080, 0x080,0x1080,0x491,
 0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2,
 0x491,0x3080,0x0A2, 0x5A2,0x3C4,0x691,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x491,
 0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
 0x491,0x1080,0x080, 0x080,0x1080,0x491,
 0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2 },
/* 6 */
{0x691,0x2C4,0x4080, 0x1080,0x1080, 0x491,
 0x5A2,0x3C4,0x291,  0x291,0x3C4,0x5A2,
 0x491,0x1080,0x1080,0x4080,0x2C4,0x691,
 0x4A2,0x2080,0x080, 0x080,0x2080,0x4A2,
 0x491,0x1080,0x080, 0x080,0x1080,0x491,
 0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2,
 0x691,0x3C4,0x1A2,  0x1A2,0x3C4,0x691,
 0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
 0x491,0x1080,0x080, 0x080,0x1080,0x491,
 0x4A2,0x2080,0x080, 0x080,0x2080,0x4A2 },
/* 7 */
{0x6080,0x1C4,0x5A2, 
 0x5A2,0x1C4,0x4080,
 0x691,0x2C4,0x4080,
 0x4080,0x2C4,0x691,
 0x4080,0x1C4,0x5A2,
 0x5A2,0x1C4,0x6080,
 0x691,0x3C4,0x1A2,

 0x5A2,0x1C4,0x6080,
 0x491,0x3080,0x0A2,

 0x4A2,0x3080,0x091,
 0x491,0x1080,0x1080,
 0x4A2,0x2080,0x2080,
 0x691,0x3C4,0x1A2,
 0x1A2,0x3C4,0x691,
 0x2080,0x2080,0x4A2,
 0x1080,0x1080,0x491, 
 0x091,0x3080,0x4A2,
 0x0A2,0x3080,0x491, 
 0x6080,0x1C4,0x5A2,
 0x1A2,0x3C4,0x691},
 
/* 8 */
{0x6080,0x0C4,0x5080, 0x5080,0x2C4, 0x291,
 0x5A2,0x3C4,0x291,  0x291,0x2C4,0x5080,
 0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
 0x5080,0x2C4,0x691, 0x091,0x3080,0x4A2,
 0x091,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x691,0x3C4,0x5A2, 0x0A2,0x3080,0x491,
 0x5A2,0x1C4,0x6080, 0x6080,0x1C4,0x5A2,

 0x691,0x3C4,0x1A2, 0x5A2,0x1C4,0x6080,
 0x1A2,0x3C4,0x291, 0x691,0x3C4,0x1A2 },

/* 9 */
{0x691,0x3C4,0x1A2,  0x1A2,0x3C4,0x691,
 0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2,
 0x291,0x3C4,0x1A2,  0x1A2,0x3C4,0x291, 
 0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080,
 0x491,0x3080,0x0A2, 0x5A2,0x3C4,0x691,
 0x6A2,0x3080,0x491, 0x491,0x3080,0x4A2, 
 0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x491,
 0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2, 
 0x291,0x3C4,0x1A2, 0x1A2,0x3C4,0x291,
 0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080,

},
/* 10 */
{0x691,0x3C4,0x1A2,  0x1A2,0x3C4,0x691, 
 0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
 0x491,0x1080,0x80, 0x80,0x1080,0x091,
 0x4A2,0x2080,0x2080, 0x4080,0x0C4,0x5080,
 0x691,0x3C4,0x1A2, 0x5A2,0x3C4,0x291,
 0x5080,0x2C4,0x691, 0x691,0x2C4,0x5080,
 0x291,0x3C4,0x5A2,  0x1A2,0x3C4,0x691, 
 0x5080,0x0C4,0x4080, 0x2080,0x2080,0x4A2,
 0x091,0x1080,0x80, 0x80,0x1080,0x491,
 0x4A2,0x2080,0x2080, 0x4080,0x1C4,0x5A2},
/* 11 */
{0x291,0x3C4,0x5A2,  0x1A2,0x3C4,0x691,
 0x4080,0x0C4,0x4080, 0x2080,0x2080,0x4A2,
 0x4080,0x0C4,0x4080, 0x1080,0x1080,0x491,
 0x5A2,0x3C4,0x291,  0x291,0x3C4,0x5A2,
 0x491,0x1080,0x1080,0x4080,0x0C4,0x4080,
 0x4A2,0x2080,0x2080,0x4080,0x0C4,0x4080,
 0x691,0x3C4,0x1A2, 0x5A2,0x3C4,0x291,
 0x1A2,0x3C4,0x691, 0x491,0x1080,0x1080,
 0x2080,0x2080,0x4A2, 0x4A2,0x2080,0x2080,
 0x1080,0x1080,0x491, 0x691,0x3C4,0x1A2
},
/* 12 */
{0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
 0x5080,0x2C4,0x691, 0x091,0x3080,0x4A2,
 0x291,0x3C4,0x5A2,  0x0A2,0x3080,0x491,
 0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2,
 0x291,0x3C4,0x1A2,  0x1A2,0x3C4,0x291,
 0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080,
 0x491,0x3080,0x0A2, 0x5A2,0x3C4,0x291,
 0x4A2,0x3080,0x091, 0x691,0x2C4,0x5080,
 0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
 0x5A2,0x3C4,0x291,  0x291,0x3C4,0x5A2
},
/* 13 */
{0x691,0x3C4,0x5A2,  0x5A2,0x3C4,0x691,
 0x5A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x091,
 0x4A2,0x3080,0x091, 0x691,0x2C4,0x5080,
 0x691,0x2C4,0x5080, 0x5080,0x2C4,0x291,
 0x5080,0x2C4,0x291, 0x291,0x2C4,0x5080,
 0x291,0x2C4,0x5080, 0x5080,0x2C4,0x691,
 0x5080,0x2C4,0x691, 0x091,0x3080,0x4A2,
 0x091,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x691,0x3C4,0x5A2
},
/* 14 */
{0x291,0x3C4,0x5A2, 0x5A2,0x3C4,0x691,
 0x5A2,0x3C4,0x691, 0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2,0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491,0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2,0x4A2,0x3080,0x091,
 0x4A2,0x3080,0x091,0x691,0x2C4,0x5080,
 0x691,0x2C4,0x5080,0x5080,0x2C4,0x291,
 0x5080,0x2C4,0x291,0x291,0x2C4,0x5080,
 0x291,0x2C4,0x5080,0x5080,0x2C4,0x691,
 0x5080,0x2C4,0x691,0x291,0x3C4,0x5A2
},
/* 15 */
{0x691,0x2C4,0x5080, 0x5080,0x0C4,0x4080,
 0x1A2,0x3C4,0x691,  0x291,0x2C4,0x4080,
 0x2080,0x2080,0x4A2,0x5A2,0x3C4,0x691,
 0x1080,0x1080,0x491,0x491,0x3080,0x4A2,
 0x091,0x3080,0x4A2, 0x4A2,0x3080,0x091,
 0x4A2,0x3080,0x491, 0x491,0x1080,0x1080,
 0x691,0x3C4,0x5A2,  0x4A2,0x2080,0x2080,
 0x4080,0x2C4,0x291, 0x691,0x3C4,0x1A2,
 0x4080,0x0C4,0x5080,0x5080,0x2C4,0x691,
 0x5A2,0x3C4,0x291,  0x291,0x3C4,0x5A2
},
/* 16 */
{0x4080,0x1C4,0x1A2, 0x1A2,0x1C4,0x4080,
 0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x4080,
 0x691,0x3C4,0x1A2,  0x5A2,0x3C4,0x691,
 0x1A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
 0x2080,0x2080,0x4A2,0x4A2,0x3080,0x091,
 0x1080,0x1080,0x491,0x491,0x1080,0x1080,
 0x091,0x3080,0x4A2, 0x4A2,0x2080,0x2080,
 0x4A2,0x3080,0x491, 0x691,0x3C4,0x1A2,
 0x691,0x3C4,0x5A2,  0x1A2,0x3C4,0x691,
 0x4080,0x0C4,0x6080,0x6080,0x1C4,0x5A2
},
/* 17 */
{0x291,0x3C4,0x1A2,  0x1A2,0x3C4,0x291,
 0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080,
 0x691,0x3C4,0x1A2,  0x5A2,0x3C4,0x691,
 0x5080,0x2C4,0x691, 0x491,0x3080,0x4A2,
 0x091,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x091,
 0x4A2,0x3080,0x491, 0x691,0x2C4,0x5080,
 0x691,0x3C4,0x5A2,  0x1A2,0x3C4,0x691,
 0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2
},
/* 18 */
{0x091,0x3080,0x4A2, 0x5A2,0x3C4,0x691,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x091,
 0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080,
 0x691,0x3C4,0x1A2,  0x1A2,0x3C4,0x691,
 0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2,
 0x091,0x3080,0x0A2, 0x5A2,0x3C4,0x691,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x691,0x3C4,0x5A2,  0x4A2,0x3080,0x091,
 0x5080,0x2C4,0x691, 0x691,0x2C4,0x5080
},
/* 19 */
{0x691,0x3C4,0x1A2,  0x1A2,0x3C4,0x691,
 0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
 0x691,0x2C4,0x4080, 0x1080,0x1080,0x091,
 0x5080,0x2C4,0x691, 0x291,0x2C4,0x5080,
 0x091,0x3080,0x4A2, 0x5A2,0x3C4,0x691,

 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x691,0x3C4,0x5A2,  0x4A2,0x3080,0x091,
 0x5080,0x2C4,0x291, 0x691,0x2C4,0x5080,
 0x091,0x1080,0x1080,0x4080,0x2C4,0x691,
 0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2
},
/* 20 */
{0x291,0x3C4,0x5A2,  0x5A2,0x3C4,0x691,
 0x5A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x091,
 0x4A2,0x3080,0x091, 0x691,0x2C4,0x5080,
 0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
 0x5080,0x2C4,0x691, 0x091,0x3080,0x4A2,
 0x091,0x3080,0x4A2, 0x4A2,0x3080,0x091,
 0x4A2,0x3080,0x091, 0x691,0x2C4,0x5080,
 0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
 0x5080,0x2C4,0x691, 0x291,0x3C4,0x5A2
},
/* 21 */
{0x691,0x3C4,0x1A2,  0x5A2,0x3C4,0x291,
 0x5080,0x2C4,0x691, 0x691,0x2C4,0x5080,
 0x291,0x3C4,0x5A2,  0x1A2,0x3C4,0x691,
 0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2,
 0x091,0x3080,0x0A2, 0x5A2,0x3C4,0x691,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
 0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
 0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x091,
 0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080
}
};
static DBL sexagonal  (VECTOR EPoint, TPATTERN *TPat)
{
   unsigned short how;   
   long xv,zv;
   DBL return_value=0;
   DBL value;
   DBL value1;
   DBL value2;
   DBL dist1;
   DBL dist2;
   DBL dist3;
   DBL x,z;
   int lng;

x=EPoint[X];
z=EPoint[Z];

 xv = floor(x);
 zv = floor(z/SQR3);
 x -= xv;
 z -= SQR3*zv;
 /* x,z is in { [0.0, 1.0 [, [0.0, SQR3 [ } 
 ** There is some mirror to reduce the problem
 */
 zv *= 2;
 xv *= 2;
 if ( z > SQR3_2 )
 {
    z -= SQR3_2;
    if (x>0.5)
    {
    x -= 0.5;
    xv++;
    }
    else 
    {
    x += 0.5;
    xv--;
    }
    zv++;
 }
    if ((x == 0.0)||(z/x>SQR3))
    { 
      z = SQR3_2 - z;
      x = 0.5 -x;
      xv--;
    }
    if ((x == 1.0)||(z/(1.0-x)>SQR3))
    {
      z = SQR3_2 - z;
      x = 1.5 -x;
      xv++;
    }
   switch(TPat->Vals.Strata.Number)
   {
      case 5:
         switch(TPat->Vals.Strata.Variant)
         {
             case 0:
             case 2:
             case 3:
             case 5:
             case 6:
             case 19:
               zv %= 10; if (zv < 0) { zv += 10; }
               xv %= 6; if (xv < 0) { xv += 6; }
               xv += 6*zv; /* 60 */
               zv = 0;
             break;
             case 1:
             case 4:
             case 9:
               zv -= 2*(((xv%30+(xv%30<0?30:0))/6));
               zv %= 10; if (zv < 0) { zv += 10; }
               xv %= 6; if (xv < 0) { xv += 6; }
               xv += 6*zv; /* 60 */
               zv = 0;
             break;
             case 7:
               zv -= 7*(((xv%60+(xv%60<0?60:0))/3)); 
               zv %= 20; if (zv < 0) { zv += 20; }
               xv %= 3; if (xv < 0) { xv += 3; }
               xv += 3*zv; /* 60 */
               zv = 0;
             break;
             case 8:
             case 10:
             case 13:
             case 14:
             case 15:
             case 17:
             case 20:
             case 21:
               zv += 2*(((xv%30+(xv%30<0?30:0))/6));
               zv %= 10; if (zv < 0) { zv += 10; }
               xv %= 6; if (xv < 0) { xv += 6; }
               xv += 6*zv; /* 60 */
               zv = 0;
             break;
             case 11:
             case 16:
               zv -= 6*(((xv%30+(xv%30<0?30:0))/6));
               zv %= 10; if (zv < 0) { zv += 10; }
               xv %= 6; if (xv < 0) { xv += 6; }
               xv += 6*zv; /* 60 */
               zv = 0;
             break;
             case 12:
             case 18:
               zv += 6*(((xv%30+(xv%30<0?30:0))/6));
               zv %= 10; if (zv < 0) { zv += 10; }
               xv %= 6; if (xv < 0) { xv += 6; }
               xv += 6*zv; /* 60 */
               zv = 0;
             break;
         }
         how = hexpentagone[TPat->Vals.Strata.Variant][xv];
         break;
      case 4:
         switch(TPat->Vals.Strata.Variant)
         {
             case 0:
               zv &= 0x07; 
               xv %= 6; if(xv <0) { xv += 6; }
               xv += 6*zv; /* 48 */
               zv = 0;
               break;
             case 3:  
               zv -= 2*(((xv%24+(xv%24<0?24:0))/6));
               zv %= 8; if (zv < 0) { zv += 8; }
               xv %= 6; if (xv < 0) { xv += 6; }
               xv += 6*zv; /* 48 */
               zv = 0;
               break;
             case 2:  
               zv &= 0x01;
               xv %= 12; if (xv < 0) { xv += 12; }
               xv += 12*zv; /* 24 */
               zv = 0;
               break;
             case 5:
             case 4:
               zv -= 2*(((xv%24+(xv%24<0?24:0))/6));
               zv %= 8; if (zv < 0) { zv += 8; }
               xv %= 6; if (xv < 0) { xv += 6; }
               xv += 6*zv; /* 48 */
               zv = 0;
               break;
             case 1:
               zv += 2*(((xv%12+(xv%12<0?12:0))/6));
               zv %= 8; if (zv < 0) { zv += 8; }
               xv %= 6; if (xv < 0) { xv += 6; }
               xv += 6*zv; /* 48 */
               zv = 0;
               break;
             case 6:
               zv %= 4; if (zv < 0) { zv += 4; }
               xv %= 12; if (xv < 0) { xv += 12; }
               xv += 12*zv; /* 48 */
               zv = 0;
               break;
         }
         how = hextetragone[TPat->Vals.Strata.Variant][xv];
         break;
      case 3:
         switch(TPat->Vals.Strata.Variant)
         {
             case 0:
               zv %= 6; if(zv <0) { zv += 6; }
               xv %= 6; if(xv <0) { xv += 6; }
               xv += 6*zv;
               zv = 0;
               break;
             case 1:
               zv += 2*(((xv%18+(xv%18<0?18:0))/6));
               zv %= 6; if(zv <0) { zv += 6; }
               xv %= 6; if(xv <0) { xv += 6; }
               xv += 6*zv;
               zv = 0;
               break;
             case 2:
               zv &= 0x01;
               xv %= 18; if (xv < 0) { xv += 18; }
               xv += 18*zv;
               zv = 0x00;
               break;
         }
         how = hextrigone[TPat->Vals.Strata.Variant][xv];
         break;
      case 2:
         zv &= 0x01;
         xv %= 6; if (xv < 0) { xv += 6; }
         how = hexdigone[TPat->Vals.Strata.Variant][xv+6*zv];
         break;
      case 1:
      default:
         zv &= 0x01;
         xv += 3*zv;
         xv %= 6; if (xv <0) { xv += 6;}
         lng = 0;
         how = hexmonogone[TPat->Vals.Strata.Variant][xv];
      break;
   }

   /* 
   **     / \
   **    1   \
   **   /     2
   **  /       \
   **  -----3-----
   ** height = sqrt(3)/2
   */
   if (how & 0x01)
   {
      dist1 = 1.0 - (fabs(SQR3 * x - z) * SQR3 );
      return_value = max(return_value,dist1);
   }
   if (how & 0x02)
   {
      dist2 = 1.0 - (fabs(SQR3 * (1.0-x) - z) * SQR3  );
      return_value = max(return_value,dist2);
   }
   if (how & 0x04)
   {
      dist3 = 1.0 - (z * 2.0 * SQR3 );
      return_value = max(return_value,dist3);
   }
 switch(TPat->Vals.Strata.Measure_int) 
 {
 case 1:
      dist1 = (1.0 - (fabs(SQR3 * z + x) ));
      dist2 = (1.0 - (fabs(SQR3 * z - x + 1.0) ));
      dist3 = (1.0 - (x * 2.0 ));
   if ((((how & 0x83) == 0x00)||((how & 0x4000) == 0x4000))
     &&(dist1<0)&&(dist2<0)) 
   {
      value1 = (3.0 / 2.0 *(fabs(SQR3 * z + x) ) - 2.0);
      value2 = (3.0 / 2.0 *(fabs(SQR3 * z - x + 1.0) ) - 2.0);
      value =  min(value1,value2);
      return_value = max(return_value,value);
   }
   if ((((how & 0x85) == 0x00)||((how & 0x2000) == 0x2000))
     &&(dist1>0)&&(dist3>0))
   {
      value1 = (1.0 - 3.0 / 2.0 * (fabs(SQR3 * z + x) ));
      value2 = (1.0 - (x * 3.0 ));
      value =  min(value1,value2);
      return_value = max(return_value,value);
   }
   if ((((how & 0x86) == 0x00)||((how & 0x1000) == 0x1000))
     &&(dist3<0)&&(dist2>0))
   {
      value1 = (1.0 - 3.0 / 2.0 *(fabs(SQR3 * z - x + 1.0) ));
      value2 = ((x * 3.0 ) - 2.0);
      value =  min(value1,value2);
      return_value = max(return_value,value);
   }
   break; 
 case 2:
   if (((how & 0x83) == 0x00)||((how & 0x4000) == 0x4000))
   {
      dist1 = x - 0.5;
      dist2 = z - SQR3_2;
      dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2))*3.0 );
      return_value = max(return_value,dist3);
   }
   if (((how & 0x85) == 0x00)||((how & 0x2000) == 0x2000))
   {
      dist1 = x;
      dist2 = z;
      dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2)) *3.0);
      return_value = max(return_value,dist3);
   }
   if (((how & 0x86) == 0x00)||((how & 0x1000) == 0x1000))
   {
      dist1 = x - 1.0;
      dist2 = z ;
      dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2)) *3.0);
      return_value = max(return_value,dist3);
   }
   break; 
 case 0:
   if (((how & 0x83) == 0x00)||((how & 0x4000) == 0x4000))
   {
      dist3 = 1.0 - ((SQR3_2 - z) * 2.0 * SQR3  );
      return_value = max(return_value,dist3);
   }
   if (((how & 0x85) == 0x00)||((how & 0x2000) == 0x2000))
   {
      dist2 = 1.0 - (fabs(SQR3 * x + z) * SQR3 );
      return_value = max(return_value,dist2);
   }
   if (((how & 0x86) == 0x00)||((how & 0x1000) == 0x1000))
   {
      dist1 = 1.0 - (fabs(SQR3 * (x -1.0) - z) * SQR3 );
      return_value = max(return_value,dist1);
   }
   break; 
 }
      dist1 = (1.0 - (fabs(SQR3 * z + x) ));
      dist2 = (1.0 - (fabs(SQR3 * z - x + 1.0) ));
      dist3 = (1.0 - (x * 2.0 ));                                               
   switch(TPat->Vals.Strata.Measure_ang)
   { 
   case 2:
   if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
   {
      value1 = x;
      value2 = z;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x140) == 0x140)&&(dist1>0)&&(dist3>0))
   {
      value1 = x - 0.5;
      value2 = z- SQR3_2;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x210) == 0x210)&&(dist1<0)&&(dist2<0))
   {
      value1 = x - 1.0;
      value2 = z;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x240) == 0x240)&&(dist3<0)&&(dist2>0))
   {
      value1 = x - 0.5;
      value2 = z - SQR3_2;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x410) == 0x410)&&(dist1>0)&&(dist3>0))
   {
      value1 = x - 1.0;
      value2 = z;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x420) == 0x420)&&(dist3<0)&&(dist2>0))
   {
      value1 = x;
      value2 = z;
      value = 2.0*SQR3*(sqrt(value1*value1+value2*value2 ) - SQR3/3.0 );
      value = max(value,0.0);
      return_value = min(value,1.0);
      return return_value;
   }
   break;
   case 1:
   if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
   {
      value = -dist1 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x140) == 0x140)&&(dist1>0)&&(dist3>0))
   {
      value = dist1 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x210) == 0x210)&&(dist1<0)&&(dist2<0))
   {
      value = -dist2 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x240) == 0x240)&&(dist3<0)&&(dist2>0))
   {
      value = dist2 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x410) == 0x410)&&(dist1>0)&&(dist3>0))
   {
      value = dist3 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x420) == 0x420)&&(dist2>0)&&(dist3<0))
   {
      value = -dist3 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
   {
      value = -dist1 * 2;
      return_value = min(value,1.0);
      return return_value;
   }
   break;
   default:
   case 0:
      break;
   }
   return return_value;
}
/* Fin de Carrelage */
